/** This file is part of LIDAR Analysis work carried as part of a PhD with Imperial College.  
 * Copyright (c) 2010 Oliver Ford
 *
 * The content, concept and implementation are the intellectual property of Oliver Ford but permission
 * is given for it's use and further development by Jakob Svensson or any person or organisation
 * directly given such permissions by Oliver Ford or by Jakob Svensson.
 */
package seed.minerva.lidar.auxiliary;

import algorithmrepository.Algorithms;
import oneLiners.AsciiMatrixFile;
import oneLiners.BinaryMatrixFile;
import oneLiners.OneLiners;
import mds.JetMDSFetcher;
import seed.minerva.DoubleImpl;
import seed.minerva.GraphicalModel;
import seed.minerva.diagnostics.ServiceManager;
import seed.minerva.lidar.dataProvision.KE9DDataProvider;

/** Attempt to use straylight pulses to get some constraint on the maximum at value and variation of the inter-channel timing */
public class TimingFromStraylight {

	public static final String outPath = "/work/lidr/ke9/postUpgrade/timingFromStray";
	public static final String pulseList = "/work/lidr/ke9/postUpgrade/dryRuns/plist.txt";
	public static final String tpulseFile = "/work/code/minerva/minerva-lidar/shapes/ke9/postUpgrade/tpulse-ke9-C25-ch0-fastIR.txt";
	
	public final int region[][] = new int[][]{ { -320, -260 }, { -200, -40 } };  
	

	public TimingFromStraylight(String[] args) {
		
		int pl[] = OneLiners.doubleToInt(AsciiMatrixFile.mustLoad(pulseList, true)[0]);
		
		
		// Gains: capital G is in (ADCU per [photoelectron per ADC bin]), little g is in (mV per [photoelectron per ns])
		double GCalcC25[] = new double[]{ 850, 415, 415, 360 };	// calculated for C25 region from xmasTree
		double gDataC25[] = new double[]{ 0.018507, 0.014703, 0.008318, 0.019103 }; // from data provider for C25 region

		
		for(int p=0; p < pl.length; p++){
			int pulse = pl[p];
		
			JetMDSFetcher jmds = ServiceManager.getInstance().getJETMDSFetcher();		
			KE9DDataProvider dataProv = new KE9DDataProvider(jmds, pulse, KE9DDataProvider.LD_SIGNAL | KE9DDataProvider.LD_BASELINE | KE9DDataProvider.LD_GAINS, 0, -1, -1);
			
			double baseline[][] = dataProv.baseLine();
			int nf = dataProv.nFrames();
			int nCh = dataProv.nChannels();
			int sigLen = dataProv.signalLength();
			
			double G[] = new double[nCh];
			double gains[] = dataProv.gains();				// mv per (photoelectron per ns)
			double ADCUnit = dataProv.ADCUnit();			// mv per ADCU  
			double signalDt = dataProv.signalDt() / 1e-9;	// dt in ns
				
			/* trust in the data provider
			for(int ch=0; ch < nCh; ch++){
				G[ch] = gains[ch] / ADCUnit / signalDt;
			}*/
			
			// we want to use the corrected GCalcC25, but the underlying 'gains' i.e the HV settings might be different to what they were in C25
			// for this pulse, so we use the ratio gDataC25 / gThisPulse to correct it
			for(int ch=0; ch < nCh; ch++)
				G[ch] = GCalcC25[ch] * gains[ch] / gDataC25[ch];


			double FWHM = 0.500 / signalDt; // FWHM of ins function is 500ps (which is 10 ADCU)
			double sigmaC = FWHM / (2*Math.sqrt(2*Math.log(2)));	// sigma of gaussian approx to ins func
			double cheese2 = 1 / (2 * sigmaC * Math.sqrt(Math.PI)); // multiplier for reduction of noise due to smoothing of ins function - 'Factor of cheeses'

			double sigmaElec = 2 / ADCUnit; //chain1's value for elec noise is 2mV which is about right
			
			
			for(int f=0; f < nf; f++){
				double sig[][] = dataProv.signalFrame(f);
				double sigma2[][] = new double[nCh][sigLen];
				
				//find first saturated point on 3rd chan (~backwall)  
				int satIdx = -1;
				for(int i=1001;i < 1500; i++)
					if(sig[2][i] > 65000){
						satIdx = i;
						break;
					}
				
				if(satIdx <= 0){
					System.err.println("p"+p+"("+pulse+"), f"+f+": No or invalid saturation point, dropping frame");
					continue;
				}	
				
				//subtract baseline, convert to photoelecs per ADCU and calc sigma
				for(int ch=0; ch < nCh; ch++){
					for(int i=0; i < sigLen; i++){
						sig[ch][i] = (sig[ch][i] - baseline[ch][i]) / G[ch];
						sigma2[ch][i] = cheese2 * sig[ch][i] + (sigmaElec*sigmaElec/G[ch]/G[ch]);
					}
				}
				
				BinaryMatrixFile.mustWrite(outPath + "/signal-p"+p+"-f"+f+".bin", sig, true);
				BinaryMatrixFile.mustWrite(outPath + "/sigma2-p"+p+"-f"+f+".bin", sigma2, true);
				OneLiners.TextToFile(outPath + "/satIdx-p"+p+"-f"+f+".txt", Integer.toString(satIdx)+"\n");
				
				int nSigma = 3; // number of sigma above 0 to consider as straylight

				//we want to look for something in the regions that might be interesting, for anything which is above 3.sigma of 0 for at least 2 of the first three channels.
				for(int i=0; i<region.length; i++){
					int j0 = Math.max(satIdx + region[i][0], 0);
					int j1 = Math.min(satIdx + region[i][1], sigLen-1);
					int chCount = 0;
					int chPointCount[] = new int[nCh];
					for(int ch=0; ch < 3; ch++){
						chPointCount[ch] = 0;
						for(int j=j0; j <= j1; j++){
							if(sig[ch][j] > nSigma * Math.sqrt(sigma2[ch][j]))
								chPointCount[ch]++;
						}
						if (chPointCount[ch] >= 5)
							chCount++;
					}
					if(chCount > 0){
						System.out.print("Region "+i+" ("+region[i][0]+" - "+region[i][1]+") has " + chCount + " channels out: ");
						for(int ch=0; ch < nCh; ch++)
							System.out.print(chPointCount[ch] + ", ");
						System.out.println();
						
						for(int ch=0; ch < nCh; ch++)
							if (chPointCount[ch] >= 5 ){
								double d[] = java.util.Arrays.copyOfRange(sig[ch], region[i][0], region[i][1]);
								double s2[] = java.util.Arrays.copyOfRange(sigma2[ch], region[i][0], region[i][1]);
								double fit[] = fitRegion(d, s2);
							}
					}
				}

				if(true)return;
			}

		}
		
	}
	
	
	public double[] fitRegion(double data[], double sigma2[]){
		int n= data.length;
		
		BinaryMatrixFile.mustWrite(outPath + "/current-data.txt", data);
		BinaryMatrixFile.mustWrite(outPath + "/current-sigma2.txt", sigma2);
		
		//find the max in data for initial guess
		int tmp[] = new int[1];
		Algorithms.max(data, tmp);
		
		GraphicalModel g = new GraphicalModel();
		DoubleImpl A = new DoubleImpl(g, "A", data[tmp[0]]);
		DoubleImpl x0 = new DoubleImpl(g, "x0", tmp[0]);
		DoubleImpl y0 = new DoubleImpl(g, "y0", 0);
		
		
		
		return null;
	}
	
	public static void main(String[] args) { new TimingFromStraylight(args); }

}
