/** This file is part of LIDAR Analysis work carried as part of a PhD with Imperial College.  
 * Copyright (c) 2010 Oliver Ford
 *
 * The content, concept and implementation are the intellectual property of Oliver Ford but permission
 * is given for it's use and further development by Jakob Svensson or any person or organisation
 * directly given such permissions by Oliver Ford or by Jakob Svensson.
 */
package seed.minerva.lidar.dataProvision;


import mds.*;
import java.text.DecimalFormat;

import oneLiners.OneLiners;

/**
 * Provides nice interface for KE3 Core LIDAR signal and calibration data from the MDSPlus reader
 * and performs all associated data hacking and sillyness to get MDSPlus read KE3 signals
 * into a sensible format.
 *  
 * Generally speaking this is a mess, the reason is simply that the LIDAR data formats, calibrations and storage methods 
 * are a mess and this is out of my control.
 * 
 * DON'T READ FURTHER UNLESS YOU REALLY HAVE TO, YOU WILL NOT GET BACK THE HOURS OF YOUR LIFE YOU WASTE!
 * 
 * Most of this is a direct reverse engineering of chain1's RDJPF.F (/home/chain1/source/lidr/fortran/rdjpf.f on JAC in 2007)
 * although I've made some extra deductions about the way the format works rather than hard-coding values as it does.
 * 
 * Where RDJPF.F has hard coded values, the values this program reads are checked against them and a warning given if it's different.
 * 
 * @author oliford <codes@oliford.co.uk>
 *
 */

//TODO: Ought to make some effort to cope with when we don't have all 6 channels  
public final class KE3DataProvider implements LidarDataProvider {
	private int frameSize; //data points per frame
	private double dt = 0.2e-9; //sampling period in seconds 
	final private int nChans = 6; //defined, because the TVS/JPF structures seem to be stuck with this!
	
	final private double adcuMax = 1024;
	private double mvMax; //max range in mV - corresponds to full ADCU range
	
	private int warningLevel = 0; //highest severity of warning issued 1-5
	
	private JetMDSFetcher jmds;
	private int pulse;
	private int nFrames;
	private double rawSig[][][]; //signal [frame][chIdx][]
	private double baseline[][]; //baseline frame [chIdx][]
	private boolean chanEnabled[]; //channel enable flags
	private boolean offline = false;
	private String baseStr = ""; 

	private double frameTimes[]; //laser pulse times
	private double normSpect[][][]; //normalised detector spectrums [ch][wave/amp][]
	private double detSpect[][][]; //detector spectrums * relSens [ch][wave/amp][]
	private double vignetting[][]; //vignetting curve [R/amp][]
	private double gain[]; //gain of channel
	private double timing[]; //tpulse to backwall times
	private double relSens[]; //relative sensitivities
	private double laserEnergy[]; //laser energy for each frame
	
	
	private double ppfne[][],ppfTe[][],ppfR[],ppfZ[],ppfTimes[];
	
	/** Create an uninitialised KE3 provider. Must call setPulse() before use */ 
	public KE3DataProvider(JetMDSFetcher jetMDS){ this.jmds = jetMDS; this.pulse = -1; }
	
	/**
	 * Creates a lidrDataProvider object that reads from the provided MDSPlus object
	 * Loads all data.
	 * @param jmds Existing JET MDSPlus data reader object.
	 * @param pulseNo JET Pulse Number.
	 */
	public KE3DataProvider(JetMDSFetcher jetMDS, int pulseNo){ this(jetMDS, pulseNo, 65535, 0, -1, -1); }
	
	public KE3DataProvider(JetMDSFetcher jetMDS, int pulseNo, int loadMask, int startFrame, int maxFrames, int singleChan){ this(jetMDS,pulseNo,loadMask,startFrame,maxFrames,singleChan,false); }
	
	public KE3DataProvider(JetMDSFetcher jetMDS, int pulseNo, boolean offline){ this(jetMDS, pulseNo, 65535, 0, -1, -1, offline); }
	
	/**
	 * Creates a lidrDataProvider object that reads from the provided MDSPlus object
	 * Only loads the data specified in loadMask.
	 * @param jmds Existing JET MDSPlus data reader object.
	 * @param pulseNo JET Pulse Number.
	 * @param loadMask Data to preload, ( mask from lidrDataProvider.LD_* )
	 */	
	public KE3DataProvider(JetMDSFetcher jetMDS, int pulseNo, int loadMask, int startFrame, int maxFrames, int singleChan, boolean offline){
		init(jetMDS, pulseNo, loadMask, startFrame, maxFrames, singleChan, offline);
	}
	
	/** Change pulse (reinit with same MDS+ provider and all frames) - provided for Minerva compatibility*/
	public void setPulse(int pulse){
		init(jmds, pulse, ~0, 0, -1, -1, offline);
	}
	
	protected void init(JetMDSFetcher jetMDS, int pulseNo, int loadMask, int startFrame, int maxFrames, int singleChan, boolean offline){
		
		knownIssues(pulseNo); //display warnings if we're attempting to read pulses with specific known problems
			
		this.jmds = jetMDS;
		this.pulse = pulseNo;
		this.offline = offline;
		this.baseStr = offline ? "jet-offline/" : "jet/";
		boolean bl,sig;
		sig = (loadMask & LD_SIGNAL) != 0;
		bl = (loadMask & LD_BASELINE) != 0;
		
		if((loadMask & LD_ENERGY_MON) != 0 && !sig)
			throw new IllegalArgumentException("LidrDataProv ERROR: Can't load energy monitor without loading signal.");		
		
		if(sig || bl)
			readSignal(sig,bl,startFrame,maxFrames,singleChan);
		if((loadMask & LD_SPECTRA) != 0)readDetSpectrums();
		if((loadMask & LD_VIGNET) != 0)readVignetting();
		if((loadMask & LD_GAINS) != 0)readGains();
		if((loadMask & LD_TIMING) != 0)readTiming();		
		if((loadMask & LD_PROFILES) != 0)readProcessedProfiles();
		if((loadMask & LD_ENERGY_MON) != 0)readEnergyMonitor();
		
		//System.gc(); //generally a good idea at this point
	}
	
	
	/** Get KE3_CMA1 from MDSPlus, break into frames and digitsers and remove 
	 * interpolated points.	 */
	private void readSignal(boolean doSignal, boolean doBaseline, int startFrame, int maxFrames, int singleChan){		
		//read the ke3_cma1 node using the special raw functions (so we can see the header)
		//The normal reading method returns the data with some of the processing done, but this only work for recent shots
		//so this now does everything it self regardless
		byte data[] = jmds.getRawJPF(baseStr + pulse + "/jpf/dc/ke3-cma1/000#raw");
		
		
		//first in the data is the JPF header...    // Chain1 Name - Description - Normal Value
		int JPFHeaderSize = UW(data[1],data[0]); 	// LenHed - Length of sructure - 19
		int maxChans = UW(data[3],data[2]);			// TexMax - Max No. ADCs - 7
		int TekEnb = UW(data[5],data[4]);			// TekEnb - Unknown (Tektronix Enable?) but < 64 means different structure - 64
		int frameArraySize = UW(data[7],data[6]);	// MPulse - Max frames (suspect this is size of TIMLAS and SHIFT??) - 200
		nFrames = UW(data[9],data[8]);				// CACPUL - No. frames (pulses from 'CAC'??) - ~140
		int yScale = UW(data[23],data[22]);			// IYSCA x 7 - ? - 0 0 0 0 0 0 500
		int xScale = UW(data[37],data[36]);			// IXSCA x 7 - ? - 0 0 0 0 0 0 512
		mvMax = yScale;
		
		//do some 'unusual situations' checks
		if( JPFHeaderSize != 19)
			warning(5,"The JPF header isn't 19 words long, this data provision will probably break horribly. Is this an invalid or very old pulse?");			
		if( maxChans < 6)
			throw new RuntimeException("LidrDataProv ERROR: Pulse doesn't have all 6 digitisers, can't cope with this yet");			
		if( maxChans != 7)
			warning(5,"The JPF Header doesn't say max of 7 digitisers, the data provision code may fail!");			
		if( TekEnb < 64 )
			throw new RuntimeException("LidrDataProv ERROR: This is apparently \"a non-TVS JPF\", I'm not sure what this means - see chain1's RDJPF.F!!");
		if( nFrames > frameArraySize){
			warning(4,"Apparently we have more frames ("+nFrames+") than can fit in the frame times data area ("+frameArraySize+"). I'll continue, but using the data size.");
			nFrames = frameArraySize;
		}
		if( frameArraySize != 200)
			warning(3,"The JPF header doesn't report 200-long time arrays, this is unusual.");
		
		
		boolean haveBaseline = false;;
		
		/*next are the two frame time arrays:		 
		*   TIMLAS x 200 - Laser(frame) times, UNSIGNED , x10^-3 after start time
		*   SHIFT x 200, - Small shift to laser times, SIGNED, x10^-6
		*   I suspect this 200 is actually MPulse i.e 'Max Pulses' so will use the variable here but throw a warning if it's not 200
		*   As Chain1 seems to always use 200 
		*/

		//start time comes from a different JPF node - this changed to early 2008 to be in the time vector, we'll check both anyway
		double startData[] = jmds.get1DData(baseStr + pulse + "/jpf/dc/e3-stw>sta");
		double startTime[] = jmds.getTimeVector(baseStr + pulse + "/jpf/dc/e3-stw>sta"); 
		double start;
		if(startData[0] < 20 && startTime[0] >= 20)
			start = startTime[0];
		else if(startData[0] >= 20 && startTime[0] < 20)
			start = startData[0];
		else{
			start = (startData[0] > startTime[0]) ? startData[0] : startTime[0];
			warning(3, "Both or niether of the data and time vector of e3-stw>sta (KE3 stopwatch start time) are > 20s. Will just use the biggest ("+start+")");
		}			
		
		int timlasIdx = JPFHeaderSize * 2;
		int shiftIdx = (JPFHeaderSize + frameArraySize) * 2;
		
		if( W(data[timlasIdx+1],data[timlasIdx]) == -1 ){ //if the first time entry is -1, it's the baseline
			haveBaseline = true; 
			nFrames--; //adjust the number of frames
			timlasIdx += 2;	shiftIdx += 2; //and move to the real first frame
		}
			
		//build each frame time
		frameTimes = new double[nFrames];
		for(int i=0;i<nFrames;i++){			 
			int timlas = UW(data[timlasIdx+1],data[timlasIdx]); timlasIdx += 2;
			int shift = W(data[shiftIdx+1],data[shiftIdx]); shiftIdx += 2;		
			frameTimes[i] = timlas * 1E-3 + shift * 1E-6 + start;
		}
		
		//next there are maxChans (always 7) pointers, all of which seem to point to directly after the pointers, but chain1 seems to just take the 7th
		//it's likely the 7th channel in some way indicates the presence of the new TVS style headers, whatever they are
		int ptrIdx = (JPFHeaderSize + 2*frameArraySize) * 2; 
		int ptrs[] = new int[maxChans];
		for(int i=0;i<maxChans;i++){
			ptrs[i] = I(data[ptrIdx+1],data[ptrIdx+0],data[ptrIdx+3],data[ptrIdx+2]); //maybe this ought to be unsigned, but the
			ptrIdx += 4;																// highest order 2 bytes are usually 0 anyway
			ptrs[i] *= 2; //multiply by 2 because we're working in 8-bit and it's in 16-bit
		}
		
		//the structure pointed to is now the "new TVS Structure"...
		int nFramesTVS = UW(data[ptrs[6]+1],data[ptrs[6]+0]);		//we have another frame count here for some reason
		int TVSHeaderSize = UW(data[ptrs[6]+3],data[ptrs[6]+2]);	//this is always 402, which i'm guessing is the TVS header size, as that's also always 402
		int chanMask = UW(data[ptrs[6]+5],data[ptrs[6]+4]);			//channel enabling bit mask
		
		int dataStart = ptrs[6] + (TVSHeaderSize*2);				//now we know the start of the actual data, in bytes 
		
		if(TVSHeaderSize != 402)
			warning(4,"TVS Structure isn't 402 words long. This is unusual, things may break.");
	 
		int xScaleTVS[] = new int[nChans];			// Real digitiser scales
		int yScaleTVS[] = new int[nChans];
		
		for(int i=0;i<nChans;i++){
			xScaleTVS[i] = UW( data[ptrs[6] + (i*2) + 7], data[ptrs[6] + (i*2) + 6] );  
			yScaleTVS[i] = UW( data[ptrs[6] + (i*2) + 19], data[ptrs[6] + (i*2) + 18] );			
		}
		
		//nframes check between TVS and JPF headers
		if(haveBaseline)nFramesTVS--;
		
		if(nFramesTVS < nFrames){
			warning(2,"TVS Structure's frame count ("+nFramesTVS+") is less than the JPF header's copy ("+nFramesTVS+"). I'll use the smallest.");			
			nFrames = nFramesTVS;			
		}
		
		//check which channels are enabled and work out where they will be in the data
		chanEnabled = new boolean[nChans];
		int chanIdx[] = new int[nChans]; //this will hold the position in the data of each channel (as they drop back when channels are turned off)
		int bitVal = 1;
		int curIdx = 0;
		int nEnabledChans = 0;
		//for each channel...
		for(int i=0;i<nChans;i++){
			chanEnabled[i] = (chanMask & bitVal) != 0; //see if this channel is enabled
			bitVal *= 2;	
			if(chanEnabled[i]){					// if so...
				chanIdx[i] = curIdx++;			//it's the next in the data
				nEnabledChans++;				
			}else
				warning(2,"Channel "+i+" is disabled.");
		}
		
		//work out the frame size
		double x = data.length;					//start with the data length, in bytes
		x -= dataStart; 						//subtract the start of the data, in bytes
		x /= 2;									//convert to number of 16-bit chunks		
		x /= nEnabledChans;						//divide by number of channels enabled
		x /= nFrames + (haveBaseline ? 1 : 0);	//divide by number of frames (including the baseline, if we had one)
		
		//this should now be the integer frame size
		frameSize = (int)Math.floor(x);
		if(x != frameSize)
			warning(5,"Calculated frame size is non-integer ("+x+")! Something is wrong.");
		
		x = Math.log(frameSize) / Math.log(2);
		if(x != Math.floor(x))
			warning(4,"Calculated frame size is not a power of 2 ("+x+")!, Something is wrong.");
		
		if(startFrame < 0) //Anything -ve means start at start
			startFrame = 0;
		
		if(startFrame >= nFrames)
			throw new IllegalArgumentException("Requested start at frame " + startFrame + " but only have " + nFrames + " frames.");
		
		//apply any maximum frame count truncation request
		if(maxFrames > 0){ 
			if((startFrame + maxFrames) < nFrames)
				nFrames = maxFrames;
			else
				nFrames = nFrames - startFrame;
		}
				
		//now we have to undo the annoying/pointless/irritating interpolation someone applies before we get hold of the data! 
				
		//pick a frame somewere in the middle and calc it's start point in the data
		int frame,ch;
		if(singleChan>=0)ch = singleChan; else ch=2;
		frame=nFrames / 2;
		
		//get the first index of this channel and this frame
		int j0 = dataStart + (((frame + (haveBaseline?1:0)) * nEnabledChans * frameSize) + (ch * frameSize) ) * 2;		
	
		double evenMatch = 0, oddMatch = 0;
		
		//we need to work out which set
		for(int j=0;j<(frameSize-3);j+=2){
			//calculate even and odd points as average of ones either side
			//if this is equal to what's actually there, increase the match count
			//for that set.
			int d0 = UW(data[j0 + ((j+0)*2) + 1],data[j0 + ((j+0)*2)]); //indexed point
			int d1 = UW(data[j0 + ((j+1)*2) + 1],data[j0 + ((j+1)*2)]); //next point
			int d2 = UW(data[j0 + ((j+2)*2) + 1],data[j0 + ((j+2)*2)]); //3rd point
			int d3 = UW(data[j0 + ((j+3)*2) + 1],data[j0 + ((j+3)*2)]); //4th point
			
			oddMatch += ( (d0+d2)/2.0 == d1 ) ? 1:0; 
			evenMatch += ( (d1+d3)/2.0 == d2 ) ? 1:0;
		}
		
		boolean interpolated = false;
		int offset = 0;
		
		if( oddMatch > ((4 * frameSize)/10) ){ //if 4/5 of the tests matched then the odd data values are probably interpolated 
			System.out.println("LidrDataProv Notice: Odd data appears to be interpolated, removing it. This is normal.");
			interpolated = true;		
			offset = 0; //odd are interpolated from even, so use the even ones
		}
		
		if( evenMatch > ((4 * frameSize)/10) ){ //if 4/5 of the tests matched then the evem data values are probably interpolated
			warning(2,"Even data appears to be interpolated, removing. Normally it's the odd ones but I'll continue anyway."); //not a problem really
			if(interpolated) //if the other test matched aswell... we have a big problem
				throw new RuntimeException("LidrDataProv ERROR: I seem to think both even and odd data are interpolated from each other. This is very wrong! Are they all 0?");
			interpolated = true;
			offset = 1; //even are interpolated from odd, so use the odd ones
		}
		
		if(!interpolated)
			warning(3,"LidrDataProv WARNING: Niether odd or even points seem to be interpolated, this is unusual");
				
		if(doSignal){
			//copy the even/odd points of each frame into appropriate place in rawSig
			rawSig = new double[nFrames][nChans][]; //this is defined as real channel indicies
			for(frame=0;frame < nFrames;frame++)
				for(ch=0;ch < nChans;ch++){
					if(singleChan>=0 && singleChan!=ch)continue;	//ignore all but the single channel, if we're only loading one
					
					//calc frame start position in data					
					j0 = dataStart + (((startFrame+frame+(haveBaseline?1:0)) * nChans * frameSize) + (chanIdx[ch] * frameSize)) * 2;
					if(interpolated){
						rawSig[frame][ch] = new double[frameSize/2];	//define the array
						for(int j=0;j < (frameSize/2);j++){
							int k = 2*j + offset;
							try{
							rawSig[frame][ch][j] = 1024 - UW(data[j0+(2*k)+1],data[j0+(2*k)]);
							}catch(ArrayIndexOutOfBoundsException e){
								System.err.println("frame="+frame+", ch="+ch+", j0="+j0+", k="+k);
							
								throw e;
							}
						}
					}else{
						rawSig[frame][ch] = new double[frameSize];	//define the array
						for(int j=0;j < frameSize;j++)
							rawSig[frame][ch][j] = 1024 - UW(data[j0+(2*j)+1],data[j0+(2*j)]);						
					}
				}
		}
		
		//and do the same with the baseline, if we have one
		if(doBaseline){
			if(!haveBaseline)
				throw new RuntimeException("No baseline frame in JPF but it was requested, don't know what todo!");

			baseline = new double[nChans][];
			for(ch=0;ch < nChans;ch++){
				if(singleChan>=0 && singleChan!=ch)continue;

				//calc frame start position in data					
				j0 = dataStart + (chanIdx[ch] * frameSize) * 2;
				if(interpolated){
					baseline[ch] = new double[frameSize/2];	
					for(int j=0;j < (frameSize/2);j++){
						int k = 2*j + offset;						
						baseline[ch][j] = 1024 - UW(data[j0+(2*k)+1],data[j0+(2*k)]);
					}
				}else{
					baseline[ch] = new double[frameSize];
					for(int j=0;j < frameSize;j++)
						baseline[ch][j] = 1024 - UW(data[j0+(2*j)+1],data[j0+(2*j)]);						
				}
			}
		}
		
		if(interpolated)
			frameSize /= 2;
		
		//we won't need this signal again, and it's rather big
		jmds.freeFromMemory(baseStr + pulse + "/jpf/dc/ke3-cma1/000#raw");

		if(pulse >= 69233 && pulse <= 69234)
			for(frame=0;frame < nFrames;frame++){
				double tmp[] = rawSig[frame][3];
				rawSig[frame][3] = rawSig[frame][5];
				rawSig[frame][5] = tmp;
			}
		
		
	}
	
	private void readTiming(){
		timing = new double[nChans];		
		
		/* investigation of what sweep linearity is, ignore
		 * byte a[] = jmds.getRawJPF(offlineStr + pulse + "/jpf/dc/ti-sweep#raw");		
		for(int ch=0;ch<nChans;ch++){
			System.out.println(ch+"\t"+ W(a[ch*2+1],a[ch*2]));
		}*/
		
		byte tishiftData[] = jmds.getRawJPF(baseStr + pulse + "/jpf/dc/ti-shift#raw");
		
		for(int ch=0;ch<nChans;ch++){
			timing[ch] = W(tishiftData[ch*2+1],tishiftData[ch*2]) * 1e-12;
		}
	}
	
	/** Read the Photomultiplier's HV setting, the gains table and the single photoelectron calibration */ 
	private void readGains(){
		
		gain = new double[nChans];
		
		double hv[],spho;
		double table[][] = new double[nChans][50];
	
		for(int ch=0;ch<nChans;ch++){
			//get gains for each HV setting HV from JPF ke3_spho{ch}, base for this is 3200 - (i*50)
			byte gainData[] = jmds.getRawJPF(baseStr + pulse + "/jpf/dc/ke3-spho"+(ch+1)+"#raw");
			
			for(int i=0;i<gainData.length;i+=2){
				table[ch][i/2] = W(gainData[i+1],gainData[i]);				
			}
			
			
			//then get current hv setting ke3_hv{ch}, (for some reason, this one works as a proper JPF, and gives sensible format numbers!
			hv =  jmds.get1DData(baseStr + pulse + "/jpf/dc/ke3-hv"+(ch+1));
			
			//use value of HV to find index into the gains table
			//the gains table, staring from position 2 (3rd) entry, runs from 3200 in -50's
			
			double f = (3200 - Math.abs(hv[0])) / 50;
			int s0 = (int)f;
			f -= s0;			
			double v0 = UW(gainData[(s0+2)*2 + 1],gainData[(s0+2)*2]);
			double v1 = UW(gainData[(s0+3)*2 + 1],gainData[(s0+3)*2]);
			//the actual gain is then the value/30000, according to chain1's ASPHEL.F
			gain[ch] = v0 + (v1 - v0) * f;
			gain[ch] /= 30000;
			
			//the single photoelectron value comes from elements 0 and 1 of the SPHO table:
			//according to chain1's RDJPF.F, this is (mV*ns) for 1 photoelectron,			
			//which is mV per (photoelectron per nanosecond) 
			spho = UW(gainData[1],gainData[0]) * Math.pow(10,W(gainData[3],gainData[2])); 						
			//this is presumably per photoelectrons hitting the annode... 
			
			// so we multiply the gain by this, to get our overall conversion from 
			// mv to photoelectrons per nanosecond leaving the photocathode.
			gain[ch] *= spho;
			
			//30/8/7 Manually checked against values read by chain1/ASPHEL.F
			//System.out.println("hv="+hv[0]+", spho="+spho+", gain="+gain[ch]);
		}	
		
		//AsciiMatrixFile.mustWrite("/work/scrap/gainTable.txt",table,true);
		
	}
	
	/** Read detector spectrums from KE3_WAVE and KE3_AMP */
	private void readDetSpectrums(){
		//Most of this is from ASLITF.F
		DecimalFormat fmt = new DecimalFormat("000");
		detSpect = new double[nChans][2][];
		normSpect = new double[nChans][2][];
		relSens = new double[nChans];
		
		byte waveData[],ampData[];

		for(int ch=0;ch<nChans;ch++){
			//At time of writing, we can't get this from the MDS+ tree directly, so use the raw JPF and convert.
			waveData = jmds.getRawJPF(baseStr + pulse + "/jpf/dc/ke3-wave<ch:"+fmt.format(ch+1)+"#raw");
			ampData = jmds.getRawJPF(baseStr + pulse + "/jpf/dc/ke3-amp<ch:"+fmt.format(ch+1)+"#raw");
			
			//first two values of ampData are value and exponent of relative sensitivity 
			relSens[ch] = W(ampData[1],ampData[0]) * Math.pow(10, W(ampData[3],ampData[2]));
			
			//find the first zero wavelength to give us our length (or use the full data length)
			int n;
			for(n=0;n<(waveData.length/2);n++)
				if( W(waveData[n*2 + 1],waveData[n*2]) == 0)
					break;
			
			
			double waveVals[] = new double[n];
			double ampVals[] = new double[n];
			
			//the first point of the spectrum is the first wavelength and an amplitude of 0
			waveVals[0] = UW(waveData[1],waveData[0]);
			ampVals[0] = 0;
						
			for(int i=1;i<(n);i++){ //now we start compiling entrues
				waveVals[i] = UW(waveData[i*2 + 1],waveData[i*2]); //wavelength are just stright in the table from 0 to n-1
				
				if( (i*2 + 3) < ampData.length) //if there is an amplitude (we arn't off the end) use it, otherwise just use 0 
					ampVals[i] = UW(ampData[i*2 + 3],ampData[i*2 + 2]); //amplitudes start from entry 1 at position 2
				else
					ampVals[i] = 0;
			}
			
			//last entry should really be 0, for integration to work. This might be in the data or we might've set it because there wasnt an amplitude
			if( ampVals[n-1] > 0 )
				warning(3,"Last amplitude value in spectum data is nonzero. " +
										"This is unusual and may cause problems building the temperature table.");
			
			//double winTransFactor[] = new double[]{ 1.00,0.98,0.97,0.95,0.914,0.883 };
			
			normSpect[ch][0] = waveVals;
			normSpect[ch][1] = ampVals;
			
			detSpect[ch][0] = normSpect[ch][0]; //these can be the same array
			detSpect[ch][1] = new double[n];			
			for(int i=0;i<n;i++)
				detSpect[ch][1][i] = normSpect[ch][1][i] * relSens[ch]; //* winTransFactor[ch];

		}
				
	}
	
	/** loads vignetting curve from KE3_FSTOP and KE3_REF */
	private void readVignetting(){
		//TODO: I think this comes from a file in chain1, ought to sort this out
		//Not that it's that beleivable anyway. 
		int i;
		byte refData[] = jmds.getRawJPF(baseStr + pulse + "/jpf/dc/ke3-ref#raw");
		byte vigData[] = jmds.getRawJPF(baseStr + pulse + "/jpf/dc/ke3-fstop#raw");
		
		//ke3_ref contains the start and step for the vignetting calibration positions
		double z0,zStep;
		z0 = (double)W(refData[1],refData[0]) / 1000.0;
		zStep = (double)W(refData[3],refData[2]) / 1000.0;
		
		//create the array and fill it
		vignetting = new double[2][vigData.length/2 - 2];
		for(i=0;i<(vigData.length/2 - 2);i++){
			vignetting[0][i] = z0 + i * zStep;			
			vignetting[1][i] = W(vigData[i*2 + 5],vigData[i*2+4]);
			
		}
		
	}	
	
	/** Read the laser energy monitor from LA_ENG<MON */
	private void readEnergyMonitor(){
		final double energySearchLength = 100e-3; //how far on the frame time to search for the laser energy
		final int baseFindLength = 10; //how far from start, in data points to use to calculate the baseline level and time step
					
		//the version of this in RDJPF.F seems to read from a 'calibration block' which I don't seem to get here, even with a jpfraw() call
		//however, the call directly from the jpf tree seems to work properly for this	
		try{
			double monitorData[] = jmds.get1DData(baseStr + pulse + "/jpf/dc/e3-la-eng<mon:004");
			double monitorTime[] = jmds.getTimeVector(baseStr + pulse + "/jpf/dc/e3-la-eng<mon:004");
			
			//some basic sanity checks
			if(monitorTime.length < 100)
				throw new RuntimeException("Laser energy monitor has less than 50 points, abandoning.");
			if(monitorTime[0] <= 10)		//check the time point is vaguely sensible
				warning(2,"First laser energy monitor time point is "+monitorTime[0]+". I expect it to be at least > 10s.");
			
			//find the time spacing and baseline by averaging the first 10 points 
			double dt = 0, base = 0;
			for(int i=0;i<baseFindLength;i++){
				base += monitorData[i];
				dt += monitorTime[i+1] - monitorTime[i];
			}
			dt /= baseFindLength; base /= baseFindLength;
			
			if(dt < 9.9e-3 || dt > 10.1e-3) 
				warning(2,"The laser energy monitor does not appear to be at 10ms intervals, this is unusual.");
			
			laserEnergy = new double[nFrames];
			
			for(int i=0;i<nFrames;i++){ //for each frame
				//find the entry corresponding to the start time of this frame
				int j0 = (int)((frameTimes[i] - monitorTime[0]) / dt);
				
				if(j0 <= baseFindLength)
					warning(2,"Frame "+i+"'s laser time falls at data point "+j0+" in the laser energy data. This is before/during the window " +
							"I'm using to find the base level. The laser energy might be wrong and hence the density calibration incorrect but I'll continue anyway.");
				
				//set laserEnergy to the highest point in our search window from here
				laserEnergy[i] = 0;
				for(int j=0;j<(energySearchLength/dt);j++)
					if( monitorData[j0+j] > laserEnergy[i])
						laserEnergy[i] = monitorData[j0+j];
							
				laserEnergy[i] -= base;
			}
			
		}catch(RuntimeException err){ //any errors in here are hopefully not fatal. They will however mean there isn't any laser energy monitor
			warning(1,"LidrDataProv WARNING: Aborted reading laser energy monitor because: " + err.getMessage());
			laserEnergy = null;
		}
				
	}
	
	/** loads Te and ne profiles from standard analysis PPF */
	private void readProcessedProfiles(){
		if(offline){
			warning(1, "This is an offline pulse, cannot read PPF profiles.");
			ppfne = null; ppfTe = null; ppfZ = null; ppfR = null; ppfTimes = null;		
		 	return;
		}
		try{
			ppfne = jmds.get2DData("jet/" + pulse + "/ppf/lidr/ne");
			ppfTe = jmds.get2DData("jet/" + pulse + "/ppf/lidr/te");
			double Ztmp[][]  = jmds.get2DData("jet/" + pulse + "/ppf/lidr/z");
			ppfZ = Ztmp[0];
			ppfR = jmds.getXVector("jet/" + pulse + "/ppf/lidr/ne");
			ppfTimes = jmds.getTimeVector("jet/" + pulse + "/ppf/lidr/ne");	
		}catch(RuntimeException err){ //any errors in here are hopefully not fatal. They will however mean there isn't any laser energy monitor
			warning(1,"Aborted reading PPF profiles. Continuing, but they will not be avaliable. Reason was: " + err.getMessage());
			ppfne = null; ppfTe = null; ppfZ = null; ppfR = null; ppfTimes = null;			
		}
	}
	
	
	//Getters... the nice sensible clean and simple interface provided to anyone still sane
	/** @return the pulse number the provider was initilaised with */
	public final long getPulse() { return pulse;	}
	
	/** @param frame Frame number (i.e laser pulse index)
	 * @return 2D array: array of data points of the specified frame: signal[channel number][]
	 */
	public final double[][] signalFrame(int frame){	return rawSig[frame];	}
		
	/** @return Laser energy for all frames */
	public final double[] laserEnergies(){ return laserEnergy; }
	
	/** @return Laser energy for a specified frame */
	public final double laserEnergy(int frame){ return (laserEnergy==null)?1:laserEnergy[frame]; }
	
	/** @return signal for all channels in all frames: signal[frame][channel][] */
	public final double[][][] signalFull(){	return rawSig;	}
	
	/** @return baseline for all channels in all frames */ 
	public final double[][] baseLine(){ return baseline; }
	
	/** @return number of channels active. */
	public final int nChannels(){ return nChans; }
	
	/** @return number of frames (laser pulses) */
	public final int nFrames(){ return nFrames; }
	
	/** @return Frame (laser pulses) time points */
	public final double[] frameTimes(){ return frameTimes; }
	
	/** @return vignetting[R / data][data point] */
	public final double[][] vignetting(){ return vignetting; }
	
	/** @return the detector spectrums [ch][wave/amp][] - normalsied multiplied by relative sensitivity*/
	public final double[][][] spectrums(){  return detSpect;  }
	
	/** @return normalised detector spectrums [ch][wave/amp][]*/
	public final double[][][] normSpectrums(){  return normSpect;  }
	
	/** @return the relative sensitivities [ch]*/
	public final double[] relativeSensitivities(){  return relSens;  }
	
	/** @return signal conversion  in mv per (photoelectron per nanosecond). */
	public final double[] gains(){ return gain; }
	
	/** @return time between timing pulse and backwall in seconds */
	public  final double[] timing(){ return timing; }
	
	/** @return The ADC unit size in mV */
	public final double ADCUnit(){ return mvMax / adcuMax; }; //full 1024 range is 0.5V //TODO: HARDCODED : fish
	
	/** @return The maximum value the signal can take in ADC Units (The saturation level) */
	public final double signalMax(){ return adcuMax; }
	
	/** @return length of signal data array */
	public final int signalLength(){ return frameSize; }
	
	/** @return data's sampling period in seconds  */
	public final double signalDt(){ return dt; }
	
	
	/** @return The level of the most severe waning issued:
	 * 0=No problems
	 * 1=Some data is unavaliable (PPF or Laser Energies) 
	 * 2=Some unusual values were found, data may be suspect.
	 * 3=Some values were not sensible , but these may or may not have an impact on output, 
	 * 4=Some very unusual values were found/things were not as expected, data is very likely to not be sensible.
	 * 5=Severe irregularlites were seen, outputs are probably complete rubbish. (the provider will most likely crash before finishing if a level 5 is issued)
	 */
	public final int warningLevel(){ return warningLevel; }
		
	public final double[] ppfFramene(int ppfFrame){ return ppfne[ppfFrame]; }
	public final double[] ppfFrameTe(int ppfFrame){ return ppfTe[ppfFrame]; }
	public final double[] ppfR(){ return ppfR; }
	public final double[] ppfZ(){ return ppfZ; }
	
	/** @return the frame index into the PPF (their frame number) that sits nearest in time to a frame in our arrays, these do not necceasrily line up exactly. */ 
	public final int pffNearestFrame(int frame){
		double t = frameTimes[frame];		
		if(ppfTimes == null)return -1;
		return OneLiners.getNearestIndex(ppfTimes, t);
	}
	
	//annoying utility functions, used throughout to deal with silly MDS/JPF/TVS formats
	private final short W(byte highByte,byte lowByte){ return (short)(( (short)highByte) * 256 + ( ((short)lowByte) & 0xff)); }
	private final int UW(byte highByte,byte lowByte){ return (((int)highByte)& 0xff) * 256 + ( ((int)lowByte) & 0xff); }
	private final short UB(byte sbyte){	return (short)(((short)sbyte) & (short)0x00ff);  }
	private final long UI(byte b4,byte b3,byte b2,byte b1){	
		return (((long)b4) & 0xff) * 0x1000000 + (((long)b3) & 0xff) * 0x1000000 + (((long)b2) & 0xff) * 0x10000 + (((long)b1) & 0xff) * 0x100;
	}
	
	private final int I(byte b4,byte b3,byte b2,byte b1){	
		return (((int)b4) * 0x1000000) + (((int)b3) & 0xff) * 0x10000 + (((int)b2) & 0xff) * 0x100 + (((int)b1) & 0xff) * 0x1;
	}
	
	private final void warning(int level, String text){
		System.err.println("KE3DataProvider WARNING (pulse "+pulse+"): "+text+" (Level "+level+")");
		if(warningLevel < level)
			warningLevel = level;			
	}


	/** All manner of special cases are given for certain pulses in the Chain1 code. I'm only going to deal with those which
	 * cover large pulse ranges. Both those dealt with and those I can't be bothered with are listed here */ 
	private void knownIssues(int pulseNo){
		if(pulseNo >= 63700 && pulseNo <= 66000)
			System.err.println("WARNING: 63700 <= pulse <= 66000 have some known timing issues!");
		if(pulseNo >= 65112 && pulseNo <= 65975)
			System.err.println("WARNING: 65112 <= pulse <= 65975 have some known timing issues!");
		if(pulseNo >= 69233 && pulseNo <= 69234)
			System.err.println("WARNING: Pulses 69233/4 have some known issues! (swapped digitisers). I've swapped the data accordingly.");		 			
		if(pulseNo < 27500)
			System.err.println("WARNING: Should be using E3-MON rather than E3-STW for time start! This particular issue is " +
					"a bit beside the point at the moment, since the data format is completely off the map for < ~yr 2000 anyway.");
	}

}
