/** This file is part of LIDAR Analysis work carried as part of a PhD with Imperial College.  
 * Copyright (c) 2010 Oliver Ford
 *
 * The content, concept and implementation are the intellectual property of Oliver Ford but permission
 * is given for it's use and further development by Jakob Svensson or any person or organisation
 * directly given such permissions by Oliver Ford or by Jakob Svensson.
 */
package seed.minerva.lidar.dataProvision;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;

import descriptors.JET.JPFDesc;
import descriptors.JET.PPFDesc;
import oneLiners.*;
import signals.JET.JPF;
import signals.JET.PPF;
import mds.JetMDSFetcher;

/**
 * Provides nice interface for KE9D Edge LIDAR signal and calibration data from the MDSPlus reader.
 *  
 * Generally speaking, this is a lot more reasonable than {@link KE3DataProvider} as all the calibration info is parsed from a text header to the data.
 * 
 * Most of this is a direct reverse engineering of chain1's RDJPF.F (/home/chain1/source/ke9d/fortran/rdjpf.f on JAC in 2008)
 * 
 * @author oliford <codes@oliford.co.uk>
 *
 */
public class KE9DDataProvider implements LidarDataProvider {
	public static final int UPGRADE_FIRST_PULSE = 70750;
	
	private int frameSize = -1; //data points per frame
	final private int nChans = 4; //defined
	private int nFrames = 6;
	
	private double dt = Double.NaN; //sampling period in seconds	
	private double mVPerADCU = Double.NaN;
	private final int ADCUMaxRange = 65536; //Hardcoded - Doesn't seem to come from anywhere, but chain1 thinks 64000, we need it exact, which appears to be 65536
	
	private int warningLevel = 0; //highest severity of warning issued 1-5
	
	private JetMDSFetcher jmds;
	
	private int pulse;
	private double rawSig[][][]; //signal [frame][chIdx][]
	private double baseline[][]; //baseline frame [chIdx][]
	//private boolean chanEnabled[]; //channel enable flags
	private boolean offline = false;
	private String baseStr = "";

	private double frameTimes[]; //laser pulse times
	private double normSpect[][][]; //normalised detector spectrums [ch][wave/amp][]
	private double detSpect[][][]; //detector spectrums * relSens [ch][wave/amp][]
	private double vignetting[][]; //vignetting curve [R/amp][]
	private double gain[]; //gain of channel
	private double timing[]; //tpulse to backwall times
	private double relSens[]; //relative sensitivities
	private double laserEnergy[]; //laser energy for each frame

	private double ppfne[][],ppfTe[][],ppfR[],ppfZ[],ppfTimes[];
	private double R0,R1,Z0,Z1; 
	
	class CalibData{ //used to pass info between parseHeader(), parseAlternateCalib() and processCalib()
		double RVig0=Double.NaN,dRVig=Double.NaN;
		double vRanges[]=null; //channel range in Volts, default and expected is 400mV
		double tRange = Double.NaN;
		int nPoints = -1;
		double tShift=Double.NaN;
		double triggerTimes[]=null; //see LaserEnergyAndTimings()
		int hvSetting[] = null;		// HV settings from jpf/dc/ke9-hv#	
		
		double dt = Double.NaN;
		double VPerADCU[] = null;
				
		// Old HV setting - pre-upgrade (70750)
		double peVal[]=null; 	// Ch#PE in header
		int hvBias[]=null;		// HVBiadCh# in header
		
		// New HV setting - post-upgrade
		int pehv = -1; // 'PEHV' in header (1 value only)
		double hvCubicCoeffs[][]; // HV gain cubic coefficients [channel][coeff{0-2}]	('ch#F' in header) 
	};
	
	/** Create an uninitialised KE3 provider. Must call setPulse() before use */ 
	public KE9DDataProvider(JetMDSFetcher jmds){ this.jmds = jmds; this.pulse = -1;	}
	
	public KE9DDataProvider(JetMDSFetcher jmds, int pulse){ this(jmds, pulse, Integer.MAX_VALUE, 0, -1, -1); }
	
	public KE9DDataProvider(JetMDSFetcher jmds, int pulse, int loadMask, int startFrame, int maxFrames, int singleChan){ this(jmds,pulse,loadMask,startFrame,maxFrames,singleChan,false); }
	
	public KE9DDataProvider(JetMDSFetcher jmds, int pulse, boolean offline){ this(jmds, pulse, Integer.MAX_VALUE, 0, -1, -1, offline); }
	
	public KE9DDataProvider(JetMDSFetcher jmds, int pulse, int loadMask, int startFrame, int maxFrames, int singleChan, boolean offline){
		init(jmds, pulse, loadMask, startFrame, maxFrames, singleChan, offline);
	}
	
	/** Change pulse (reinit with same MDS+ provider and all frames) - provided for Minerva compatibility*/
	public void setPulse(int pulse){
		init(jmds, pulse, ~0, 0, -1, -1, offline);
	}
	
	protected void init(JetMDSFetcher jmds, int pulse, int loadMask, int startFrame, int maxFrames, int singleChan, boolean offline){
		try{
		knownIssues(pulse); //display warnings if we're attempting to read pulses with specific known problems
		
		this.pulse = pulse;
		this.jmds = jmds;
		this.offline = offline;
		this.baseStr = offline ? "jet-offline/" : "jet/";
		
		//read the PPF profiles if requested
		if((loadMask & LD_PROFILES) != 0)readProcessedProfiles();

		boolean bl,sig;
		sig = (loadMask & LD_SIGNAL) != 0;
		bl = (loadMask & LD_BASELINE) != 0;
		
		//everything else comes from the signal reader as the calibration stuff is in the jpf/dd/e9-spec:001 header
		// or needs the calibration, so basically call this anyway
		if(sig || bl || (loadMask & (LD_SPECTRA | LD_VIGNET | LD_GAINS | LD_TIMING | LD_ENERGY_MON)) != 0) 
			readSignal(sig,bl,startFrame,maxFrames,singleChan);
				
		//System.gc(); //generally a good idea at this point
		}catch(RuntimeException err){
			System.err.println("RuntimeException in KE9DDataProvider while processing pulse "+pulse);
			throw new RuntimeException(err);
		}
	}
	
	/** Reads all the appropriate JPFs, fills in and processes the calibration files and reads the baseline and signal.	 */
	private void readSignal(boolean doSignal, boolean doBaseline, int startFrame, int maxFrames, int singleChan){		
		byte jpfData[] = jmds.getRawJPF(baseStr + pulse + "/jpf/dd/e9-spec:001#raw");
				
		byte data[] = new byte[jpfData.length];
		System.arraycopy(jpfData, 0, data, 0, jpfData.length);
		
		jmds.freeFromMemory(baseStr + pulse + "/jpf/dd/e9-spec:001#raw");
		
		//swap every pair of bytes
		for(int i=0;i<Math.floor(data.length / 2.0);i++){
			byte tmp = data[2*i];
			data[2*i] = data[2*i+1];
			data[2*i+1]=tmp;
		}
		
		int ptr = findData(data);						
		if(ptr < 0)
			throw new RuntimeException("Didn't find 'EndOfHeader' in /jpf/dd/e9-spec:001#raw");	
		
		String headerText = new String(data,0,ptr);
		
		CalibData cal = parseHeader(headerText);
		parseAlternateCalib(cal);
		readHV(cal);
		processCalib(cal);
		TriggersAndTiming(cal.triggerTimes);
		LaserEnergy();
		
		//if they wanted less frames than the complete, adjust the frame times accordingly
		if((maxFrames > 0 & maxFrames < frameTimes.length) || startFrame > 0){
			double newTimes[] = new double[maxFrames];
			double newLasEng[] = new double[maxFrames];
			System.arraycopy(frameTimes, startFrame, newTimes, 0, maxFrames);
			System.arraycopy(laserEnergy, startFrame, newLasEng, 0, maxFrames);
			frameTimes = newTimes;
			laserEnergy = newLasEng;
		}
		
		int expectedLen = (nFrames+1)*nChans*frameSize*2;
		int actualLen = (data.length - ptr);
		if( actualLen > expectedLen )
			warning(3,"Expecting "+expectedLen+" bytes of data for "+nFrames+" frames x "+nChans+" channels x "+frameSize+" integers but got more: " + actualLen + ", ignoring extra.");
		else if(actualLen < expectedLen ){
			nFrames = (int)(actualLen / (nChans*frameSize*2)) - 1;
			warning(3,"Expecting "+expectedLen+" bytes of data for "+nFrames+" frames x "+nChans+" channels x "+frameSize+" integers but got less: " + actualLen + ".\n" +
					"We have enough data for " + nFrames + " frames.");
		}

		if(startFrame < 0 || startFrame >= nFrames)
			throw new IllegalArgumentException("startFrame is below 0 or out of range (0 < "+startFrame+" < " + nFrames + ").");		

		if(doBaseline){
			baseline = new double[nChans][frameSize];
			for(int ch=0;ch<nChans;ch++)
				for(int i=0;i<frameSize;i++){
					int j = ptr + (((ch)*frameSize)+i)*2;
					baseline[ch][i] = 32768 - data[j]*256 + data[j+1];
				}
		}else baseline = null;
		
		if(doSignal){
			if(maxFrames > 0 && maxFrames < nFrames)
				nFrames = maxFrames;
			rawSig = new double[nFrames][nChans][frameSize];
			for(int f=0;f<nFrames;f++)
				for(int ch=0;ch<nChans;ch++)
					for(int i=0;i<frameSize;i++){
						int j = ptr + ((((f+1+startFrame)*nChans + ch)*frameSize)+i)*2;
						rawSig[f][ch][i] = 32768 - data[j]*256 + data[j+1];
					}
		}else rawSig = null;
	}
	
	/** Search the data for 'EndOfHeader' and return the index */
	private int findData(byte data[]){
		//data is now split into a text and a binary part
		//binary parts starts after "EndOfHeader"
		int ptr;
		String cmp = "EndOfHeader";
		
		for(ptr=0;ptr<(data.length - cmp.length());ptr++){			
			int j;
			for(j=0;j<cmp.length();j++)
				if( cmp.charAt(j) != data[ptr+j] )
					break;
			
			if(j >= cmp.length()){				
				ptr += cmp.length() + 2; //+2 for CR+LF
				return ptr;
			}
		}
		return -1;		
	}
	
	private CalibData parseHeader(String txt){		
		String lines[] = txt.split("\n");
		
		CalibData cal = new CalibData();
		
		
		for(int l=0;l<lines.length;l++){			
			String line = cleanLine(lines[l]);
			if(line.length() <= 0)continue; //ignore blank lines
			
			String parts[] = line.split("\\s"); //split by spaces (that might've been :,=,\t or space)
						
			if(parts[0].equalsIgnoreCase("Analysis")){ }//err... yea?				
			else if(parts[0].equalsIgnoreCase("R1")){ R0 = Double.parseDouble(parts[1]); }//LOS definition 
			else if(parts[0].equalsIgnoreCase("R2")){ R1 = Double.parseDouble(parts[1]); }//	not sure what todo with these yet
			else if(parts[0].equalsIgnoreCase("Z1")){ Z0 = Double.parseDouble(parts[1]); }
			else if(parts[0].equalsIgnoreCase("Z2")){ Z1 = Double.parseDouble(parts[1]); }
			//else if(parts[0].equalsIgnoreCase("dOmega")){ dOmega = Double.parseDouble(parts[1]); }
			//else if(parts[0].equalsIgnoreCase("TShift")){ Z1 = Double.parseDouble(parts[1]); }
			
			else if(parts[0].matches("Ch[0-9]Eta")){ //relative sensitivities
				int ch = getCh(parts[0],2);			
				if(relSens==null)relSens = new double[nChans];
				relSens[ch] = Double.parseDouble(parts[1]);
			}
			else if(parts[0].matches("Ch[0-9]PE")){
				int ch = getCh(parts[0],2);				
				if(cal.peVal==null)cal.peVal = new double[nChans];
				cal.peVal[ch] = Double.parseDouble(parts[1]); 
				
			}
			else if(parts[0].matches("HVBiasCh[0-9]")){
				int ch = getCh(parts[0],8);				
				if(cal.hvBias==null)cal.hvBias = new int[nChans];
				cal.hvBias[ch] = Integer.parseInt(parts[1]);
			}
			else if(parts[0].matches("PEHV")){				
				cal.pehv = Integer.parseInt(parts[1]);
			}
			else if(parts[0].matches("Ch[0-9]F")){
				int ch = getCh(parts[0], 2);				
				if(cal.hvCubicCoeffs==null)cal.hvCubicCoeffs = new double[nChans][3];
				cal.hvCubicCoeffs[ch][0] = Double.parseDouble(parts[1]);
				cal.hvCubicCoeffs[ch][1] = Double.parseDouble(parts[2]);
				cal.hvCubicCoeffs[ch][2] = Double.parseDouble(parts[3]);
			}
			else if(parts[0].matches("channel[0-9]") && parts[1].equalsIgnoreCase("range")){
				int ch = getCh(parts[0],7);				
				if(cal.vRanges==null)cal.vRanges = new double[nChans];
				cal.vRanges[ch] = Double.parseDouble(parts[2]);
			} 
			else if(parts[0].equalsIgnoreCase("timebase") && parts[1].equalsIgnoreCase("range"))
				cal.tRange = Double.parseDouble(parts[2]);
			else if(parts[0].equalsIgnoreCase("acquire") && parts[1].equalsIgnoreCase("points"))
				cal.nPoints = Integer.parseInt(parts[2]);
			
			else if(parts[0].equalsIgnoreCase("dOmega")){ } //TODO: Do something with this
			else if(parts[0].equalsIgnoreCase("TShift"))
				cal.tShift = Double.parseDouble(parts[1]);
			else if(parts[0].equalsIgnoreCase("INVALID"))
				warning(4, "Calibration is apparently invalid");					
			else if(parts[0].matches("NCh[0-9]SlitFn")){ //slit functions
				int ch = getCh(parts[0],3);				
				int nEntries = Integer.parseInt(parts[1]);
				if(nEntries <=0 )
					throw new IllegalArgumentException("Invalid number of entries '"+parts[1]+"' for '"+parts[0]+"'");
				if(l+nEntries > lines.length)
					throw new IllegalArgumentException("Not enough lines left for "+nEntries+" of '"+parts[0]+"'");
				
				if(normSpect == null)
					normSpect = new double[nChans][][];
				normSpect[ch] = new double[2][nEntries];
				
				for(int i=0;i<nEntries;i++){					
					String line2 = cleanLine(lines[l+1+i]);
					String parts2[] = line2.split("\\s+"); //split by multiple spaces
					normSpect[ch][0][i] = Double.parseDouble(parts2[0]) * 10; //wavelength comes in nm, we want it in angstrom, just to be difficult
					normSpect[ch][1][i] = Double.parseDouble(parts2[1]);
				}
				l += nEntries; //move on to line of last entry so main for loop takes us to next cmd
			}
			else if(parts[0].equalsIgnoreCase("Xinc")) cal.dt = Double.parseDouble(parts[1]);
			else if(parts[0].matches("Yinc[0-9]")){
				int ch = getCh(parts[0], 4);				
				if(cal.VPerADCU==null)cal.VPerADCU = new double[nChans];
				cal.VPerADCU[ch] = Double.parseDouble(parts[1]);
			}
			else if(parts[0].equalsIgnoreCase("RVig1"))	cal.RVig0 = Double.parseDouble(parts[1]);
			else if(parts[0].equalsIgnoreCase("dRVig")) cal.dRVig = Double.parseDouble(parts[1]);
			else if(parts[0].equalsIgnoreCase("NVig")){
				int nEntries = Integer.parseInt(parts[1]);
				if(nEntries <=0 )
					throw new IllegalArgumentException("Invalid number of entries '"+parts[1]+"' for '"+parts[0]+"'");
				if(l+nEntries > lines.length)
					throw new IllegalArgumentException("Not enough lines left for "+nEntries+" of '"+parts[0]+"'");
				
				vignetting = new double[2][nEntries];
				for(int i=0;i<nEntries;i++)				
					vignetting[1][i] = Double.parseDouble(cleanLine(lines[l+1+i]));
				
				l += nEntries; //move on to line of last entry so main for loop takes us to next cmd				
			}
			else if(parts[0].equalsIgnoreCase("LaserTriggers")){
				int n = Integer.parseInt(parts[1]);
				if(n <= 0){
					warning(3,"No laser triggers in header (NTriggers <=0 )!");
					cal.triggerTimes = null;
				}else cal.triggerTimes = new double[n];
			}else if(parts[0].equalsIgnoreCase("t_Trigger")){
				/* Old tigger times were like:
				 * t_Trigger 1 = 0.998168;
				 *
				 * New trigger times are like:
				 * t_Trigger = 5.23758e+007 5.33757e+007 5.43757e+007 5.53757e+007 5.63757e+007 5.73757e+007 ;
				 * 
				 * So try to parse first arg as an integer channel number, if that fails, assume it's post-upgrade
				 */
				try{
					int i = Integer.parseInt(parts[1]) - 1;				
					if(cal.triggerTimes == null || i < 0 || i >= cal.triggerTimes.length)
						warning(3,"No 'LaserTriggers' command found in header or 't_Trigger' command issued with index ("+(i+1)+") which is outside 0 <= i < LaserTriggers");
					else
						cal.triggerTimes[i] = Double.parseDouble(parts[2]);
				}catch(NumberFormatException err){
					if( (parts.length - 1) != cal.triggerTimes.length)
						warning(2, "Post 2007/8 upgrade style t_Trigger command has different number of triggers to 'LaserTriggers' command.");
					
					for(int ch=0; ch < Math.min(cal.triggerTimes.length, (parts.length - 1)); ch++)
						cal.triggerTimes[ch] = Double.parseDouble(parts[ch+1]);				
				}
			}
		}		
		return cal;		
	}
			
	private String cleanLine(String line){		
		line = line.split("//")[0]; //get rid of comment (take anything before it)
		line = line.replaceAll("\r", ""); //remove CRs 
		line = line.replaceAll("\t", " "); //replace tabs with spaces
		line = line.replaceAll(",", " "); //replace , with spaces
		line = line.replaceAll(":", " "); //replace : with spaces			
		line = line.replaceAll("=", " "); //replace = with spaces			
		line = line.replaceAll("^\\s+", ""); //remove all leading spaces
		line = line.replaceAll("\\s+$", ""); //remove all trailing spaces
		line = line.replaceAll(";$", ""); //strip trailing semicolons			
		line = line.replaceAll("\\s+", " "); //remove duplicate spaces
		return line;
	}
	
	/** Get channel number from inside command and check against channel limits etc */
	private final int getCh(String cmd, int idx){
		int ch = cmd.charAt(idx) - '1';
		if(ch < 0 || ch >= nChans)
			throw new IllegalArgumentException("Invalid channel number in '"+cmd+"'");
		return ch;
	}
	
	/** Read configuration amendmants from input file 'chain1/input/ke9d/e9calib' as in 'chain1/source/ke9d/fortran/rdcalib.f' */
	private void parseAlternateCalib(CalibData cal){
		String fileName = jmds.getCacheRoot() + "/common/ke9/e9calib"; //the alternate calib is kept in the data root (MDS+ cache)
		try{
			
			FileInputStream fIn = new FileInputStream(fileName);
			BufferedReader reader = new BufferedReader( new InputStreamReader(fIn));
			
			int pulse0 = 1, pulse1 = -1; //start non-matchable pulse range
			boolean inRange = false;
			do{
				String line = reader.readLine();
				if(line == null)
					break;
				
				String parts[] = line.split("[\\s,]+");
				
				if(parts[0].equalsIgnoreCase("PULSENO:")){ //if this is a pulse range line
					pulse0 = Integer.parseInt(parts[1]);
					pulse1 = Integer.parseInt(parts[2]);
					inRange = (pulse >= pulse0 && pulse <= pulse1);
				}
				else if(inRange){ //if another command, only process if we were in range of the last PULSENO command
					if(parts[0].equalsIgnoreCase("INVALID:"))
						warning(4,"Alternative calibration file says this pulse is invalid");
					else if(parts[0].equalsIgnoreCase("DATE:"))
						System.out.println("Processing alternate calibration of "+parts[1]+" ("+pulse0+"-"+pulse1+")");
					else if(parts[0].equalsIgnoreCase("ETA:"))
						for(int ch=0;ch<nChans;ch++)
							relSens[ch] = Double.parseDouble(parts[ch+1]);
					else if(parts[0].equalsIgnoreCase("PE:"))
						for(int ch=0;ch<nChans;ch++)
							cal.peVal[ch] = Double.parseDouble(parts[ch+1]);
					else if(parts[0].equalsIgnoreCase("HVBIAS:"))
						for(int ch=0;ch<nChans;ch++)
							cal.hvBias[ch] = Integer.parseInt(parts[ch+1]);
					else if(parts[0].equalsIgnoreCase("TSHIFT:"))
						for(int ch=0;ch<nChans;ch++)
							cal.tShift = Double.parseDouble(parts[ch+1]);
					else if(parts[0].equalsIgnoreCase("DOM:")){
						//TODO: Do something with this
					}else if(parts[0].equalsIgnoreCase("VIG0:")){
						cal.RVig0 = Double.parseDouble(parts[1]);
					}else if(parts[0].equalsIgnoreCase("VIGNET01:"))
						warning(3,"Alternate instruction 'VIGNET01' not yet handled.");
					else if(parts[0].equalsIgnoreCase("VIGNET11:"))
						warning(3,"Alternate instruction 'VIGNET11' not yet handled.");
					else if(parts[0].equalsIgnoreCase("VIGNET21:"))
						warning(3,"Alternate instruction 'VIGNET21' not yet handled.");					
					else if(parts[0].equalsIgnoreCase("RIW:")){ //TODO: Do something with this
					}else if(parts[0].equalsIgnoreCase("ZIW:")){ //TODO: Do something with this
					}else if(parts[0].equalsIgnoreCase("ABSCALIB:")){  
						//TODO: Do something with this - it's their version of smfs.LogProb.uberConst, calibrated by KG1V 
					}					
				}
				
			}while(true); //broken by null return from readLine()
			
		}
		catch(IOException err){
			warning(2,"Error processing alternate KE9 calibration file '"+fileName+"':\n"+err.getMessage());
		}
	}
	
	/** After reading everything in from the header (or otherwise), this does our conversions to the things we need for the proper interface */ 
	private final void processCalib(CalibData cal){
		if(cal.tRange != 5e-8)
			warning(3,"Expecting timebase range of 5x10^-8 but have been told its "+cal.tRange+", probably by ':timebase:range' line in header");
		
		
		/* In theory cal.tRange / cal.nPoints should give us the signal Dt, but
		 * it seems to be 1/2 the correct value for post-upgrade. 
		 * It also gets it slightly off as it does eg 50ns / 250 for pre-upgrade
		 * which is not precisely 0.2ns
		 * 
		 * I suspect the scope actually takes 256 x 0.2ns samples, ending up with an actual range of 5.12e-8 s
		 * TODO: ask people at some point
		 * So we just do specific cases here
		 */
		
		frameSize = cal.nPoints; //This one seems to be actually correct.
		
		if(pulse >= UPGRADE_FIRST_PULSE){
			if(cal.nPoints != 2048)
				warning(3,"Expecting framesize of 2048 (post-upgrade) but have been told its "+cal.nPoints +", probably by ':aquire:points' line in header");
			//dt = 50e-12; // 50ps	
		}else{
			if(cal.nPoints != 256 && cal.nPoints != 2048)
				warning(3,"Expecting framesize of 256 (pre-upgrade) but have been told its "+cal.nPoints +", probably by ':aquire:points' line in header");
			//dt = 200e-12; // 200ps
		}
		
		/*
		//we only have a single global voltage range, so check they're all the same and are the expected 400mV 
		double vRange = 0.4;
		if(cal.vRanges==null)
			warning(5,"No channel voltage ranges in header, using 0.4V");
		else{
			if(cal.vRanges[0] != 0.4){
				warning(3,"Expecting voltage range of 0.4V but ch0's is "+cal.vRanges[0]+"V. This will be used instead");
				vRange = cal.vRanges[0];
			}
				
			for(int ch=1;ch<nChans;ch++)
				if(vRange != cal.vRanges[ch])
					warning(3,"Voltage range for channel "+ch+" was "+cal.vRanges[ch]+"V which isn't the same as ch0's ("+cal.vRanges[0]+"V). Ch0's will be used");
		}
		mVPerADCU = (vRange*1000.0) / ADCUMaxRange; //calc mv Per ADCU		
			*/
		
		dt = cal.dt;
		if(pulse >= UPGRADE_FIRST_PULSE && dt != 50e-12 || pulse < UPGRADE_FIRST_PULSE && dt != 250e-12)
			warning(3, "Timebase is "+dt+", which isn't expected (50ps for post-upgrade, 250ps for pre-upgrade");
		
		for(int ch=1; ch<nChans; ch++)
			if(cal.VPerADCU[ch] != cal.VPerADCU[0])
				warning(4, "Channel voltages ranges are not the same.");
		
		mVPerADCU = cal.VPerADCU[0] * 1000;
		
		//create the RVector for the vignetting
		if(Double.isNaN(cal.RVig0))throw new IllegalArgumentException("Never found RVig1 in header");
		if(Double.isNaN(cal.dRVig))throw new IllegalArgumentException("Never found dRVig in header");
		for(int i=0;i<vignetting[0].length;i++)
			vignetting[0][i] = cal.RVig0 + i* cal.dRVig;
		
		//AsciiMatrixFile.mustWrite("/work/scrap/ke9Vignet.txt", vignetting, true);
		
		if(relSens==null)
			warning(5,"No relative sensitivies in header.");
		else
			for(int ch=0;ch<nChans;ch++){
				//AsciiMatrixFile.mustWrite("/work/scrap/ke9DetSpect"+ch+".txt", normSpect[ch], true);
			
				//create spectrums from normalised ones
				if(normSpect[ch]==null)
					warning(5,"No normalised detector spectrum for channel "+ch+"in header");
				else if(relSens[ch]==0)
					warning(5,"Relative sensitivity for channel " + ch+ " is 0");
				else{
					//if backwards, we will need to reverse the spectrum
					int n = normSpect[ch][0].length;
					if(normSpect[ch][0][0] > normSpect[ch][0][n-1]){
						double newWave[] = new double[n];
						double newAmp[] = new double[n];
						for(int i=0;i<n;i++){
							newWave[i] = normSpect[ch][0][n-i-1];
							newAmp[i] = normSpect[ch][1][n-i-1];
						}
						normSpect[ch][0] = newWave;
						normSpect[ch][1] = newAmp;
						
					}	
						
					detSpect = new double[nChans][2][];
					detSpect[ch][0] = normSpect[ch][0];
					detSpect[ch][1] = new double[n];
					for(int i=0;i<n;i++)
						detSpect[ch][1][i] = normSpect[ch][1][i] * relSens[ch]; //* winTransFactor[ch];
				}
			}
		
		// calculate gain (mV per photoelectron per ns)
		// If we have Ch#F we assume its the post-upgrade type
		// if not we should have the HVBiasCh# lines
		// we need Ch#PE in both cases
		if(cal.peVal == null)
			warning(5,"No PE Values in header, there will be no gains.");
		else {
			if(cal.hvCubicCoeffs != null) {
				gain = new double[nChans];
				for(int ch=0;ch<nChans;ch++){				
					//calculation as in chain1/source/ke9d/source/rdjpf.f.									
					double x = Math.pow(10, cal.hvCubicCoeffs[ch][0] * cal.hvSetting[ch] * cal.hvSetting[ch]
   					                      + cal.hvCubicCoeffs[ch][1] * cal.hvSetting[ch]
   					                      + cal.hvCubicCoeffs[ch][2]);
					
					double y = Math.pow(10, cal.hvCubicCoeffs[ch][0] * cal.pehv * cal.pehv
										  + cal.hvCubicCoeffs[ch][1] * cal.pehv
										  + cal.hvCubicCoeffs[ch][2]);		                                                                 
                                                               					
					gain[ch] = x / y;
					System.out.println("gain/peval "+ch+" = " + gain[ch]);
					gain[ch] *= cal.peVal[ch];
					
				}
			} else if(cal.hvBias != null) {
				gain = new double[nChans];
				for(int ch=0;ch<nChans;ch++){				
					//calculation as in chain1/source/ke9d/fortran/rdjpf.f. Not sure where is comes from/what it does
					cal.hvSetting[ch] = Math.abs(cal.hvSetting[ch]); //check there are +ve
					cal.hvBias[ch] = Math.abs(cal.hvBias[ch]);
					gain[ch] = (cal.hvSetting[ch] - cal.hvBias[ch]);
					gain[ch] = Math.pow(gain[ch]/(double)(3300 - cal.hvBias[ch]),6);
					gain[ch] *= cal.peVal[ch];
				}
			} else {
				warning(5,"No HV biases in header (no 'HVBiasCh#' or 'Ch#F' lines). There will be no gains.");
			}
		}
		
		
		//timings are the same for all channels as only 1 scope is used
		timing = new double[nChans];
		for(int ch=0;ch<nChans;ch++)
			timing[ch] = cal.tShift * 1e-9; //calib data appear to give them in ns, convert to secs
		
	}
	
	/** read the HV settings from their seperate JPFs */
	private void readHV(CalibData cal){
		cal.hvSetting = new int[nChans];
		for(int ch=0;ch<nChans;ch++){
			double hv[] =  jmds.get1DData(baseStr + pulse + "/jpf/dc/ke9-hv"+(ch+1));
			cal.hvSetting[ch] = (int)hv[0]; //only first entry means anything			
		}
	}
	
	private void TriggersAndTiming(double triggerTimes[]){
		
		if(triggerTimes != null && triggerTimes.length > 0 && triggerTimes[0] >= 1e6){
			//trigger times look like sensible times (in us) so just use them as they are
			
			if(triggerTimes.length >= nFrames){
				frameTimes = new double[triggerTimes.length];
				for(int i=0; i < triggerTimes.length; i++)
					frameTimes[i] = triggerTimes[i] / 1e6;
				nFrames = frameTimes.length;
				
			}else{			
				warning(1, "Trigger times are avaliable but there are only " + triggerTimes.length + " of them, instead of " + nFrames + ".\n" +
									"The rest will be filled in by adding 1s to the first, just incase there is done data there.");
				frameTimes = new double[nFrames];
				
				for(int i=0; i < triggerTimes.length; i++)
					frameTimes[i] = triggerTimes[i] / 1e6;
				for(int i=triggerTimes.length; i < frameTimes.length; i++)
					frameTimes[i] = frameTimes[i-1] + 1.000;
			}
			
			return;
		}
		
		//no probably pre 2007/8 upgrade or no trigger times, we cheat and use the PPF ones (as the process is complicated and I CBA)
		if(ppfTimes==null)
			readProcessedProfiles();
		
		warning(1, "Trigger times are not avaliable or are < 1s, will attempt to use PPFs for accurate frame times...");
		
		if(ppfTimes==null){					
			if(triggerTimes == null){
				warning(4,"No trigger times and no PPFs either. Setting frames times to 1,2,3,4,5,6 and assuming 6 frames (Last ditch attempt)");
				frameTimes = new double[]{ 1,2,3,4,5,6 };
			}else{
				warning(4,"Trigger times from header are present but < 1s, which means this is probably a pre 2007/8 upgrade pulse which can't be processed here. " +
						" Normally the PPF frameTimes would be used but they are unavaliable. This will now use the trigger times added to the main " +
						" start time for an approx estimate. THE FRAME TIMES ARE ALMOST CERTAINLY WRONG.");
				
			
				//read start time for stopwatch and add to the trigger times 
				// THIS DOES NOT WORK PROPERLY
				
				double trigStart[] = jmds.getTimeVector(baseStr + pulse + "/jpf/dc/e3-la-tim>");				
				for(int i=0;i<triggerTimes.length;i++){
					triggerTimes[i] += trigStart[0];
					//System.out.println("Trigger time "+i+": "+triggerTimes[i]);
				}
				frameTimes = triggerTimes;
			}
		}else if(ppfTimes.length < nFrames){
				warning(4,"PPF times arn't long enough for every frame, Setting frames times to 1,2,3,4,5,6 and assuming 6 frames (Last ditch attempt)");
				frameTimes = new double[]{ 1,2,3,4,5,6 };				
		}else{
			warning(1,"Using PPFs for frame times which means this is probably a pre 2007/8 upgrade pulse." +
					"They do have the same number as we have frames but this is now trusting AJC's Chain1 code which he comments himself is probably wrong.");
		
			frameTimes = ppfTimes;
		}		
	}
	
	/** Processes the laser energy monitor to find laser firing times and laser energy.
	 * These are then lined up with the trigger times.
	 * This is all done almost exactly as chain1/source/ke9d/fortran/la{pulse,stats,trig}.f because I can't be bothered
	 * to go through and work out how it works. 
	 */
	private void LaserEnergy(){
		
		// in the first instance, use the PPF laser energy because the algorithm they have to do this is a bit too complicated for me to bother here
		try{
			laserEnergy = (double[])(((PPF)jmds.getSig(new PPFDesc(pulse, "ke9d", "elas"))).getDataAsType(double.class));			
			if(laserEnergy.length == nFrames)
				return;
			
			warning(2, "Laser energy from PPF doesn't have same length as frame times so will use a crude measurement of JPF monitor instead. Laser energy (Absolute density) might be wrong");
			
		}catch(RuntimeException err){
			warning(2, "Error reading PPF laser energy so will use crude measurement of JPF instead. Laser energy (Absolute density) might be wrong: " + err.getMessage());			
		}
		
		/* Ok, try it from JPF. This method seems to get ~2% higher than chain1 for nice pulses.
		 * Actually I think this is more accurate, but it probably will fail easily if things are
		 * unusual, so we'll stick with chain1's for consistency and soak up any problem in the uberConst.
		 */
		
		JPF laserMon = (JPF)jmds.getSig(new JPFDesc(pulse, "dc", "e3-ulaeng<mon:006", offline, false));
		double laEng[] = (double[])laserMon.getDataAsType(double.class);
		double laTime[] = (double[])laserMon.getTVectorAsType(double.class);
		
		laserEnergy = new double[nFrames];

		//find baseline level from first and last 100 points (~1s each)
		if(frameTimes[0] < laTime[100] || frameTimes[nFrames-1] > laTime[laTime.length-100])
			warning(2, "First or last frame times overlap with averaging window for laser energy monitor. Laser energy (Absolute density) might be wrong.");
		double baseline = 0;
		for(int i=0;i<100;i++){
			baseline += laEng[i];
			baseline += laEng[laTime.length-100+i];
		}
		baseline /= 200;
				
		//now get max between -50ms and +300ms of the trigger times
		for(int i=0;i<nFrames;i++){
			int idx0 = OneLiners.getNearestLowerIndex(laTime, frameTimes[i] - 0.050);
			int idx1 = OneLiners.getNearestLowerIndex(laTime, frameTimes[i] + 0.300);
			
			//find max
			double max = Double.NEGATIVE_INFINITY;			
			for(int idx=idx0; idx <= idx1; idx++){
				if(laEng[idx] > max)
					max = laEng[idx] - baseline;
			}
			
			//go back over and find average of all points within 90% of max
			int nPointsInAvg = 0;
			for(int idx=idx0; idx <= idx1; idx++){
				double l = laEng[idx] - baseline;
				if(l > 0.8*max){
					nPointsInAvg++;
					laserEnergy[i] += l;
				}
			}
			laserEnergy[i] /= nPointsInAvg;
		}
		
	}
	
	/** loads Te and ne profiles from standard analysis PPF */
	private void readProcessedProfiles(){
		if(offline){
			warning(1, "This is an offline pulse, cannot read PPF profiles.");
			ppfne = null; ppfTe = null; ppfZ = null; ppfR = null; ppfTimes = null;		
		 	return;
		}
		try{
			ppfne = jmds.get2DData("jet/" + pulse + "/ppf/ke9d/ne");
			ppfTe = jmds.get2DData("jet/" + pulse + "/ppf/ke9d/te");
			double Ztmp[][]  = jmds.get2DData("jet/" + pulse + "/ppf/ke9d/zvec");
			ppfZ = Ztmp[0];
			ppfR = jmds.getXVector("jet/" + pulse + "/ppf/ke9d/ne");
			ppfTimes = jmds.getTimeVector("jet/" + pulse + "/ppf/ke9d/ne");	
		}catch(RuntimeException err){ //any errors in here are hopefully not fatal. They will however mean there isn't any laser energy monitor
			warning(1,"Aborted reading PPF profiles. Continuing, but they will not be avaliable. Reason was: " + err.getMessage());
			ppfne = null; ppfTe = null; ppfZ = null; ppfR = null; ppfTimes = null;			
		}
	}
	
	/** Getters specific to {@link KE9DDataProvider} (i.e not part of the {@link LidarDataProvider} interface) */
	public final double getR0() { return R0; }
	public final double getR1() { return R1; }
	public final double getZ0() { return Z0; }
	public final double getZ1() { return Z1; }
	
	//Getters... the nice sensible clean and simple interface provided to anyone still sane
//	TODO got to check these as they were copied from KE3 provider
	
	/** @return the pulse number the provider was initilaised with */
	public final long getPulse() { return pulse;	}
	
	/** @param frame Frame number (i.e laser pulse index)
	 * @return 2D array: array of data points of the specified frame: signal[channel number][]
	 */
	public final double[][] signalFrame(int frame){	return rawSig[frame];	}
		
	/** @return Laser energy for all frames */
	public final double[] laserEnergies(){ return laserEnergy; }
	
	/** @return Laser energy for a specified frame */
	public final double laserEnergy(int frame){ return laserEnergy[frame]; }
	
	/** @return signal for all channels in all frames: signal[frame][channel][] */
	public final double[][][] signalFull(){	return rawSig;	}
	
	/** @return baseline for all channels in all frames */ 
	public final double[][] baseLine(){ return baseline; }
	
	/** @return number of channels active. */
	public final int nChannels(){ return nChans; }
	
	/** @return number of frames (laser pulses) */
	public final int nFrames(){ return nFrames; }
	
	/** @return Frame (laser pulses) time points */
	public final double[] frameTimes(){ return frameTimes; }
	
	/** @return vignetting[R / data][data point] */
	public final double[][] vignetting(){ return vignetting; }
	
	/** @return the detector spectrums [ch][wave/amp][] - normalsied multiplied by relative sensitivity*/
	public final double[][][] spectrums(){  return detSpect;  }
	
	/** @return normalised detector spectrums [ch][wave/amp][]*/
	public final double[][][] normSpectrums(){  return normSpect;  }
	
	/** @return the relative sensitivities [ch]*/
	public final double[] relativeSensitivities(){  return relSens;  }
	
	/** @return signal conversion  in mv per (photoelectron per nanosecond). */
	public final double[] gains(){ return gain; }
	
	/** @return time between timing pulse and backwall in seconds */
	public  final double[] timing(){ return timing; }
	
	/** @return The ADC unit size in mV */
	public final double ADCUnit(){ return mVPerADCU; };
	
	/** @return The maximum value the signal can take in ADC Units (The saturation level) */
	public final double signalMax(){ return ADCUMaxRange; }
	
	/** @return length of signal data array */
	public final int signalLength(){ return frameSize; }
	
	/** @return data's sampling period in seconds  */
	public final double signalDt(){ return dt; }
		
	public final double[] ppfFramene(int ppfFrame){ return ppfne[ppfFrame]; }
	public final double[] ppfFrameTe(int ppfFrame){ return ppfTe[ppfFrame]; }
	public final double[] ppfR(){ return ppfR; }
	public final double[] ppfZ(){ return ppfZ; }
	/** @return the frame index into the PPF (their frame number) that sits nearest in time to a frame in our arrays, these do not necceasrily line up exactly. */ 
	public final int pffNearestFrame(int frame){
		double t = frameTimes[frame];		
		if(ppfTimes == null)return -1;
		return OneLiners.getNearestIndex(ppfTimes, t);
	}
	
	/** @return The level of the most severe warning issued:
	 * 0=No problems
	 * 1=Some data is unavaliable (PPF or Laser Energies) 
	 * 2=Some unusual values were found, data may be suspect.
	 * 3=Some values were not sensible , but these may or may not have an impact on output, 
	 * 4=Some very unusual values were found/things were not as expected, data is very likely to not be sensible.
	 * 5=Severe irregularlites were seen, outputs are probably complete rubbish. (the provider will most likely crash before finishing if a level 5 is issued)
	 */
	public final int warningLevel(){ return warningLevel; }
	
	private final void warning(int level, String text){
		System.err.println("lidrDataProvider WARNING (pulse "+pulse+"): "+text+" (Level "+level+")");
		if(level > 3 && warningLevel <= 3) //first warning over level 3
			System.err.println("For warning levels > 3, if the data provider or anything using the data crashes, this is most likely the reason why.");
		if(warningLevel < level)
			warningLevel = level;
	}	
	
	/** All manner of special cases are given for certain pulses in the Chain1 code. I'm only going to deal with those which
	 * cover large pulse ranges. Both those dealt with and those I can't be bothered with are listed here */ 
	private void knownIssues(int pulseNo){
		if(pulseNo >= 63700 && pulseNo <= 66000)
			System.err.println("WARNING: 52250 has a known issue! See chain1's ke9d/fortran/keywrds.f");
		
	}
	
}
