/** This file is part of LIDAR Analysis work carried as part of a PhD with Imperial College.  
 * Copyright (c) 2010 Oliver Ford
 *
 * The content, concept and implementation are the intellectual property of Oliver Ford but permission
 * is given for it's use and further development by Jakob Svensson or any person or organisation
 * directly given such permissions by Oliver Ford or by Jakob Svensson.
 */
package seed.minerva.lidar.dataProvision;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;

import algorithmrepository.LinearInterpolation1D;

import oneLiners.AsciiMatrixFile;
import oneLiners.OneLiners;

import otherSupport.SimpleScriptParser;

import seed.minerva.*;
import seed.minerva.nodetypes.*;

/**
 * Provides all the configuration defaults for the JET LIDAR TS systems which do not come from the DataSource (PPFs etc) 
 * Currently uses files under lidar config dir defined in minerva-settings
 * 
 * It will load first the file named
 * 		config-[diag].txt
 * then it will load any file
 * 		config-[diag]-p[pulse0]_[pulse1].txt
 * 	which range includes this pulse (inclusive)
 * then it will look for and load
 * 		config-[diag]-p[pulse].txt
 * and finally
 * 		config-[diag]-p[pulse]-f[frame].txt 
 * 
 * Files include any of the following directives:
 * 		Warning 		text		this pulse/frame/range has problems
 * 		Abort			text		this pulse/frame/range doesn't work
 * 
 * 
 * Things for each:
 * 		init
 * 		mean (for gauss)
 * 		sigma (for gauss)
 * 		min (for trunc. gauss or uniform)
 * 		max (for trunc. gauss or uniform)
 * 		typMin (typ. range)
 * 		typMax (typ. range)
 * 
 * Outbound Connection Points:
 * 		[baseline/tpulse/switchOn]ShapeData[X/Y] (value only)
 * 		[baseline/tpulse/switchOn][X0/Y0/A][Min/Max/Value/TypMin/TypMax]
 * 		ambientBegin[Min/Max/Value/TypMin/TypMax]
 * 		ambientLevel[Min/Max/Value/TypMin/TypMax]
 * 		backwall[X0/A/ScaleLength][Min/Max/Value/TypMin/TypMax]
 * 
 * 		uberConst[Mean/Sigma/Value/Min/Max/TypMin/TypMax] (trunc.gauss)
 * 		elecSigma	(basic on-the-fly calc on data or override)
 * 		windowTransmission (value only)
 * 		vignettingProfile[X/Y] (value only) (pipe/override)
 * 		relativeSensitivity[Mean/Sigma/Value/Min/Max/TypMin/TypMax] (pipe/override)
 * 		timingAdjust[Mean/Sigma/Value/Min/Max/TypMin/TypMax] (0 or override)
 * 		tpulseToBackwallTime[Mean/Sigma/Value/Min/Max/TypMin/TypMax] (pipe/override)
 * 		photoElectronFactors (pipe/override, value only)
 * 		lineOfSightCoords (value only) (pipe/override for KE9, specified for KE3, all slicing etc internal)
 * 		plasmaRegion (value only)
 * 		channelEnable (value only) 
 * 		
 *  
 * Copyright (c) 2010 Oliver Ford
 * This file is part of LIDAR Analysis work carried as part of a PhD with Imperial College.
 * The content, concept and implementation are the intellectual property of Oliver Ford but permission
 * is given for it's use and further development by Jakob Svensson or any person or organisation
 * directly given such permissions by Oliver Ford or by Jakob Svensson.
 * 
 *  
 * TODO: Replace with XBase based lookup someday,
 */
public class LidarConfigSource extends StateFullNodeImpl {
	public static final String defaultName = "LidarConfigSource";
	
	
	private boolean isInitied = false;
	private boolean isFrozen = false;
	private int configedPulse = -1;
	private int configedFrame = -1;
	
	//Parents
	//private LidarDataProvider dataProv;
	private IntegerValue pulse;
	private IntegerValue frame;
	private IntegerValue nChannels;
	private IntegerValue diag;
	private DoubleMatrix dSrcBaselineX;
	private DoubleMatrix dSrcBaselineY;
	private DoubleMatrix signal;
	private DoubleValue saturationLevel;
	private DoubleArray dSrcRelativeSensitivities;
	private DoubleArray dSrcTPulseToBackwallTimes;
	private DoubleArray dSrcPhotoElectronFactors;
	private DoubleArray dSrcVignettingR;
	private DoubleArray dSrcVignettingAmp;
	private DoubleMatrix dSrcLineOfSightCoords;
	
	//State	
	private double[][] baselineShapeDataX, baselineShapeDataY;
	private double[][] tpulseShapeDataX, tpulseShapeDataY;
	private double[][] switchOnShapeDataX, switchOnShapeDataY;
	private LinearInterpolation1D[] baselineShapes, tpulseShapes, switchOnShapes;
	private double[] baselineX0Min, baselineX0Max, baselineX0Value, baselineX0TypMin, baselineX0TypMax;
	private double[] baselineY0Min, baselineY0Max, baselineY0Value, baselineY0TypMin, baselineY0TypMax;
	private double[] baselineAMin, baselineAMax, baselineAValue, baselineATypMin, baselineATypMax;
	private double[] tpulseX0Min, tpulseX0Max, tpulseX0Value, tpulseX0TypMin, tpulseX0TypMax;
	private double[] tpulseY0Min ,tpulseY0Max, tpulseY0Value, tpulseY0TypMin, tpulseY0TypMax;
	private double[] tpulseAMin, tpulseAMax, tpulseAValue, tpulseATypMin, tpulseATypMax;
	private double[] switchOnX0Min, switchOnX0Max, switchOnX0Value, switchOnX0TypMin, switchOnX0TypMax;
	private double[] switchOnY0Min, switchOnY0Max, switchOnY0Value, switchOnY0TypMin, switchOnY0TypMax;
	private double[] switchOnAMin, switchOnAMax, switchOnAValue, switchOnATypMin, switchOnATypMax;
	private double[] ambientLevelMin, ambientLevelMax, ambientLevelValue, ambientLevelTypMin, ambientLevelTypMax;
	private double[] ambientBeginMin, ambientBeginMax, ambientBeginValue, ambientBeginTypMin, ambientBeginTypMax;
	private double[] backwallX0Min, backwallX0Max, backwallX0Value, backwallX0TypMin, backwallX0TypMax;
	private double[] backwallAMin, backwallAMax, backwallAValue, backwallATypMin, backwallATypMax;
	private double[] backwallScaleLengthMin, backwallScaleLengthMax, backwallScaleLengthValue;
	private double[] backwallScaleLengthTypMin, backwallScaleLengthTypMax;
	private double uberConstMean, uberConstSigma, uberConstValue;
	private double uberConstMin, uberConstMax, uberConstTypMin, uberConstTypMax;
	private double elecSigma;
	private double[] windowTransmission;
	private double[] vignettingProfileX, vignettingProfileY;
	private double[] relativeSensitivitiesMean, relativeSensitivitiesSigma, relativeSensitivitiesValue;
	private double[] relativeSensitivitiesMin, relativeSensitivitiesMax, relativeSensitivitiesTypMin, relativeSensitivitiesTypMax;
	private double laserWidthAdjustMean, laserWidthAdjustSigma, laserWidthAdjustValue; 
	private double laserWidthAdjustMin, laserWidthAdjustMax, laserWidthAdjustTypMin, laserWidthAdjustTypMax;
	private double timingAdjustMean, timingAdjustSigma, timingAdjustValue;
	private double timingAdjustMin, timingAdjustMax, timingAdjustTypMin, timingAdjustTypMax;
	private double[] tpulseToBackwallTimeMean, tpulseToBackwallTimeSigma, tpulseToBackwallTimeValue;
	private double[] tpulseToBackwallTimeMin, tpulseToBackwallTimeMax, tpulseToBackwallTimeTypMin, tpulseToBackwallTimeTypMax;
	private double[] photoElectronFactors;
	private double[][] lineOfSightCoords;
	private int[][] plasmaRegion;
	private boolean[] channelEnable;
	private double vignetPivotR;
	private double vignetShiftMean, vignetShiftSigma, vignetShiftValue;
	private double vignetShiftMin, vignetShiftMax, vignetShiftTypMin, vignetShiftTypMax;
	private double vignetAdjustGradMean, vignetAdjustGradSigma, vignetAdjustGradValue;
	private double vignetAdjustGradMin, vignetAdjustGradMax, vignetAdjustGradTypMin, vignetAdjustGradTypMax;
	
	/** each entry is a peak, outer dim is [4]{channel, A, x0, FWHM}, inner dim is [5]{ value, mean, sigma, min, max }, channel array has inner dim 1*/   
	private ArrayList<double [][]> strayLightPeaks; 
	
	//internal state for saturation relative init
	private int[] satIdx;
	
		
	public LidarConfigSource() { this(defaultName);	}
	
	public LidarConfigSource(String name) {
		super(name);
		//addConnectionPoint(new ConnectionPoint("dataProvider", LidarDataProvider.class, false, getField("dataProv")));
		addConnectionPoint(new ConnectionPoint("frame", IntegerValue.class, false, getField("frame")));
		addConnectionPoint(new ConnectionPoint("signal", DoubleMatrix.class, false, getField("signal")));
		addConnectionPoint(new ConnectionPoint("dSrcBaselineX", DoubleMatrix.class, false, getField("dSrcBaselineX")));
		addConnectionPoint(new ConnectionPoint("dSrcBaselineY", DoubleMatrix.class, false, getField("dSrcBaselineY")));
		addConnectionPoint(new ConnectionPoint("dSrcRelativeSensitivities", DoubleArray.class, false, getField("dSrcRelativeSensitivities")));
		addConnectionPoint(new ConnectionPoint("dSrcTPulseToBackwallTimes", DoubleArray.class, false, getField("dSrcTPulseToBackwallTimes")));
		addConnectionPoint(new ConnectionPoint("dSrcPhotoElectronFactors", DoubleArray.class, false, getField("dSrcPhotoElectronFactors")));		
		addConnectionPoint(new ConnectionPoint("nChannels", IntegerValue.class, false, getField("nChannels")));
		addConnectionPoint(new ConnectionPoint("diag", IntegerValue.class, false, getField("diag")));
		addConnectionPoint(new ConnectionPoint("pulse", IntegerValue.class, false, getField("pulse")));
		addConnectionPoint(new ConnectionPoint("saturationLevel", DoubleValue.class, false, getField("saturationLevel")));
		addConnectionPoint(new ConnectionPoint("dSrcVignettingR", DoubleArray.class, false, getField("dSrcVignettingR")));		
		addConnectionPoint(new ConnectionPoint("dSrcVignettingAmp", DoubleArray.class, false, getField("dSrcVignettingAmp")));
		addConnectionPoint(new ConnectionPoint("dSrcLineOfSightCoords", DoubleMatrix.class, false, getField("dSrcLineOfSightCoords")));	

	}
	
	@Override
	public void updateState() {
		
		
		/* Update logic is a little odd here.
		 * 
		 * Typically, the pulse/frame gets set and the node inits based on the config scripts.
		 * In this process it will take some information from the data source, but the config scripts will change it.
		 * 
		 * The user/app/graph manipulation program may call setters and getters to override these quantities.
		 * 
		 * Any ACTUAL change to the pulse/frame will cause a complete reload.
		 * 
		 *  If the user wishes to avoid this, they can call freezeConfig();
		 */
		
		if( !isInitied || 
			(isAncestorChanged("pulse") && pulse.getInteger() != configedPulse) ||
			(isAncestorChanged("frame") && frame.getInteger() != configedFrame)  ){

			if(isFrozen){
				System.err.println("WARNING: Pulse and/or frame have changed and configuration is fronzen. Calibrations etc will not change to defaults for theis pulse/frame.");
				return;
			}
			
			if(isInitied)
				System.err.println("WARNING: Pulse and/or frame have change, Lidar configuration is being reloaded. All nodes connected with value connected will now be reset.");
			
			
			//changes in the data source/provider cause a full reset of everything
			doEmptyInit();
			doConfigInit();
			if(!isConfigComplete())
				throw new RuntimeException("Config is incomplete!.");
			
			isInitied = true;
			configedPulse = pulse.getInteger();
			configedFrame = frame.getInteger();
			
		}
	}
	
	public void doEmptyInit(){
		int nCh = nChannels.getInteger();

		baselineShapeDataX = new double[nCh][];
		baselineShapeDataY = new double[nCh][];
		tpulseShapeDataX = new double[nCh][];
		tpulseShapeDataY = new double[nCh][];
		switchOnShapeDataX = new double[nCh][];
		switchOnShapeDataY = new double[nCh][];
		baselineShapes = new LinearInterpolation1D[nCh];
		tpulseShapes = new LinearInterpolation1D[nCh];
		switchOnShapes = new LinearInterpolation1D[nCh];
		baselineX0Min = OneLiners.fillArray(Double.NaN, nCh);
		baselineX0Max = OneLiners.fillArray(Double.NaN, nCh);
		baselineX0Value = OneLiners.fillArray(Double.NaN, nCh);
		baselineX0TypMin = OneLiners.fillArray(Double.NaN, nCh);
		baselineX0TypMax = OneLiners.fillArray(Double.NaN, nCh);
		baselineY0Min = OneLiners.fillArray(Double.NaN, nCh);
		baselineY0Max = OneLiners.fillArray(Double.NaN, nCh);
		baselineY0Value = OneLiners.fillArray(Double.NaN, nCh);
		baselineY0TypMin = OneLiners.fillArray(Double.NaN, nCh);
		baselineY0TypMax = OneLiners.fillArray(Double.NaN, nCh);
		baselineAMin = OneLiners.fillArray(Double.NaN, nCh);
		baselineAMax = OneLiners.fillArray(Double.NaN, nCh);
		baselineAValue = OneLiners.fillArray(Double.NaN, nCh);
		baselineATypMin = OneLiners.fillArray(Double.NaN, nCh);
		baselineATypMax = OneLiners.fillArray(Double.NaN, nCh);
		tpulseX0Min = OneLiners.fillArray(Double.NaN, nCh);
		tpulseX0Max = OneLiners.fillArray(Double.NaN, nCh);
		tpulseX0Value = OneLiners.fillArray(Double.NaN, nCh);
		tpulseX0TypMin = OneLiners.fillArray(Double.NaN, nCh);
		tpulseX0TypMax = OneLiners.fillArray(Double.NaN, nCh);
		tpulseY0Min = OneLiners.fillArray(Double.NaN, nCh);
		tpulseY0Max = OneLiners.fillArray(Double.NaN, nCh);
		tpulseY0Value = OneLiners.fillArray(Double.NaN, nCh);
		tpulseY0TypMin = OneLiners.fillArray(Double.NaN, nCh);
		tpulseY0TypMax = OneLiners.fillArray(Double.NaN, nCh);
		tpulseAMin = OneLiners.fillArray(Double.NaN, nCh);
		tpulseAMax = OneLiners.fillArray(Double.NaN, nCh);
		tpulseAValue = OneLiners.fillArray(Double.NaN, nCh);
		tpulseATypMin = OneLiners.fillArray(Double.NaN, nCh);
		tpulseATypMax = OneLiners.fillArray(Double.NaN, nCh);
		switchOnX0Min = OneLiners.fillArray(Double.NaN, nCh);
		switchOnX0Max = OneLiners.fillArray(Double.NaN, nCh);
		switchOnX0Value = OneLiners.fillArray(Double.NaN, nCh);
		switchOnX0TypMin = OneLiners.fillArray(Double.NaN, nCh);
		switchOnX0TypMax = OneLiners.fillArray(Double.NaN, nCh);
		switchOnY0Min = OneLiners.fillArray(Double.NaN, nCh);
		switchOnY0Max = OneLiners.fillArray(Double.NaN, nCh);
		switchOnY0Value = OneLiners.fillArray(Double.NaN, nCh);
		switchOnY0TypMin = OneLiners.fillArray(Double.NaN, nCh);
		switchOnY0TypMax = OneLiners.fillArray(Double.NaN, nCh);
		switchOnAMin = OneLiners.fillArray(Double.NaN, nCh);
		switchOnAMax = OneLiners.fillArray(Double.NaN, nCh);
		switchOnAValue = OneLiners.fillArray(Double.NaN, nCh);
		switchOnATypMin = OneLiners.fillArray(Double.NaN, nCh);
		switchOnATypMax = OneLiners.fillArray(Double.NaN, nCh);
		ambientLevelMin = OneLiners.fillArray(Double.NaN, nCh);
		ambientLevelMax = OneLiners.fillArray(Double.NaN, nCh);
		ambientLevelValue = OneLiners.fillArray(Double.NaN, nCh);
		ambientLevelTypMin = OneLiners.fillArray(Double.NaN, nCh);
		ambientLevelTypMax = OneLiners.fillArray(Double.NaN, nCh);
		ambientBeginMin = OneLiners.fillArray(Double.NaN, nCh);
		ambientBeginMax = OneLiners.fillArray(Double.NaN, nCh);
		ambientBeginValue = OneLiners.fillArray(Double.NaN, nCh);
		ambientBeginTypMin = OneLiners.fillArray(Double.NaN, nCh);
		ambientBeginTypMax = OneLiners.fillArray(Double.NaN, nCh);
		backwallX0Min = OneLiners.fillArray(Double.NaN, nCh);
		backwallX0Max = OneLiners.fillArray(Double.NaN, nCh);
		backwallX0Value = OneLiners.fillArray(Double.NaN, nCh);
		backwallX0TypMin = OneLiners.fillArray(Double.NaN, nCh);
		backwallX0TypMax = OneLiners.fillArray(Double.NaN, nCh);
		backwallAMin = OneLiners.fillArray(Double.NaN, nCh);
		backwallAMax = OneLiners.fillArray(Double.NaN, nCh);
		backwallAValue = OneLiners.fillArray(Double.NaN, nCh);
		backwallATypMin = OneLiners.fillArray(Double.NaN, nCh);
		backwallATypMax = OneLiners.fillArray(Double.NaN, nCh);
		backwallScaleLengthMin = OneLiners.fillArray(Double.NaN, nCh);
		backwallScaleLengthMax = OneLiners.fillArray(Double.NaN, nCh);
		backwallScaleLengthValue = OneLiners.fillArray(Double.NaN, nCh);
		backwallScaleLengthTypMin = OneLiners.fillArray(Double.NaN, nCh);
		backwallScaleLengthTypMax = OneLiners.fillArray(Double.NaN, nCh);
		uberConstMean = Double.NaN; uberConstSigma = Double.NaN; uberConstValue = Double.NaN;
		uberConstMin = Double.NaN; uberConstMax = Double.NaN; uberConstTypMin = Double.NaN; uberConstTypMax = Double.NaN;
		elecSigma = Double.NaN;
		windowTransmission = OneLiners.fillArray(Double.NaN, nCh);
		vignettingProfileX = null;
		vignettingProfileY = null;
		relativeSensitivitiesMean = OneLiners.fillArray(Double.NaN, nCh);
		relativeSensitivitiesSigma = OneLiners.fillArray(Double.NaN, nCh);
		relativeSensitivitiesValue = OneLiners.fillArray(Double.NaN, nCh);
		relativeSensitivitiesMin = OneLiners.fillArray(Double.NaN, nCh);
		relativeSensitivitiesMax = OneLiners.fillArray(Double.NaN, nCh);
		relativeSensitivitiesTypMin = OneLiners.fillArray(Double.NaN, nCh);
		relativeSensitivitiesTypMax = OneLiners.fillArray(Double.NaN, nCh);
		laserWidthAdjustMean = Double.NaN; laserWidthAdjustSigma = Double.NaN; laserWidthAdjustValue = Double.NaN;
		laserWidthAdjustMin = Double.NaN; laserWidthAdjustMax = Double.NaN;
		laserWidthAdjustTypMin = Double.NaN; laserWidthAdjustMax = Double.NaN;
		timingAdjustMean = Double.NaN; timingAdjustSigma = Double.NaN; timingAdjustValue = Double.NaN;
		timingAdjustTypMin = Double.NaN; timingAdjustTypMax = Double.NaN;
		tpulseToBackwallTimeMean = OneLiners.fillArray(Double.NaN, nCh);
		tpulseToBackwallTimeSigma = OneLiners.fillArray(Double.NaN, nCh);
		tpulseToBackwallTimeValue = OneLiners.fillArray(Double.NaN, nCh);
		tpulseToBackwallTimeMin = OneLiners.fillArray(Double.NaN, nCh);
		tpulseToBackwallTimeMax = OneLiners.fillArray(Double.NaN, nCh);
		tpulseToBackwallTimeTypMin = OneLiners.fillArray(Double.NaN, nCh);
		tpulseToBackwallTimeTypMax = OneLiners.fillArray(Double.NaN, nCh);
		photoElectronFactors = OneLiners.fillArray(Double.NaN, nCh);
		lineOfSightCoords = new double[][]{ { Double.NaN, Double.NaN, Double.NaN, }, { Double.NaN, Double.NaN, Double.NaN } };  
		plasmaRegion = new int[nCh][2];
		channelEnable = new boolean[nCh];
		vignetPivotR = Double.NaN;
		vignetShiftMean = Double.NaN;
		vignetShiftSigma = Double.NaN;
		vignetShiftValue = Double.NaN;
		vignetShiftMin = Double.NaN;
		vignetShiftMax = Double.NaN;
		vignetShiftTypMin = Double.NaN;
		vignetShiftTypMax = Double.NaN;
		vignetAdjustGradMean = Double.NaN;
		vignetAdjustGradSigma = Double.NaN;
		vignetAdjustGradValue = Double.NaN;
		vignetAdjustGradMin = Double.NaN;
		vignetAdjustGradMax = Double.NaN;
		vignetAdjustGradTypMin = Double.NaN;
		vignetAdjustGradTypMax = Double.NaN;
		strayLightPeaks = new ArrayList<double[][]>();
	
		satIdx = new int[nCh]; 
		for(int ch=0; ch < nCh; ch++)
			satIdx[ch] = -1;
	}
	
	/** Sets up configuration from config files */
	private void doConfigInit(){
		String configPath = MinervaSettings.instance().getProperty("minerva.lidar.configPath");
		String diagStr = (diag.getInteger() == MinervaLidarDataSource.DIAG_KE3_CORE_LIDAR) ? "ke3" : "ke9";
		String prefix = configPath + "/config-" + diagStr;
		int p = pulse.getInteger();
		int fr = frame.getInteger();
				
		parseScript(prefix + ".txt");	//diag common config first
		
		//get config files list
		File dir = new File(configPath);
		File files[] = dir.listFiles();
		if(files == null){
			System.err.println("WARNING: No files found in LIDAR configs dir '"+configPath+"'");
			return;
		}
		Arrays.sort(files);
				
		//pulse range configs
		for(File f : files){
			//look for any file named "config-[diag][anything]-p####.txt" and split the range by the underscore
			if(f.getName().matches("config-"+diagStr+".*-p[0-9]*_[0-9]*\\.txt")){
				String parts[] = f.getName().replaceAll("config-"+diagStr+".*-p([0-9]*_[0-9]*)\\.txt", "$1").split("_");
				
				if(p >= Integer.parseInt(parts[0]) && p <= Integer.parseInt(parts[1])){
					parseScript(f.getAbsolutePath());
				}
			}
		}
		
		//pulse specific configs
		for(File f : files){
			//look for any file named "config-[diag][anything]-p####.txt" and split the range by the underscore
			if(f.getName().matches("config-"+diagStr+".*-p"+p+"\\.txt"))
				parseScript(f.getAbsolutePath());			
		}
		
		//frame specific configs
		for(File f : files){
			//look for any file named "config-[diag][anything]-p####-f####.txt" and split the range by the underscore
			if(f.getName().matches("config-"+diagStr+".*-p"+p+"-f"+fr+"\\.txt"))
				parseScript(f.getAbsolutePath());
		}
	}
	
	/** Checks the configuration is complete */
	private boolean isConfigComplete(){
		int nCh = nChannels.getInteger();
		if(nCh <= 0)
			return false;
		
		boolean cfgOK = true;

		cfgOK |= checkChannelArray2D("baselineShapeDataX", nCh, baselineShapeDataX);
		cfgOK &= checkChannelArray2D("baselineShapeDataY", nCh, baselineShapeDataY);
		cfgOK &= checkChannelArray2D("tpulseShapeDataX", nCh, tpulseShapeDataX);
		cfgOK &= checkChannelArray2D("tpulseShapeDataY", nCh, tpulseShapeDataY);
		cfgOK &= checkChannelArray2D("switchOnShapeDataX", nCh, switchOnShapeDataX);
		cfgOK &= checkChannelArray2D("switchOnShapeDataY", nCh, switchOnShapeDataY);
		
		//cfgOK &= checkChannelArray1D("baselineShapes", nCh, baselineShapes);
		//cfgOK &= checkChannelArray1D("tpulseShapes", nCh, tpulseShapes);		
		//cfgOK &= checkChannelArray1D("switchOnShapes", nCh, switchOnShapes);
		
		cfgOK &= checkChannelArray1D("baselineX0Min", nCh, baselineX0Min);
		cfgOK &= checkChannelArray1D("baselineX0Max", nCh, baselineX0Max);
		cfgOK &= checkChannelArray1D("baselineX0Value", nCh, baselineX0Value);
		cfgOK &= checkChannelArray1D("baselineX0TypMin", nCh, baselineX0TypMin);
		cfgOK &= checkChannelArray1D("baselineX0TypMax", nCh, baselineX0TypMax);
		cfgOK &= checkChannelArray1D("baselineY0Min", nCh, baselineY0Min);
		cfgOK &= checkChannelArray1D("baselineY0Max", nCh, baselineY0Max);
		cfgOK &= checkChannelArray1D("baselineY0Value", nCh, baselineY0Value);
		cfgOK &= checkChannelArray1D("baselineY0TypMin", nCh, baselineY0TypMin);
		cfgOK &= checkChannelArray1D("baselineY0TypMax", nCh, baselineY0TypMax);
		cfgOK &= checkChannelArray1D("baselineAMin", nCh, baselineAMin);
		cfgOK &= checkChannelArray1D("baselineAMax", nCh, baselineAMax);
		cfgOK &= checkChannelArray1D("baselineAValue", nCh, baselineAValue);
		cfgOK &= checkChannelArray1D("baselineATypMin", nCh, baselineATypMin);
		cfgOK &= checkChannelArray1D("baselineATypMax", nCh, baselineATypMax);
		cfgOK &= checkChannelArray1D("switchOnX0Min", nCh, switchOnX0Min);
		cfgOK &= checkChannelArray1D("switchOnX0Max", nCh, switchOnX0Max);
		cfgOK &= checkChannelArray1D("switchOnX0Value", nCh, switchOnX0Value);
		cfgOK &= checkChannelArray1D("switchOnX0TypMin", nCh, switchOnX0TypMin);
		cfgOK &= checkChannelArray1D("switchOnX0TypMax", nCh, switchOnX0TypMax);
		cfgOK &= checkChannelArray1D("switchOnY0Min", nCh, switchOnY0Min);
		cfgOK &= checkChannelArray1D("switchOnY0Max", nCh, switchOnY0Max);
		cfgOK &= checkChannelArray1D("switchOnY0Value", nCh, switchOnY0Value);
		cfgOK &= checkChannelArray1D("switchOnY0TypMin", nCh, switchOnY0TypMin);
		cfgOK &= checkChannelArray1D("switchOnY0TypMax", nCh, switchOnY0TypMax);
		cfgOK &= checkChannelArray1D("switchOnAMin", nCh, switchOnAMin);
		cfgOK &= checkChannelArray1D("switchOnAMax", nCh, switchOnAMax);
		cfgOK &= checkChannelArray1D("switchOnAValue", nCh, switchOnAValue);
		cfgOK &= checkChannelArray1D("switchOnATypMin", nCh, switchOnATypMin);
		cfgOK &= checkChannelArray1D("switchOnATypMax", nCh, switchOnATypMax);
		cfgOK &= checkChannelArray1D("ambientLevelMin", nCh, ambientLevelMin);
		cfgOK &= checkChannelArray1D("ambientLevelMax", nCh, ambientLevelMax);
		cfgOK &= checkChannelArray1D("ambientLevelValue", nCh, ambientLevelValue);
		cfgOK &= checkChannelArray1D("ambientLevelTypMin", nCh, ambientLevelTypMin);
		cfgOK &= checkChannelArray1D("ambientLevelTypMax", nCh, ambientLevelTypMax);
		cfgOK &= checkChannelArray1D("ambientBeginMin", nCh, ambientBeginMin);
		cfgOK &= checkChannelArray1D("ambientBeginMax", nCh, ambientBeginMax);
		cfgOK &= checkChannelArray1D("ambientBeginValue", nCh, ambientBeginValue);
		cfgOK &= checkChannelArray1D("ambientBeginTypMin", nCh, ambientBeginTypMin);
		cfgOK &= checkChannelArray1D("ambientBeginTypMax", nCh, ambientBeginTypMax);
		cfgOK &= checkChannelArray1D("backwallX0Min", nCh, backwallX0Min);
		cfgOK &= checkChannelArray1D("backwallX0Max", nCh, backwallX0Max);
		cfgOK &= checkChannelArray1D("backwallX0Value", nCh, backwallX0Value);
		cfgOK &= checkChannelArray1D("backwallX0TypMin", nCh, backwallX0TypMin);
		cfgOK &= checkChannelArray1D("backwallX0TypMax", nCh, backwallX0TypMax);
		cfgOK &= checkChannelArray1D("backwallAMin", nCh, backwallAMin);
		cfgOK &= checkChannelArray1D("backwallAMax", nCh, backwallAMax);
		cfgOK &= checkChannelArray1D("backwallAValue", nCh, backwallAValue);
		cfgOK &= checkChannelArray1D("backwallATypMin", nCh, backwallATypMin);
		cfgOK &= checkChannelArray1D("backwallATypMax", nCh, backwallATypMax);
		cfgOK &= checkChannelArray1D("backwallScaleLengthMin", nCh, backwallScaleLengthMin);
		cfgOK &= checkChannelArray1D("backwallScaleLengthMax", nCh, backwallScaleLengthMax);
		cfgOK &= checkChannelArray1D("backwallScaleLengthValue", nCh, backwallScaleLengthValue);
		cfgOK &= checkChannelArray1D("backwallScaleLengthTypMin", nCh, backwallScaleLengthTypMin);
		cfgOK &= checkChannelArray1D("backwallScaleLengthTypMax", nCh, backwallScaleLengthTypMax);
		
		cfgOK &= checkChannelArray1D("relativeSensitivitiesMean", nCh, relativeSensitivitiesMean);
		cfgOK &= checkChannelArray1D("relativeSensitivitiesSigma", nCh, relativeSensitivitiesSigma);
		cfgOK &= checkChannelArray1D("relativeSensitivitiesValue", nCh, relativeSensitivitiesValue);
		cfgOK &= checkChannelArray1D("relativeSensitivitiesMin", nCh, relativeSensitivitiesMin);
		cfgOK &= checkChannelArray1D("relativeSensitivitiesMax", nCh, relativeSensitivitiesMax);
		cfgOK &= checkChannelArray1D("relativeSensitivitiesTypMin", nCh, relativeSensitivitiesTypMin);
		cfgOK &= checkChannelArray1D("relativeSensitivitiesTypMax", nCh, relativeSensitivitiesTypMax);
		cfgOK &= checkChannelArray1D("tpulseToBackwallTimeMean", nCh, tpulseToBackwallTimeMean);
		cfgOK &= checkChannelArray1D("tpulseToBackwallTimeSigma", nCh, tpulseToBackwallTimeSigma);
		cfgOK &= checkChannelArray1D("tpulseToBackwallTimeValue", nCh, tpulseToBackwallTimeValue);
		cfgOK &= checkChannelArray1D("tpulseToBackwallTimeMin", nCh, tpulseToBackwallTimeMin);
		cfgOK &= checkChannelArray1D("tpulseToBackwallTimeMax", nCh, tpulseToBackwallTimeMax);
		cfgOK &= checkChannelArray1D("tpulseToBackwallTimeTypMin", nCh, tpulseToBackwallTimeTypMin);
		cfgOK &= checkChannelArray1D("tpulseToBackwallTimeTypMax", nCh, tpulseToBackwallTimeTypMax);
		cfgOK &= checkChannelArray1D("photoElectronFactors", nCh, photoElectronFactors);
		cfgOK &= checkChannelArray1D("windowTransmission", nCh, windowTransmission);
		
		//tpulse arrays can have different dim to nChannels
		cfgOK &= checkArray1D("tpulseX0Min", tpulseX0Min);
		cfgOK &= checkArray1D("tpulseX0Max", tpulseX0Max);
		cfgOK &= checkArray1D("tpulseX0Value", tpulseX0Value);
		cfgOK &= checkArray1D("tpulseX0TypMin", tpulseX0TypMin);
		cfgOK &= checkArray1D("tpulseX0TypMax", tpulseX0TypMax);
		cfgOK &= checkArray1D("tpulseY0Min", tpulseY0Min);
		cfgOK &= checkArray1D("tpulseY0Max", tpulseY0Max);
		cfgOK &= checkArray1D("tpulseY0Value", tpulseY0Value);
		cfgOK &= checkArray1D("tpulseY0TypMin", tpulseY0TypMin);
		cfgOK &= checkArray1D("tpulseY0TypMax", tpulseY0TypMax);
		cfgOK &= checkArray1D("tpulseAMin", tpulseAMin);
		cfgOK &= checkArray1D("tpulseAMax", tpulseAMax);
		cfgOK &= checkArray1D("tpulseAValue", tpulseAValue);
		cfgOK &= checkArray1D("tpulseATypMin", tpulseATypMin);
		cfgOK &= checkArray1D("tpulseATypMax", tpulseATypMax);
		

		cfgOK &= checkValue("uberConstMean", uberConstMean);
		cfgOK &= checkValue("uberConstSigma", uberConstSigma);
		cfgOK &= checkValue("uberConstValue", uberConstValue);
		cfgOK &= checkValue("uberConstMin", uberConstMin);
		cfgOK &= checkValue("uberConstMax", uberConstMax);
		cfgOK &= checkValue("uberConstTypMin", uberConstTypMin);
		cfgOK &= checkValue("uberConstTypMax", uberConstTypMax);
		cfgOK &= checkValue("elecSigma", elecSigma);
		cfgOK &= checkValue("vignetPivotR", vignetPivotR);
		cfgOK &= checkValue("vignetShiftMean", vignetShiftMean);
		cfgOK &= checkValue("vignetShiftSigma", vignetShiftSigma);
		cfgOK &= checkValue("vignetShiftValue", vignetShiftValue);
		cfgOK &= checkValue("vignetShiftMin", vignetShiftMin);
		cfgOK &= checkValue("vignetShiftMax", vignetShiftMax);
		cfgOK &= checkValue("vignetShiftTypMin", vignetShiftTypMin);
		cfgOK &= checkValue("vignetShiftTypMax", vignetShiftTypMax);
		cfgOK &= checkValue("vignetAdjustGradMean", vignetAdjustGradMean);
		cfgOK &= checkValue("vignetAdjustGradSigma", vignetAdjustGradSigma);
		cfgOK &= checkValue("vignetAdjustGradValue", vignetAdjustGradValue);
		cfgOK &= checkValue("vignetAdjustGradMin", vignetAdjustGradMin);
		cfgOK &= checkValue("vignetAdjustGradMax", vignetAdjustGradMax);
		cfgOK &= checkValue("vignetAdjustGradTypMin", vignetAdjustGradTypMin);
		cfgOK &= checkValue("vignetAdjustGradTypMax", vignetAdjustGradTypMax);
		cfgOK &= checkValue("timingAdjustMean", timingAdjustMean);
		cfgOK &= checkValue("timingAdjustSigma", timingAdjustSigma);
		cfgOK &= checkValue("timingAdjustValue", timingAdjustValue);
		cfgOK &= checkValue("timingAdjustMin", timingAdjustMin);
		cfgOK &= checkValue("timingAdjustMax", timingAdjustMax);
		cfgOK &= checkValue("timingAdjustTypMin", timingAdjustTypMin);
		cfgOK &= checkValue("timingAdjustTypMax", timingAdjustTypMax);		
		cfgOK &= checkValue("laserWidthAdjustMean", laserWidthAdjustMean);
		cfgOK &= checkValue("laserWidthAdjustSigma", laserWidthAdjustSigma);
		cfgOK &= checkValue("laserWidthAdjustValue", laserWidthAdjustValue);
		cfgOK &= checkValue("laserWidthAdjustMin", laserWidthAdjustMin);
		cfgOK &= checkValue("laserWidthAdjustMax", laserWidthAdjustMax);
		cfgOK &= checkValue("laserWidthAdjustTypMin", laserWidthAdjustTypMin);
		cfgOK &= checkValue("laserWidthAdjustTypMax", laserWidthAdjustTypMax);

		cfgOK &= checkArray1D("vignettingProfileX", vignettingProfileX);
		cfgOK &= checkArray1D("vignettingProfileY", vignettingProfileY);
		
		cfgOK &= checkArray2D("lineOfSightCoords", lineOfSightCoords);
		
		if(plasmaRegion == null || plasmaRegion.length != nCh){
			System.err.println("CONFIG ERROR: Config incomplete/invalid: plasmaRegion is null or wrong length");
			cfgOK = false;
		}
		
		if(channelEnable == null || channelEnable.length != nCh){
			System.err.println("CONFIG ERROR: Config incomplete/invalid: channelEnables is null or wrong length");
			cfgOK = false;
		}
		
		return cfgOK;

	}
	
	private boolean checkChannelArray1D(String name, int nCh, double array[]){
		if(array == null){
			System.err.println("CONFIG ERROR: Config incomplete/invalid: " + name + " is null.");			
			return false;
		}
		if(array.length != nCh){
			System.err.println("CONFIG ERROR: Config incomplete/invalid: " + name + " has different dimension ("+array.length+") to nChannels + ("+nCh+")");
			return false;
		}
		return checkArray1D(name, array);
	}
	
	private boolean checkArray1D(String name, double array[]){
		if(array == null){
			System.err.println("CONFIG ERROR: Config incomplete/invalid: " + name + " is null.");			
			return false;
		}
		for(int i=0; i < array.length; i++)
			if(!checkValue(name+"["+i+"]", array[i]))
				return false;
		return true;
	}
	
	private boolean checkValue(String name, double value){
		if(Double.isNaN(value)){
			System.err.println("CONFIG ERROR: Config incomplete/invalid '" + name + "'.");
			return false;
		}
		return true;
	}
	
	private boolean checkChannelArray2D(String name, int nCh, double array[][]){
		if(array == null){
			System.err.println("CONFIG ERROR: Config incomplete/invalid: " + name + " is null.");			
			return false;
		}
		if(array.length != nCh){
			System.err.println("CONFIG ERROR: Config incomplete/invalid: " + name + " has different dimension ("+array.length+") to nChannels ("+nCh+")");
			return false;
		}
		for(int i=0; i < nCh; i++){			
			if(!checkArray1D(name + "[ch="+i+"]", array[i]))
				return false;
		}
		return true;
	}
	
	private boolean checkArray2D(String name, double array[][]){		
		if(array == null){
			System.err.println("CONFIG ERROR: Config incomplete/invalid: " + name + " is null.");			
			return false;
		}
		for(int i=0; i < array.length; i++){			
			if(!checkArray1D(name + "[ch="+i+"]", array[i]))
				return false;
		}
		return true;
	}
	
	
	private boolean parseScript(String fileName){
		
		if(!new File(fileName).canRead())
			return false;
		
		if(MinervaSettings.getDbgLevel() >= 3)
			System.out.println("Parsing LIDAR config script '"+fileName+"'.");
		
		SimpleScriptParser.parse(this, fileName, "script");
			
		return true;
	}
	
	/* ************* Config Ops (Typically from Scripts) ************ */
	
	public void scriptLoadTPulses(int ch, String pathTemplate){
		loadShapeData(ch, pathTemplate, tpulseShapeDataX, tpulseShapeDataY, tpulseShapes, LinearInterpolation1D.EXTRAPOLATE_CONSTANT_VALUE);
		setChanged();		
	}

	public void scriptLoadSwitchOnPulses(int ch, String pathTemplate){
		loadShapeData(ch, pathTemplate, switchOnShapeDataX, switchOnShapeDataY, switchOnShapes, LinearInterpolation1D.EXTRAPOLATE_CONSTANT_VALUE);
		setChanged();
	}		

	public void scriptLoadBaselines(int ch, String pathTemplate){
		loadShapeData(ch, pathTemplate, baselineShapeDataX, baselineShapeDataY, baselineShapes, LinearInterpolation1D.EXTRAPOLATE_CONSTANT_END_KNOT);
		setChanged();
	}
	
	public void loadShapeData(int ch, String pathTemplate, double dataX[][], double dataY[][],
			LinearInterpolation1D shapes[], int extrapolationType){
		int nCh = nChannels.getInteger();
		
		if(ch < 0){
			for(int i=0; i < nCh; i++)
				scriptLoadTPulses(i, pathTemplate);
			
		}else{				
			String shapesPath = MinervaSettings.instance().getProperty("minerva.lidar.shapesPath");
			String fileName = pathTemplate.replaceAll("%CH%", Integer.toString(ch)).replaceAll("%SHAPESPATH%", shapesPath);
			double dataIn[][] = AsciiMatrixFile.mustLoad(fileName, true);
			
			dataX[ch] = dataIn[0];
			dataY[ch] = dataIn[1];
			
			shapes[ch] = new LinearInterpolation1D(dataX[ch], dataY[ch], extrapolationType);
		}
		
	}
	
	
	
	public void scriptSingleTimingPulse(){
		tpulseX0Min = new double[]{ Double.NaN };
		tpulseX0Max = new double[]{ Double.NaN };
		tpulseX0Value = new double[]{ Double.NaN };
		tpulseX0TypMin = new double[]{ Double.NaN };
		tpulseX0TypMax = new double[]{ Double.NaN };
		tpulseY0Min = new double[]{ Double.NaN };
		tpulseY0Max = new double[]{ Double.NaN };
		tpulseY0Value = new double[]{ Double.NaN };
		tpulseY0TypMin = new double[]{ Double.NaN };
		tpulseY0TypMax = new double[]{ Double.NaN };
		tpulseAMin = new double[]{ Double.NaN };
		tpulseAMax = new double[]{ Double.NaN };
		tpulseAValue = new double[]{ Double.NaN };
		tpulseATypMin = new double[]{ Double.NaN };
		tpulseATypMax = new double[]{ Double.NaN };
		

		setChanged();
	}
	
	public void scriptSwitchOnFromTPulses(){
		switchOnShapeDataX = OneLiners.copyArray(tpulseShapeDataX);
		switchOnShapeDataY = OneLiners.copyArray(tpulseShapeDataY);
		switchOnShapes = tpulseShapes;
		
		setChanged();
	}
	
	public void scriptBaselinesFromDataSource(){
		
		baselineShapeDataX = OneLiners.copyArray(dSrcBaselineX.getDoubleMatrix());
		baselineShapeDataY =  OneLiners.copyArray(dSrcBaselineY.getDoubleMatrix());
		
		for(int ch=0; ch < baselineShapeDataX.length; ch++)
			baselineShapes[ch] = new LinearInterpolation1D(baselineShapeDataX[ch], baselineShapeDataY[ch], LinearInterpolation1D.EXTRAPOLATE_CONSTANT_END_KNOT);
		
		setChanged();
	}
	
	public void scriptUseFlatBaselines(){
		int nCh = nChannels.getInteger(); 
		int sigLen = signal.getDoubleMatrix()[0].length;
		baselineShapeDataX = new double[nCh][];
		baselineShapeDataY = new double[nCh][];
				
		for(int ch=0; ch < nCh; ch++){
			baselineShapeDataX[ch] = new double[]{ 0, sigLen };
			baselineShapeDataY[ch] = new double[]{ 0, 0 };
				
			baselineShapes[ch] = new LinearInterpolation1D(baselineShapeDataX[ch], baselineShapeDataY[ch], LinearInterpolation1D.EXTRAPOLATE_CONSTANT_END_KNOT);
		}
		
		setChanged();
		
	}

	/** calc the photo electron factors (ADCU per photoelectron per ADCU bin) */
	public void scriptPhotoElectronFactorsFromDataSource(){
		photoElectronFactors = dSrcPhotoElectronFactors.getDoubleArray().clone();
		setChanged();
	}
	
	/** Set relativeSensitivity mean and init value to the one from the data provider.
	 */
	public void scriptRelativeSensitivitiesFromDataSource(double fractionalSigma){
		//
		relativeSensitivitiesValue = dSrcRelativeSensitivities.getDoubleArray().clone();		
		for(int ch=0; ch < relativeSensitivitiesValue.length; ch++){
			relativeSensitivitiesMean[ch] = relativeSensitivitiesValue[ch];
			relativeSensitivitiesSigma[ch] = relativeSensitivitiesValue[ch] * fractionalSigma;
			relativeSensitivitiesMin[ch] = Math.max(0, relativeSensitivitiesValue[ch] * (1 - 5 * fractionalSigma));
			relativeSensitivitiesMax[ch] = relativeSensitivitiesValue[ch] * (1 + 5 * fractionalSigma);
			relativeSensitivitiesTypMin[ch] = Math.max(0, relativeSensitivitiesValue[ch] * (1 - 3 * fractionalSigma));
			relativeSensitivitiesTypMax[ch] = relativeSensitivitiesValue[ch] * (1 + 3 * fractionalSigma);
		}
	}
	
	public void  scriptVignettingFromDataSource(){
		vignettingProfileX = dSrcVignettingR.getDoubleArray().clone();
		vignettingProfileY = dSrcVignettingAmp.getDoubleArray().clone();
		setChanged();
	}
	
	/** Set tpulseToBackwallTimeValue mean and init value to the one from the data provider.
	 * Use +/-1sigma = 5% for uncertainty.
	 */
	public void scriptTPulseToBackwallFromDataSource(double sigma){
		tpulseToBackwallTimeValue = dSrcTPulseToBackwallTimes.getDoubleArray().clone();		
		for(int ch=0; ch < tpulseToBackwallTimeValue.length; ch++){
			tpulseToBackwallTimeMean[ch] = tpulseToBackwallTimeValue[ch];
			tpulseToBackwallTimeSigma[ch] = sigma;
			tpulseToBackwallTimeMin[ch] = tpulseToBackwallTimeValue[ch] - 5 * sigma;
			tpulseToBackwallTimeMax[ch] = tpulseToBackwallTimeValue[ch] + 5 * sigma;
			tpulseToBackwallTimeTypMin[ch] = tpulseToBackwallTimeValue[ch] - 3 * sigma;
			tpulseToBackwallTimeTypMax[ch] = tpulseToBackwallTimeValue[ch] + 3 * sigma;
		}	
		setChanged();
	}
	
	public void scriptLineOfSightFromDataSource(){
		lineOfSightCoords = OneLiners.copyArray(dSrcLineOfSightCoords.getDoubleMatrix());		
		setChanged();
	}
	
	public void scriptLineOfSightFromDataSource(double minR, double maxR){
		double los[][] = dSrcLineOfSightCoords.getDoubleMatrix();
		
		if(minR >= maxR)
			throw new IllegalArgumentException("minimumR >= maximumR ("+minR+" >= "+maxR+").");
	
		double R0 = los[0][0], Z0 = los[1][0];		
		double R1 = los[0][1], Z1 = los[1][1];
		double Rbackwall = los[0][2], Zbackwall = los[1][2];

		if(!Double.isNaN(minR)){			
			if(R0 < minR){
				if(R1 < minR)
					throw new IllegalArgumentException("Both ends of LOS R=("+R0+", "+R1+") are < minimumR ("+minR+").");
				Z0 += (minR - R0) * (Z1 - Z0) / (R1 - R0);
				R0 = minR;
			}else if(R1 < minR){
				Z1 += (minR - R1) * (Z0 - Z1) / (R0 - R1);
				R1 = minR;
			}
		}
		if(!Double.isNaN(maxR)){			
			if(R0 > maxR){
				if(R1 > maxR)
					throw new IllegalArgumentException("Both ends of LOS R=("+R0+", "+R1+") are > maximumR ("+maxR+").");
				Z0 += (maxR - R0) * (Z1 - Z0) / (R1 - R0);
				R0 = maxR;
			}else if(R1 > maxR){
				Z1 += (maxR - R1) * (Z0 - Z1) / (R0 - R1);
				R1 = maxR;
			}
		}
		
		lineOfSightCoords = new double[][]{ {R0, R1, Rbackwall }, {Z0, Z1, Zbackwall} };
		setChanged();
	}
	
	public void scriptSaturationRelativeInitBaselineFromData(int ch, int X0, int X1, double rangeA, double rangeX, double rangeY){
		if(ch < 0){
			for(int i=0; i < nChannels.getInteger(); i++)
				scriptSaturationRelativeInitBaselineFromData(i, X0, X1, rangeA, rangeX, rangeY);
			return;
		}
		if(satIdx[ch] < 0)
			throw new RuntimeException("No saturation point stored for channel "+ch);
		scriptInitBaselineFromData(ch, satIdx[ch]+X0, satIdx[ch]+X1, rangeA, rangeX, rangeY);
		setChanged();
	}
	
	/** Initialise baseline parameters from basic calculations on the data.
	 * The X shift is set to 0.
	 * @param X0 Start of averaging window for Y level
	 * @param X1 End of averaging window for Y level
	 * @param rangeA Normal variance range for ampltiude (0 for fixed at 1) 
	 * @param rangeX Normal variance range for X shift
	 * @param rangeY Normal variance range for Y shift 
	 */
	public void scriptInitBaselineFromData(int ch, int X0, int X1, double rangeA, double rangeX, double rangeY){
		
		if(ch < 0){
			for(int i=0; i < nChannels.getInteger(); i++)
				scriptInitBaselineFromData(i, X0, X1, rangeA, rangeX, rangeY);
			return;
		}

		double a=0;	// average of data in window
		double b=0; // average of baseline shape in window
		
		double data[][] = signal.getDoubleMatrix();  
								
		for(int i=X0;i<=X1;i++){
			a += data[ch][i];
			b += baselineShapes[ch].eval(i);
		}
			
		a /= (X1+1-X0);
		b /= (X1+1-X0);
			
		baselineY0Value[ch] = a-b; 
		baselineY0Min[ch] = a-b-rangeY/2; 
		baselineY0Max[ch] = a-b+rangeY/2;
		baselineY0TypMin[ch] = baselineY0Min[ch]; 
		baselineY0TypMax[ch] = baselineY0Max[ch];
		 
		baselineX0Value[ch]	= 0.0; 
		baselineX0Min[ch] = -rangeX/2; 
		baselineX0Max[ch] = rangeX/2;
		baselineX0TypMin[ch] = baselineX0Min[ch]; 
		baselineX0TypMax[ch] = baselineX0Max[ch];
		
		baselineAValue[ch] = 1.0;
		baselineAMin[ch] = 0;
		baselineAMax[ch] = (rangeA == 0) ? 2 : rangeA;
		baselineATypMin[ch] = baselineAMin[ch]; 
		baselineATypMax[ch] = baselineAMax[ch];			
		setChanged();		
	}
	
	public void scriptSaturationRelativeInitAmbientFromData(int ch, int X0, int X1, double threshold){
		if(ch < 0){
			for(int i=0; i < nChannels.getInteger(); i++)
				scriptSaturationRelativeInitAmbientFromData(i, X0, X1, threshold);
			return;
		}
	
		if(satIdx[ch] < 0)
			throw new RuntimeException("No saturation point stored for channel "+ch);
		
		scriptInitAmbientFromData(ch, satIdx[ch]+X0, satIdx[ch]+X1, threshold);
		setChanged();
	}
	
	public void scriptInitAmbientFromData(int ch, int X0, int X1, double threshold){
		if(ch < 0){
			for(int i=0; i < nChannels.getInteger(); i++)
				scriptInitAmbientFromData(i, X0, X1, threshold);
			return;
		}
		
		double data[][] = signal.getDoubleMatrix();  
			
		//find the average of 150-160 with baseline subtracted and use as ambient light level
		double a=0;
		for(int i=X0;i<=X1;i++){
			double baseline = baselineY0Value[ch] + baselineAValue[ch] * baselineShapes[ch].eval(i - baselineX0Value[ch]);
			a += data[ch][i] - baseline;
		}
		a /= (X1+1-X0);		
			
		ambientLevelMin[ch] = 0;
		if(a > threshold){				
			a /= photoElectronFactors[ch];
			ambientLevelValue[ch] = a;
			ambientLevelMax[ch] = 2 * a;			
		}else{ //too small to tell, use 0 and allow +ve adjustment
			ambientLevelValue[ch] = 0;
			ambientLevelMax[ch] = 20 / photoElectronFactors[ch];
			
		}	
		ambientLevelTypMin[ch] = ambientLevelMin[ch];
		ambientLevelTypMax[ch] = ambientLevelMax[ch];
		setChanged();
	}

	public void scriptSaturationRelativeInitTimingPulseFromData(int  ch, int X0, int X1){
		if(ch < 0){
			for(int i=0; i < nChannels.getInteger(); i++)
				scriptSaturationRelativeInitTimingPulseFromData(i, X0, X1);
			return;
		}
		
		if(satIdx[ch] < 0)
			throw new RuntimeException("No saturation point stored for channel "+ch);
		
		scriptInitTimingPulseFromData(ch, satIdx[ch]+X0, satIdx[ch]+X1);
		setChanged();
	}
	
	
	public void scriptInitTimingPulseFromData(int  ch, int X0, int X1){
		if(ch < 0){
			for(int i=0; i < nChannels.getInteger(); i++)
				scriptInitTimingPulseFromData(i, X0, X1);
			return;
		}
		
		if(X1<=X0 || X0 < 0 || X1 < 0){ //no timing pulse on this channel
			tpulseX0Value[ch] = Double.NaN;
			tpulseX0Min[ch] = Double.NaN;
			tpulseX0Max[ch] = Double.NaN;

			tpulseX0TypMin[ch] = tpulseX0Min[ch]; 
			tpulseX0TypMax[ch] = tpulseX0Max[ch];				
		}else{			
			matchShape(ch, tpulseShapes,
					tpulseX0Min, tpulseAMin,
					tpulseX0Max, tpulseAMax,
					tpulseX0TypMin, tpulseATypMin,
					tpulseX0TypMax, tpulseATypMax,
					tpulseX0Value, tpulseAValue, X0, X1);			
		}
		
		tpulseY0Value[ch] = 0;
		tpulseY0Min[ch] = 0;
		tpulseY0Max[ch] = 1;
		tpulseY0TypMin[ch] = tpulseY0Min[ch];
		tpulseY0TypMax[ch] = tpulseY0Max[ch];	
		setChanged();
		
	}
	
	public void scriptSaturationRelativeInitSwitchOnPulseFromData(int ch, int X0, int X1){
		if(ch < 0){
			for(int i=0; i < nChannels.getInteger(); i++)
				scriptSaturationRelativeInitSwitchOnPulseFromData(i, X0, X1);
			return;
		}
		
		if(satIdx[ch] < 0)
			throw new RuntimeException("No saturation point stored for channel "+ch);
		
		scriptInitSwitchOnPulseFromData(ch, satIdx[ch]+X0, satIdx[ch]+X1);
		setChanged();
	}
	
	public void scriptInitSwitchOnPulseFromData(int ch, int X0, int X1){
		if(ch < 0){
			for(int i=0; i < nChannels.getInteger(); i++)
				scriptInitSwitchOnPulseFromData(i, X0, X1);
			return;
		}
		
		double a = matchShape(ch, switchOnShapes,
				switchOnX0Min, switchOnAMin,
				switchOnX0Max, switchOnAMax,
				switchOnX0TypMin, switchOnATypMin,
				switchOnX0TypMax, switchOnATypMax,
				switchOnX0Value, switchOnAValue, X0, X1);		
		
		//double a = switchonX0[ch].get();
		ambientBeginValue[ch] = a - 5;
		ambientBeginMin[ch] = a - 10;
		ambientBeginMax[ch] = a;
		ambientBeginTypMin[ch] = ambientBeginMin[ch];
		ambientBeginTypMax[ch] = ambientBeginMax[ch];
		
		switchOnY0Value[ch] = 0;
		switchOnY0Min[ch] = 0;
		switchOnY0Max[ch] = 1;
		switchOnY0TypMin[ch] = switchOnY0Min[ch]; 
		switchOnY0TypMax[ch] = switchOnY0Max[ch];
		setChanged();
	}
	
	public void scriptFindSaturationPoint(int ch, int X0, int X1){
		if(ch < 0){
			for(int i=0; i < nChannels.getInteger(); i++)
				scriptFindSaturationPoint(i, X0, X1);
			return;
		}
		
		double data[][] = signal.getDoubleMatrix();  
		double satLevel = saturationLevel.getDouble();
		
		//find the first point of saturation inthe given Range
		
		for(satIdx[ch]=X0; satIdx[ch] <= X1; satIdx[ch]++){
			if(data[ch][satIdx[ch]] >= satLevel)
				break;
		}
		if(satIdx[ch] > X1){
			//throw new RuntimeException("Signal doesn't seem to saturate, can't determine backwall position.");
			System.err.println("Signal doesn't seem to saturate, can't determine backwall position. " +
					"Will try to use highest value in the given range ("+X0+" <= X <= "+X1+")");
			satIdx[ch]=X0;			
			for(int i=X0; i < X1; i++)
				if(data[ch][i] > data[ch][satIdx[ch]])
					satIdx[ch] = i;
		}
		setChanged();
	}
	
	public void scriptInitBackwall(int ch, int X0, int X1, int rangeX, double L0, double L, double L1){
		if(ch < 0){
			for(int i=0; i < nChannels.getInteger(); i++)
				scriptInitBackwall(i, X0, X1, rangeX, L0, L, L1);
			return;
		}
		
		double data[][] = signal.getDoubleMatrix();  
		double satLevel = saturationLevel.getDouble();
		
		/*int maxIdx = 0;
		double maxVal = Double.NEGATIVE_INFINITY;
		for(int i=220;i<256;i++){
			tVar.set(i);
			//get data height and subtract ambient light and baseline
			double val = data[ch][i] - baselineTransShape[ch].evaluate() - (ambientLightLevel[ch].get() * facts[ch]);
			if(val > 1.20*maxVal){ //has to go up by more than 20% - to stop changes in baseline against saturation level affecting it
				maxVal = val;
				maxIdx = i;					
			}
		}*/
	
		if(satIdx[ch] <= 0)
			scriptFindSaturationPoint(ch, X0, X1);
	
		double a = satLevel / photoElectronFactors[ch]; //set amplitude as saturation level
		backwallAValue[ch] = a;
		backwallAMin[ch] = 0.9 * a;
		backwallAMax[ch] = 1.1 * a;
		backwallATypMin[ch] = backwallAMin[ch];
		backwallATypMax[ch] = backwallAMax[ch]; 
		
		backwallX0Value[ch] = satIdx[ch];
		backwallX0Min[ch] = satIdx[ch] - rangeX / 2;
		backwallX0Max[ch] = satIdx[ch] + rangeX / 2;
		backwallX0TypMin[ch] = backwallX0Min[ch];
		backwallX0TypMax[ch] = backwallX0Max[ch]; 
		 
		backwallScaleLengthValue[ch] = L;
		backwallScaleLengthMin[ch] = L0;
		backwallScaleLengthMax[ch] = L1;
		backwallScaleLengthTypMin[ch] = backwallScaleLengthMin[ch];
		backwallScaleLengthTypMax[ch] = backwallScaleLengthMax[ch];
		setChanged();
		
	}
	
	public void scriptClearStrayLight(){
		strayLightPeaks = new ArrayList<double[][]>();
		setChanged();
	}
	
	public void scriptAddSaturationRelativeStrayLight(int ch,
			double A, double AMean, double ASigma, double AMin, double AMax,
			double X, double XMean, double XSigma, double XMin, double XMax,
			double W, double WMean, double WSigma, double WMin, double WMax ){
		
		int sigLen = signal.getDoubleMatrix()[0].length;
		
		X += satIdx[ch]; XMean += satIdx[ch]; XMin += satIdx[ch]; XMax += satIdx[ch];
		
		if(X<0)X=0; if(X>=sigLen)X=sigLen-1;
		if(XMean<0)XMean=0; if(XMean>=sigLen)XMean=sigLen-1;
		if(XMin<0)XMin=0; if(XMin>=sigLen)XMin=sigLen-1;
		if(XMax<0)XMax=0; if(XMax>=sigLen)XMax=sigLen-1;
		
		strayLightPeaks.add(new double[][]{
						{ ch },
						{ A, AMean, ASigma, AMin, AMax }, 
						{ X, XMean, XSigma, XMin, XMax }, 
						{ W, WMean, WSigma, WMin, WMax }	});

		setChanged();
	}
						
	public void scriptElecSigma(double val){ elecSigma = val; }

	public void scriptLineOfSightCoords(double R0, double R1, double R2, double Z0, double Z1, double Z2){
		lineOfSightCoords = new double[][]{ {R0, R1, R2}, {Z0, Z1, Z2} };
		setChanged();
	}
			
	/** matches a given shape's X0 and A to the provided data in range i0-i1 inclusive, including baseline and ambient light*/
	private double matchShape(int ch, LinearInterpolation1D shape[],
			double X0Min[], double AMin[],
			double X0Max[], double AMax[],
			double X0TypMin[], double ATypMin[],
			double X0TypMax[], double ATypMax[],
			double X0Value[], double AValue[],			
			int i0, int i1){
		
		double data[][] = signal.getDoubleMatrix();  
		
		//first find the peak height of the given shape
		double shapePeak = Double.NEGATIVE_INFINITY;
		double shapePeakPos = 0;
		double nodes[] = shape[ch].getF();
		double nodeX[] = shape[ch].getX();
		for(int i=0;i<nodes.length;i++){
			double val = nodes[i];
			if(val > shapePeak){
				shapePeakPos = nodeX[i];
				shapePeak = val;
			}
		}
		
		//now find the peak in the data
		double dataPeak = Double.NEGATIVE_INFINITY;
		int dataPeakPos = 0;
		for(int i=i0;i<i1;i++){
			//get data height and subtract ambient light and baseline
			double baseline = baselineY0Value[ch] + baselineAValue[ch] * baselineShapes[ch].eval(i - baselineX0Value[ch]);			
			double val = data[ch][i] - baseline - (ambientLevelValue[ch] * photoElectronFactors[ch]);
			if(val > dataPeak){
				dataPeakPos = i;
				dataPeak = val;
			}				
		}
		
		//now set the shape translation to the peak position (assume shape's peak is at it's internal 0)
		X0Value[ch] =  dataPeakPos-shapePeakPos;
		X0Min[ch] = dataPeakPos - 10;
		X0Max[ch] = dataPeakPos + 10;
		X0TypMin[ch] = X0Min[ch];
		X0TypMax[ch] = X0Max[ch];
		
		if(dataPeak > 10){		
			//and set the scale to get the heights to match
			double scale = dataPeak / (shapePeak * photoElectronFactors[ch]);
			AValue[ch] =  scale;
			AMin[ch] = 0.5 * scale;
			AMax[ch] = 2 * scale;			
		}else{ //too small to tell, use 0 and allow +ve adjustment
			AValue[ch] =  0;
			AMin[ch] = 0;
			AMax[ch] = 2;			
		}
		ATypMin[ch] = AMin[ch];
		ATypMax[ch] = AMax[ch];
		
		return dataPeakPos;
	}
	
	public void scriptWindowTransmission(String p[]){
		for(int i=0; i < windowTransmission.length; i++)
			windowTransmission[i] = Double.parseDouble(p[i]);
		setChanged();
	}
				
	public void scriptChannelEnable(boolean enable){ 
		for(int i=0; i < channelEnable.length; i++)
			channelEnable[i] = enable;
		setChanged();
	}
	
	public void scriptChannelEnable(int ch, boolean enable){ 
		if(ch < 0)
			scriptChannelEnable(enable);
		else
			channelEnable[ch] = enable;
		setChanged();
	}	
	
	/** Replaces the data providers photo electron factors */
	public void scriptOverridePhotoElectronFactor(int ch, double newPEFactor, 
			boolean adjustRelativeSensitivities, boolean adjustSignalComponents){
		
		if(adjustRelativeSensitivities){
			/* The calibrations done actually measure:
			 * a) the peFactors (photocathode to ADC) and 
			 * b) the overall transmission excluding windows (windows to ADC). 
			 * 
			 * The relative sensitivities are calculated by (c) = (b) / (a), giving the transmission
			 * of the path from the windows to photocathode. Here, we adjust (a) but still believe (b),
			 * so we recalculate (c) as if we had the measurement of (b) and are using our
			 * new overridden value for (a).
			 */
						
			relativeSensitivitiesValue[ch] *= photoElectronFactors[ch] / newPEFactor;
			
			//rescale whatever may have already been set as the relSens sigmas and ranges etc
			relativeSensitivitiesMean[ch] = relativeSensitivitiesValue[ch];
			relativeSensitivitiesSigma[ch] *= photoElectronFactors[ch] / newPEFactor;
			relativeSensitivitiesMin[ch] *= photoElectronFactors[ch] / newPEFactor;
			relativeSensitivitiesMax[ch] *= photoElectronFactors[ch] / newPEFactor;
			relativeSensitivitiesTypMin[ch] *= photoElectronFactors[ch] / newPEFactor;
			relativeSensitivitiesTypMax[ch] *= photoElectronFactors[ch] / newPEFactor;
			
			
			//and use +/- 1sigma = 5%, they can override to something static if neccessary
			/*relativeSensitivitiesMean[ch] = relativeSensitivitiesValue[ch];
			relativeSensitivitiesSigma[ch] = 0.05 * relativeSensitivitiesValue[ch];
			relativeSensitivitiesMin[ch] = 0.80 * relativeSensitivitiesValue[ch];
			relativeSensitivitiesMax[ch] = 1.20 * relativeSensitivitiesValue[ch];
			relativeSensitivitiesTypMin[ch] = 0.75 * relativeSensitivitiesValue[ch];
			relativeSensitivitiesTypMax[ch] = 1.15 * relativeSensitivitiesValue[ch];
			*/  
		}
		
		if(adjustSignalComponents){
			//signal component amplitudes are in photoelectrons
			//so also may need adjusting
			ambientLevelValue[ch] *= photoElectronFactors[ch] / newPEFactor;
			ambientLevelMin[ch] *= photoElectronFactors[ch] / newPEFactor;
			ambientLevelMax[ch] *= photoElectronFactors[ch] / newPEFactor;
			ambientLevelTypMin[ch] *= photoElectronFactors[ch] / newPEFactor;
			ambientLevelTypMax[ch] *= photoElectronFactors[ch] / newPEFactor;
			
			if(ch < tpulseAValue.length){ //skip if single timing pulse
				tpulseAValue[ch] *= photoElectronFactors[ch] / newPEFactor;
				tpulseAMin[ch] *= photoElectronFactors[ch] / newPEFactor;
				tpulseAMax[ch] *= photoElectronFactors[ch] / newPEFactor;
				tpulseATypMin[ch] *= photoElectronFactors[ch] / newPEFactor;
				tpulseATypMax[ch] *= photoElectronFactors[ch] / newPEFactor;
			}

			switchOnAValue[ch] *= photoElectronFactors[ch] / newPEFactor;
			switchOnAMin[ch] *= photoElectronFactors[ch] / newPEFactor;
			switchOnAMax[ch] *= photoElectronFactors[ch] / newPEFactor;
			switchOnATypMin[ch] *= photoElectronFactors[ch] / newPEFactor;
			switchOnATypMax[ch] *= photoElectronFactors[ch] / newPEFactor;
			
			backwallAValue[ch] *= photoElectronFactors[ch] / newPEFactor;
			backwallAMin[ch] *= photoElectronFactors[ch] / newPEFactor;
			backwallAMax[ch] *= photoElectronFactors[ch] / newPEFactor;
			backwallATypMin[ch] *= photoElectronFactors[ch] / newPEFactor;
			backwallATypMax[ch] *= photoElectronFactors[ch] / newPEFactor;
		}
		
		photoElectronFactors[ch] = newPEFactor;	
		setChanged();	
	}
	
	public void scriptSaturationRelativePlasmaRegion(	String p[] ){
		int sigLen = signal.getDoubleMatrix()[0].length;
		for(int i=0; i < plasmaRegion.length; i++){
			if(satIdx[i] < 0)
				throw new RuntimeException("No saturation point stored for channel "+i);
			int i0 = satIdx[i]+Integer.parseInt(p[i]);
			int i1 = satIdx[i]+Integer.parseInt(p[plasmaRegion.length + i]);
			
			plasmaRegion[i] = new int[]{ Math.max(i0, 0), Math.min(i1, sigLen-1) };
		}
		setChanged();
	}
	
	public void scriptPlasmaRegion(	String p[] ){
		for(int i=0; i < plasmaRegion.length; i++)
			plasmaRegion[i] = new int[]{ Integer.parseInt(p[i]), Integer.parseInt(p[plasmaRegion.length + i]) };

		setChanged();
	}
	
	public void scriptWarning( String s ){
		System.err.println("WARNING: " + s);
	}
	
	public void scriptError( String s ){
		throw new RuntimeException("Script induced error: " + s);
	}
			
	private void setChannelArray(double array[], int ch, double val){		
		if(ch >= 0)
			array[ch] = val;
		else
			for(int i=0;i < array.length; i++)
				array[i] = val;

		setChanged();
	}	
	
	/** ************ bulk script signalcomponents setting *********** **/
	public void scriptBaselineX0Min(int ch, double val){ setChannelArray(baselineX0Min, ch, val); setChanged(); }
	public void scriptBaselineX0Max(int ch, double val){ setChannelArray(baselineX0Max, ch, val); setChanged(); }
	public void scriptBaselineX0Value(int ch, double val){ setChannelArray(baselineX0Value, ch, val); setChanged(); }
	public void scriptBaselineX0TypMin(int ch, double val){ setChannelArray(baselineX0TypMin, ch, val); setChanged(); }
	public void scriptBaselineX0TypMax(int ch, double val){ setChannelArray(baselineX0TypMax, ch, val); setChanged(); }
	public void scriptBaselineY0Min(int ch, double val){ setChannelArray(baselineY0Min, ch, val); setChanged(); }
	public void scriptBaselineY0Max(int ch, double val){ setChannelArray(baselineY0Max, ch, val); setChanged(); }
	public void scriptBaselineY0Value(int ch, double val){ setChannelArray(baselineY0Value, ch, val); setChanged(); }
	public void scriptBaselineY0TypMin(int ch, double val){ setChannelArray(baselineY0TypMin, ch, val); setChanged(); }
	public void scriptBaselineY0TypMax(int ch, double val){ setChannelArray(baselineY0TypMax, ch, val); setChanged(); }
	public void scriptBaselineAMin(int ch, double val){ setChannelArray(baselineAMin, ch, val); setChanged(); }
	public void scriptBaselineAMax(int ch, double val){ setChannelArray(baselineAMax, ch, val); setChanged(); }
	public void scriptBaselineAValue(int ch, double val){ setChannelArray(baselineAValue, ch, val); setChanged(); }
	public void scriptBaselineATypMin(int ch, double val){ setChannelArray(baselineATypMin, ch, val); setChanged(); }
	public void scriptBaselineATypMax(int ch, double val){ setChannelArray(baselineATypMax, ch, val); setChanged(); }
	public void scriptTpulseX0Min(int ch, double val){ setChannelArray(tpulseX0Min, ch, val); setChanged(); }
	public void scriptTpulseX0Max(int ch, double val){ setChannelArray(tpulseX0Max, ch, val); setChanged(); }
	public void scriptTpulseX0Value(int ch, double val){ setChannelArray(tpulseX0Value, ch, val); setChanged(); }
	public void scriptTpulseX0TypMin(int ch, double val){ setChannelArray(tpulseX0TypMin, ch, val); setChanged(); }
	public void scriptTpulseX0TypMax(int ch, double val){ setChannelArray(tpulseX0TypMax, ch, val); setChanged(); }
	public void scriptTpulseY0Min(int ch, double val){ setChannelArray(tpulseY0Min, ch, val); setChanged(); }
	public void scriptTpulseY0Max(int ch, double val){ setChannelArray(tpulseY0Max, ch, val); setChanged(); }
	public void scriptTpulseY0Value(int ch, double val){ setChannelArray(tpulseY0Value, ch, val); setChanged(); }
	public void scriptTpulseY0TypMin(int ch, double val){ setChannelArray(tpulseY0TypMin, ch, val); setChanged(); }
	public void scriptTpulseY0TypMax(int ch, double val){ setChannelArray(tpulseY0TypMax, ch, val); setChanged(); }
	public void scriptTpulseAMin(int ch, double val){ setChannelArray(tpulseAMin, ch, val); setChanged(); }
	public void scriptTpulseAMax(int ch, double val){ setChannelArray(tpulseAMax, ch, val); setChanged(); }
	public void scriptTpulseAValue(int ch, double val){ setChannelArray(tpulseAValue, ch, val); setChanged(); }
	public void scriptTpulseATypMin(int ch, double val){ setChannelArray(tpulseATypMin, ch, val); setChanged(); }
	public void scriptTpulseATypMax(int ch, double val){ setChannelArray(tpulseATypMax, ch, val); setChanged(); }
	public void scriptSwitchOnX0Min(int ch, double val){ setChannelArray(switchOnX0Min, ch, val); setChanged(); }
	public void scriptSwitchOnX0Max(int ch, double val){ setChannelArray(switchOnX0Max, ch, val); setChanged(); }
	public void scriptSwitchOnX0Value(int ch, double val){ setChannelArray(switchOnX0Value, ch, val); setChanged(); }
	public void scriptSwitchOnX0TypMin(int ch, double val){ setChannelArray(switchOnX0TypMin, ch, val); setChanged(); }
	public void scriptSwitchOnX0TypMax(int ch, double val){ setChannelArray(switchOnX0TypMax, ch, val); setChanged(); }
	public void scriptSwitchOnY0Min(int ch, double val){ setChannelArray(switchOnY0Min, ch, val); setChanged(); }
	public void scriptSwitchOnY0Max(int ch, double val){ setChannelArray(switchOnY0Max, ch, val); setChanged(); }
	public void scriptSwitchOnY0Value(int ch, double val){ setChannelArray(switchOnY0Value, ch, val); setChanged(); }
	public void scriptSwitchOnY0TypMin(int ch, double val){ setChannelArray(switchOnY0TypMin, ch, val); setChanged(); }
	public void scriptSwitchOnY0TypMax(int ch, double val){ setChannelArray(switchOnY0TypMax, ch, val); setChanged(); }
	public void scriptSwitchOnAMin(int ch, double val){ setChannelArray(switchOnAMin, ch, val); setChanged(); }
	public void scriptSwitchOnAMax(int ch, double val){ setChannelArray(switchOnAMax, ch, val); setChanged(); }
	public void scriptSwitchOnAValue(int ch, double val){ setChannelArray(switchOnAValue, ch, val); setChanged(); }
	public void scriptSwitchOnATypMin(int ch, double val){ setChannelArray(switchOnATypMin, ch, val); setChanged(); }
	public void scriptSwitchOnATypMax(int ch, double val){ setChannelArray(switchOnATypMax, ch, val); setChanged(); }
	public void scriptAmbientLevelMin(int ch, double val){ setChannelArray(ambientLevelMin, ch, val); setChanged(); }
	public void scriptAmbientLevelMax(int ch, double val){ setChannelArray(ambientLevelMax, ch, val); setChanged(); }
	public void scriptAmbientLevelValue(int ch, double val){ setChannelArray(ambientLevelValue, ch, val); setChanged(); }
	public void scriptAmbientLevelTypMin(int ch, double val){ setChannelArray(ambientLevelTypMin, ch, val); setChanged(); }
	public void scriptAmbientLevelTypMax(int ch, double val){ setChannelArray(ambientLevelTypMax, ch, val); setChanged(); }
	public void scriptAmbientBeginMin(int ch, double val){ setChannelArray(ambientBeginMin, ch, val); setChanged(); }
	public void scriptAmbientBeginMax(int ch, double val){ setChannelArray(ambientBeginMax, ch, val); setChanged(); }
	public void scriptAmbientBeginValue(int ch, double val){ setChannelArray(ambientBeginValue, ch, val); setChanged(); }
	public void scriptAmbientBeginTypMin(int ch, double val){ setChannelArray(ambientBeginTypMin, ch, val); setChanged(); }
	public void scriptAmbientBeginTypMax(int ch, double val){ setChannelArray(ambientBeginTypMax, ch, val); setChanged(); }
	public void scriptBackwallX0Min(int ch, double val){ setChannelArray(backwallX0Min, ch, val); setChanged(); }
	public void scriptBackwallX0Max(int ch, double val){ setChannelArray(backwallX0Max, ch, val); setChanged(); }
	public void scriptBackwallX0Value(int ch, double val){ setChannelArray(backwallX0Value, ch, val); setChanged(); }
	public void scriptBackwallX0TypMin(int ch, double val){ setChannelArray(backwallX0TypMin, ch, val); setChanged(); }
	public void scriptBackwallX0TypMax(int ch, double val){ setChannelArray(backwallX0TypMax, ch, val); setChanged(); }
	public void scriptBackwallAMin(int ch, double val){ setChannelArray(backwallAMin, ch, val); setChanged(); }
	public void scriptBackwallAMax(int ch, double val){ setChannelArray(backwallAMax, ch, val); setChanged(); }
	public void scriptBackwallAValue(int ch, double val){ setChannelArray(backwallAValue, ch, val); setChanged(); }
	public void scriptBackwallATypMin(int ch, double val){ setChannelArray(backwallATypMin, ch, val); setChanged(); }
	public void scriptBackwallATypMax(int ch, double val){ setChannelArray(backwallATypMax, ch, val); setChanged(); }
	public void scriptBackwallScaleLengthMin(int ch, double val){ setChannelArray(backwallScaleLengthMin, ch, val); setChanged(); }
	public void scriptBackwallScaleLengthMax(int ch, double val){ setChannelArray(backwallScaleLengthMax, ch, val); setChanged(); }
	public void scriptBackwallScaleLengthValue(int ch, double val){ setChannelArray(backwallScaleLengthValue, ch, val); setChanged(); }
	public void scriptBackwallScaleLengthTypMin(int ch, double val){ setChannelArray(backwallScaleLengthTypMin, ch, val); setChanged(); }
	public void scriptBackwallScaleLengthTypMax(int ch, double val){ setChannelArray(backwallScaleLengthTypMax, ch, val); setChanged(); }
	public void scriptRelativeSensitivitiesMean(int ch, double val){ setChannelArray(relativeSensitivitiesMean, ch, val); setChanged(); }
	public void scriptRelativeSensitivitiesSigma(int ch, double val){ setChannelArray(relativeSensitivitiesSigma, ch, val); setChanged(); }
	public void scriptRelativeSensitivitiesValue(int ch, double val){ setChannelArray(relativeSensitivitiesValue, ch, val); setChanged(); }
	public void scriptRelativeSensitivitiesMin(int ch, double val){ setChannelArray(relativeSensitivitiesMin, ch, val); setChanged(); }
	public void scriptRelativeSensitivitiesMax(int ch, double val){ setChannelArray(relativeSensitivitiesMax, ch, val); setChanged(); }
	public void scriptRelativeSensitivitiesTypMin(int ch, double val){ setChannelArray(relativeSensitivitiesTypMin, ch, val); setChanged(); }
	public void scriptRelativeSensitivitiesTypMax(int ch, double val){ setChannelArray(relativeSensitivitiesTypMax, ch, val); setChanged(); }
	public void scriptTpulseToBackwallTimeMean(int ch, double val){ setChannelArray(tpulseToBackwallTimeMean, ch, val); setChanged(); }
	public void scriptTpulseToBackwallTimeSigma(int ch, double val){ update(); setChannelArray(tpulseToBackwallTimeSigma, ch, val); setChanged(); }
	public void scriptTpulseToBackwallTimeValue(int ch, double val){ setChannelArray(tpulseToBackwallTimeValue, ch, val); setChanged(); }
	public void scriptTpulseToBackwallTimeMin(int ch, double val){ setChannelArray(tpulseToBackwallTimeMin, ch, val); setChanged(); }
	public void scriptTpulseToBackwallTimeMax(int ch, double val){ setChannelArray(tpulseToBackwallTimeMax, ch, val); setChanged(); }
	public void scriptTpulseToBackwallTimeTypMin(int ch, double val){ setChannelArray(tpulseToBackwallTimeTypMin, ch, val); setChanged(); }
	public void scriptTpulseToBackwallTimeTypMax(int ch, double val){ setChannelArray(tpulseToBackwallTimeTypMax, ch, val); setChanged(); }
	public void scriptTimingAdjustMean(double val){ timingAdjustMean = val; setChanged(); };
	public void scriptTimingAdjustSigma(double val){ timingAdjustSigma = val; setChanged(); };
	public void scriptTimingAdjustValue(double val){ timingAdjustValue = val; setChanged(); };
	public void scriptTimingAdjustMin(double val){ timingAdjustMin = val; setChanged(); };
	public void scriptTimingAdjustMax(double val){ timingAdjustMax = val; setChanged(); };
	public void scriptTimingAdjustTypMin(double val){ timingAdjustTypMin = val; setChanged(); };
	public void scriptTimingAdjustTypMax(double val){ timingAdjustTypMax = val; setChanged(); };
	public void scriptLaserWidthAdjustMean(double val){ laserWidthAdjustMean = val; setChanged(); };
	public void scriptLaserWidthAdjustSigma(double val){ laserWidthAdjustSigma = val; setChanged(); };
	public void scriptLaserWidthAdjustValue(double val){ laserWidthAdjustValue = val; setChanged(); };
	public void scriptLaserWidthAdjustMin(double val){ laserWidthAdjustMin = val; setChanged(); };
	public void scriptLaserWidthAdjustMax(double val){ laserWidthAdjustMax = val; setChanged(); };
	public void scriptLaserWidthAdjustTypMin(double val){ laserWidthAdjustTypMin = val; setChanged(); };
	public void scriptLaserWidthAdjustTypMax(double val){ laserWidthAdjustTypMax = val; setChanged(); };
	public void scriptUberConstMean(double val){ uberConstMean = val; setChanged(); }
	public void scriptUberConstSigma(double val){ uberConstSigma = val; setChanged(); }
	public void scriptUberConstValue(double val){ uberConstValue = val; setChanged(); }
	public void scriptUberConstMin(double val){ uberConstMin = val; setChanged(); }
	public void scriptUberConstMax(double val){ uberConstMax = val; setChanged(); }
	public void scriptUberConstTypMin(double val){ uberConstTypMin = val; setChanged(); }
	public void scriptUberConstTypMax(double val){ uberConstTypMax = val; setChanged(); }	
	public void scriptVignetPivotR(double val){ vignetPivotR = val; setChanged(); }	
	public void scriptVignetShiftMean(double val){ vignetShiftMean = val; setChanged(); }
	public void scriptVignetShiftSigma(double val){ vignetShiftSigma = val; setChanged(); }
	public void scriptVignetShiftValue(double val){ vignetShiftValue = val; setChanged(); }
	public void scriptVignetShiftMin(double val){ vignetShiftMin = val; setChanged(); }
	public void scriptVignetShiftMax(double val){ vignetShiftMax = val; setChanged(); }
	public void scriptVignetShiftTypMin(double val){ vignetShiftTypMin = val; setChanged(); }
	public void scriptVignetShiftTypMax(double val){ vignetShiftTypMax = val; setChanged(); }	
	public void scriptVignetAdjustGradMean(double val){ vignetAdjustGradMean = val; setChanged(); }
	public void scriptVignetAdjustGradSigma(double val){ vignetAdjustGradSigma = val; setChanged(); }
	public void scriptVignetAdjustGradValue(double val){ vignetAdjustGradValue = val; setChanged(); }
	public void scriptVignetAdjustGradMin(double val){ vignetAdjustGradMin = val; setChanged(); }
	public void scriptVignetAdjustGradMax(double val){ vignetAdjustGradMax = val; setChanged(); }
	public void scriptVignetAdjustGradTypMin(double val){ vignetAdjustGradTypMin = val; setChanged(); }
	public void scriptVignetAdjustGradTypMax(double val){ vignetAdjustGradTypMax = val; setChanged(); }
	
	/** ************* Direct Config Setters ******************* */
	public void setBaselineShapeDataX(double[][] baselineShapeDataX){ this.baselineShapeDataX = baselineShapeDataX; setChanged(); }
	public void setBaselineShapeDataY(double[][] baselineShapeDataY){ this.baselineShapeDataY = baselineShapeDataY; setChanged(); }
	public void setTpulseShapeDataX(double[][] tpulseShapeDataX){ this.tpulseShapeDataX = tpulseShapeDataX; setChanged(); }
	public void setTpulseShapeDataY(double[][] tpulseShapeDataY){ this.tpulseShapeDataY = tpulseShapeDataY; setChanged(); }
	public void setSwitchOnShapeDataX(double[][] switchOnShapeDataX){ this.switchOnShapeDataX = switchOnShapeDataX; setChanged(); }
	public void setSwitchOnShapeDataY(double[][] switchOnShapeDataY){ this.switchOnShapeDataY = switchOnShapeDataY; setChanged(); }
	public void setBaselineX0Min(double[] baselineX0Min){ this.baselineX0Min = baselineX0Min; setChanged(); }
	public void setBaselineX0Max(double[] baselineX0Max){ this.baselineX0Max = baselineX0Max; setChanged(); }
	public void setBaselineX0Value(double[] baselineX0Value){ this.baselineX0Value = baselineX0Value; setChanged(); }
	public void setBaselineX0TypMin(double[] baselineX0TypMin){ this.baselineX0TypMin = baselineX0TypMin; setChanged(); }
	public void setBaselineX0TypMax(double[] baselineX0TypMax){ this.baselineX0TypMax = baselineX0TypMax; setChanged(); }
	public void setBaselineY0Min(double[] baselineY0Min){ this.baselineY0Min = baselineY0Min; setChanged(); }
	public void setBaselineY0Max(double[] baselineY0Max){ this.baselineY0Max = baselineY0Max; setChanged(); }
	public void setBaselineY0Value(double[] baselineY0Value){ this.baselineY0Value = baselineY0Value; setChanged(); }
	public void setBaselineY0TypMin(double[] baselineY0TypMin){ this.baselineY0TypMin = baselineY0TypMin; setChanged(); }
	public void setBaselineY0TypMax(double[] baselineY0TypMax){ this.baselineY0TypMax = baselineY0TypMax; setChanged(); }
	public void setBaselineAMin(double[] baselineAMin){ this.baselineAMin = baselineAMin; setChanged(); }
	public void setBaselineAMax(double[] baselineAMax){ this.baselineAMax = baselineAMax; setChanged(); }
	public void setBaselineAValue(double[] baselineAValue){ this.baselineAValue = baselineAValue; setChanged(); }
	public void setBaselineATypMin(double[] baselineATypMin){ this.baselineATypMin = baselineATypMin; setChanged(); }
	public void setBaselineATypMax(double[] baselineATypMax){ this.baselineATypMax = baselineATypMax; setChanged(); }
	public void setTpulseX0Min(double[] tpulseX0Min){ this.tpulseX0Min = tpulseX0Min; setChanged(); }
	public void setTpulseX0Max(double[] tpulseX0Max){ this.tpulseX0Max = tpulseX0Max; setChanged(); }
	public void setTpulseX0Value(double[] tpulseX0Value){ this.tpulseX0Value = tpulseX0Value; setChanged(); }
	public void setTpulseX0TypMin(double[] tpulseX0TypMin){ this.tpulseX0TypMin = tpulseX0TypMin; setChanged(); }
	public void setTpulseX0TypMax(double[] tpulseX0TypMax){ this.tpulseX0TypMax = tpulseX0TypMax; setChanged(); }
	public void setTpulseY0Min(double[] tpulseY0Min){ this.tpulseY0Min = tpulseY0Min; setChanged(); }
	public void setTpulseY0Max(double[] tpulseY0Max){ this.tpulseY0Max = tpulseY0Max; setChanged(); }
	public void setTpulseY0Value(double[] tpulseY0Value){ this.tpulseY0Value = tpulseY0Value; setChanged(); }
	public void setTpulseY0TypMin(double[] tpulseY0TypMin){ this.tpulseY0TypMin = tpulseY0TypMin; setChanged(); }
	public void setTpulseY0TypMax(double[] tpulseY0TypMax){ this.tpulseY0TypMax = tpulseY0TypMax; setChanged(); }
	public void setTpulseAMin(double[] tpulseAMin){ this.tpulseAMin = tpulseAMin; setChanged(); }
	public void setTpulseAMax(double[] tpulseAMax){ this.tpulseAMax = tpulseAMax; setChanged(); }
	public void setTpulseAValue(double[] tpulseAValue){ this.tpulseAValue = tpulseAValue; setChanged(); }
	public void setTpulseATypMin(double[] tpulseATypMin){ this.tpulseATypMin = tpulseATypMin; setChanged(); }
	public void setTpulseATypMax(double[] tpulseATypMax){ this.tpulseATypMax = tpulseATypMax; setChanged(); }
	public void setSwitchOnX0Min(double[] switchOnX0Min){ this.switchOnX0Min = switchOnX0Min; setChanged(); }
	public void setSwitchOnX0Max(double[] switchOnX0Max){ this.switchOnX0Max = switchOnX0Max; setChanged(); }
	public void setSwitchOnX0Value(double[] switchOnX0Value){ this.switchOnX0Value = switchOnX0Value; setChanged(); }
	public void setSwitchOnX0TypMin(double[] switchOnX0TypMin){ this.switchOnX0TypMin = switchOnX0TypMin; setChanged(); }
	public void setSwitchOnX0TypMax(double[] switchOnX0TypMax){ this.switchOnX0TypMax = switchOnX0TypMax; setChanged(); }
	public void setSwitchOnY0Min(double[] switchOnY0Min){ this.switchOnY0Min = switchOnY0Min; setChanged(); }
	public void setSwitchOnY0Max(double[] switchOnY0Max){ this.switchOnY0Max = switchOnY0Max; setChanged(); }
	public void setSwitchOnY0Value(double[] switchOnY0Value){ this.switchOnY0Value = switchOnY0Value; setChanged(); }
	public void setSwitchOnY0TypMin(double[] switchOnY0TypMin){ this.switchOnY0TypMin = switchOnY0TypMin; setChanged(); }
	public void setSwitchOnY0TypMax(double[] switchOnY0TypMax){ this.switchOnY0TypMax = switchOnY0TypMax; setChanged(); }
	public void setSwitchOnAMin(double[] switchOnAMin){ this.switchOnAMin = switchOnAMin; setChanged(); }
	public void setSwitchOnAMax(double[] switchOnAMax){ this.switchOnAMax = switchOnAMax; setChanged(); }
	public void setSwitchOnAValue(double[] switchOnAValue){ this.switchOnAValue = switchOnAValue; setChanged(); }
	public void setSwitchOnATypMin(double[] switchOnATypMin){ this.switchOnATypMin = switchOnATypMin; setChanged(); }
	public void setSwitchOnATypMax(double[] switchOnATypMax){ this.switchOnATypMax = switchOnATypMax; setChanged(); }
	public void setAmbientLevelMin(double[] ambientLevelMin){ this.ambientLevelMin = ambientLevelMin; setChanged(); }
	public void setAmbientLevelMax(double[] ambientLevelMax){ this.ambientLevelMax = ambientLevelMax; setChanged(); }
	public void setAmbientLevelValue(double[] ambientLevelValue){ this.ambientLevelValue = ambientLevelValue; setChanged(); }
	public void setAmbientLevelTypMin(double[] ambientLevelTypMin){ this.ambientLevelTypMin = ambientLevelTypMin; setChanged(); }
	public void setAmbientLevelTypMax(double[] ambientLevelTypMax){ this.ambientLevelTypMax = ambientLevelTypMax; setChanged(); }
	public void setAmbientBeginMin(double[] ambientBeginMin){ this.ambientBeginMin = ambientBeginMin; setChanged(); }
	public void setAmbientBeginMax(double[] ambientBeginMax){ this.ambientBeginMax = ambientBeginMax; setChanged(); }
	public void setAmbientBeginValue(double[] ambientBeginValue){ this.ambientBeginValue = ambientBeginValue; setChanged(); }
	public void setAmbientBeginTypMin(double[] ambientBeginTypMin){ this.ambientBeginTypMin = ambientBeginTypMin; setChanged(); }
	public void setAmbientBeginTypMax(double[] ambientBeginTypMax){ this.ambientBeginTypMax = ambientBeginTypMax; setChanged(); }
	public void setBackwallX0Min(double[] backwallX0Min){ this.backwallX0Min = backwallX0Min; setChanged(); }
	public void setBackwallX0Max(double[] backwallX0Max){ this.backwallX0Max = backwallX0Max; setChanged(); }
	public void setBackwallX0Value(double[] backwallX0Value){ this.backwallX0Value = backwallX0Value; setChanged(); }
	public void setBackwallX0TypMin(double[] backwallX0TypMin){ this.backwallX0TypMin = backwallX0TypMin; setChanged(); }
	public void setBackwallX0TypMax(double[] backwallX0TypMax){ this.backwallX0TypMax = backwallX0TypMax; setChanged(); }
	public void setBackwallAMin(double[] backwallAMin){ this.backwallAMin = backwallAMin; setChanged(); }
	public void setBackwallAMax(double[] backwallAMax){ this.backwallAMax = backwallAMax; setChanged(); }
	public void setBackwallAValue(double[] backwallAValue){ this.backwallAValue = backwallAValue; setChanged(); }
	public void setBackwallATypMin(double[] backwallATypMin){ this.backwallATypMin = backwallATypMin; setChanged(); }
	public void setBackwallATypMax(double[] backwallATypMax){ this.backwallATypMax = backwallATypMax; setChanged(); }
	public void setBackwallScaleLengthMin(double[] backwallScaleLengthMin){ this.backwallScaleLengthMin = backwallScaleLengthMin; setChanged(); }
	public void setBackwallScaleLengthMax(double[] backwallScaleLengthMax){ this.backwallScaleLengthMax = backwallScaleLengthMax; setChanged(); }
	public void setBackwallScaleLengthValue(double[] backwallScaleLengthValue){ this.backwallScaleLengthValue = backwallScaleLengthValue; setChanged(); }
	public void setBackwallScaleLengthTypMin(double[] backwallScaleLengthTypMin){ this.backwallScaleLengthTypMin = backwallScaleLengthTypMin; setChanged(); }
	public void setBackwallScaleLengthTypMax(double[] backwallScaleLengthTypMax){ this.backwallScaleLengthTypMax = backwallScaleLengthTypMax; setChanged(); }
	public void setUberConstMean(double uberConstMean){ this.uberConstMean = uberConstMean; setChanged(); }
	public void setUberConstSigma(double uberConstSigma){ this.uberConstSigma = uberConstSigma; setChanged(); }
	public void setUberConstValue(double uberConstValue){ this.uberConstValue = uberConstValue; setChanged(); }
	public void setUberConstMin(double uberConstMin){ this.uberConstMin = uberConstMin; setChanged(); }
	public void setUberConstMax(double uberConstMax){ this.uberConstMax = uberConstMax; setChanged(); }
	public void setUberConstTypMin(double uberConstTypMin){ this.uberConstTypMin = uberConstTypMin; setChanged(); }
	public void setUberConstTypMax(double uberConstTypMax){ this.uberConstTypMax = uberConstTypMax; setChanged(); }
	public void setElecSigma(double elecSigma){ this.elecSigma = elecSigma; setChanged(); }
	public void setWindowTransmission(double[] windowTransmission){ this.windowTransmission = windowTransmission; setChanged(); }
	public void setVignettingProfileX(double[] vignettingProfileX){ this.vignettingProfileX = vignettingProfileX; setChanged(); }
	public void setVignettingProfileY(double[] vignettingProfileY){ this.vignettingProfileY = vignettingProfileY; setChanged(); }
	public void setRelativeSensitivitiesMean(double[] relativeSensitivitiesMean){ this.relativeSensitivitiesMean = relativeSensitivitiesMean; setChanged(); }
	public void setRelativeSensitivitiesSigma(double[] relativeSensitivitiesSigma){ this.relativeSensitivitiesSigma = relativeSensitivitiesSigma; setChanged(); }
	public void setRelativeSensitivitiesValue(double[] relativeSensitivitiesValue){ this.relativeSensitivitiesValue = relativeSensitivitiesValue; setChanged(); }
	public void setRelativeSensitivitiesMin(double[] relativeSensitivitiesMin){ this.relativeSensitivitiesMin = relativeSensitivitiesMin; setChanged(); }
	public void setRelativeSensitivitiesMax(double[] relativeSensitivitiesMax){ this.relativeSensitivitiesMax = relativeSensitivitiesMax; setChanged(); }
	public void setRelativeSensitivitiesTypMin(double[] relativeSensitivitiesTypMin){ this.relativeSensitivitiesTypMin = relativeSensitivitiesTypMin; setChanged(); }
	public void setRelativeSensitivitiesTypMax(double[] relativeSensitivitiesTypMax){ this.relativeSensitivitiesTypMax = relativeSensitivitiesTypMax; setChanged(); }	
	public void setTimingAdjustMean(double timingAdjustMean){ this.timingAdjustMean = timingAdjustMean; setChanged(); }
	public void setTimingAdjustSigma(double timingAdjustSigma){ this.timingAdjustSigma = timingAdjustSigma; setChanged(); }
	public void setTimingAdjustValue(double timingAdjustValue){ this.timingAdjustValue = timingAdjustValue; setChanged(); }
	public void setTimingAdjustMin(double timingAdjustMin){ this.timingAdjustMin = timingAdjustMin; setChanged(); }
	public void setTimingAdjustMax(double timingAdjustMax){ this.timingAdjustMax = timingAdjustMax; setChanged(); }
	public void setTimingAdjustTypMin(double timingAdjustTypMin){ this.timingAdjustTypMin = timingAdjustTypMin; setChanged(); }
	public void setTimingAdjustTypMax(double timingAdjustTypMax){ this.timingAdjustTypMax = timingAdjustTypMax; setChanged(); }	
	public void setLaserWidthAdjustMean(double laserWidthAdjustMean){ this.laserWidthAdjustMean = laserWidthAdjustMean; setChanged(); }
	public void setLaserWidthAdjustSigma(double laserWidthAdjustSigma){ this.laserWidthAdjustSigma = laserWidthAdjustSigma; setChanged(); }
	public void setLaserWidthAdjustValue(double laserWidthAdjustValue){ this.laserWidthAdjustValue = laserWidthAdjustValue; setChanged(); }
	public void setLaserWidthAdjustMin(double laserWidthAdjustMin){ this.laserWidthAdjustMin = laserWidthAdjustMin; setChanged(); }
	public void setLaserWidthAdjustMax(double laserWidthAdjustMax){ this.laserWidthAdjustMax = laserWidthAdjustMax; setChanged(); }
	public void setLaserWidthAdjustTypMin(double laserWidthAdjustTypMin){ this.laserWidthAdjustTypMin = laserWidthAdjustTypMin; setChanged(); }
	public void setLaserWidthAdjustTypMax(double laserWidthAdjustTypMax){ this.laserWidthAdjustTypMax = laserWidthAdjustTypMax; setChanged(); }
	public void setTpulseToBackwallTimeMean(double[] tpulseToBackwallTimeMean){ this.tpulseToBackwallTimeMean = tpulseToBackwallTimeMean; setChanged(); }
	public void setTpulseToBackwallTimeSigma(double[] tpulseToBackwallTimeSigma){ this.tpulseToBackwallTimeSigma = tpulseToBackwallTimeSigma; setChanged(); }
	public void setTpulseToBackwallTimeValue(double[] tpulseToBackwallTimeValue){ this.tpulseToBackwallTimeValue = tpulseToBackwallTimeValue; setChanged(); }
	public void setTpulseToBackwallTimeMin(double[] tpulseToBackwallTimeMin){ this.tpulseToBackwallTimeMin = tpulseToBackwallTimeMin; setChanged(); }
	public void setTpulseToBackwallTimeMax(double[] tpulseToBackwallTimeMax){ this.tpulseToBackwallTimeMax = tpulseToBackwallTimeMax; setChanged(); }
	public void setTpulseToBackwallTimeTypMin(double[] tpulseToBackwallTimeTypMin){ this.tpulseToBackwallTimeTypMin = tpulseToBackwallTimeTypMin; setChanged(); }
	public void setTpulseToBackwallTimeTypMax(double[] tpulseToBackwallTimeTypMax){ this.tpulseToBackwallTimeTypMax = tpulseToBackwallTimeTypMax; setChanged(); }
	public void setPhotoElectronFactors(double[] photoElectronFactors){ this.photoElectronFactors = photoElectronFactors; setChanged(); }
	public void setLineOfSightCoords(double[][] lineOfSightCoords){ this.lineOfSightCoords = lineOfSightCoords; setChanged(); }
	public void setPlasmaRegion(int[][] plasmaRegion){ this.plasmaRegion = plasmaRegion; setChanged(); }
	public void setChannelEnable(boolean[] channelEnable){ this.channelEnable = channelEnable; setChanged(); }
	public void setVignetPivotR(double vignetPivotR){ this.vignetPivotR = vignetPivotR; setChanged(); }
	public void setVignetShiftMin(double vignetShiftMin){ this.vignetShiftMin = vignetShiftMin; setChanged(); }
	public void setVignetShiftMax(double vignetShiftMax){ this.vignetShiftMax = vignetShiftMax; setChanged(); }
	public void setVignetShiftMean(double vignetShiftMean){ this.vignetShiftMean = vignetShiftMean; setChanged(); }
	public void setVignetShiftSigma(double vignetShiftSigma){ this.vignetShiftSigma = vignetShiftSigma; setChanged(); }
	public void setVignetShiftValue(double vignetShiftValue){ this.vignetShiftValue = vignetShiftValue; setChanged(); }
	public void setVignetShiftTypMin(double vignetShiftTypMin){ this.vignetShiftTypMin = vignetShiftTypMin; setChanged(); }
	public void setVignetShiftTypMax(double vignetShiftTypMax){ this.vignetShiftTypMax = vignetShiftTypMax; setChanged(); }
	public void setVignetAdjustGradMean(double vignetAdjustGradMean){ this.vignetAdjustGradMean = vignetAdjustGradMean; setChanged(); }
	public void setVignetAdjustGradSigma(double vignetAdjustGradSigma){ this.vignetAdjustGradSigma = vignetAdjustGradSigma; setChanged(); }
	public void setVignetAdjustGradValue(double vignetAdjustGradValue){ this.vignetAdjustGradValue = vignetAdjustGradValue; setChanged(); }
	public void setVignetAdjustGradMin(double vignetAdjustGradMin){ this.vignetAdjustGradMin = vignetAdjustGradMin; setChanged(); }
	public void setVignetAdjustGradMax(double vignetAdjustGradMax){ this.vignetAdjustGradMax = vignetAdjustGradMax; setChanged(); }
	public void setVignetAdjustGradTypMin(double vignetAdjustGradTypMin){ this.vignetAdjustGradTypMin = vignetAdjustGradTypMin; setChanged(); }
	public void setVignetAdjustGradTypMax(double vignetAdjustGradTypMax){ this.vignetAdjustGradTypMax = vignetAdjustGradTypMax; setChanged(); }	
	
	
	/** ************* Current Config Getters ******************* */
	public double[][] getBaselineShapeDataX() { update(); return baselineShapeDataX; }
	public double[][] getBaselineShapeDataY() { update(); return baselineShapeDataY; }
	public double[][] getTpulseShapeDataX() { update(); return tpulseShapeDataX; }
	public double[][] getTpulseShapeDataY() { update(); return tpulseShapeDataY; }
	public double[][] getSwitchOnShapeDataX() { update(); return switchOnShapeDataX; }
	public double[][] getSwitchOnShapeDataY() { update(); return switchOnShapeDataY; }
	public double[] getBaselineX0Min() { update(); return baselineX0Min; }
	public double[] getBaselineX0Max() { update(); return baselineX0Max; }
	public double[] getBaselineX0Value() { update(); return baselineX0Value; }
	public double[] getBaselineX0TypMin() { update(); return baselineX0TypMin; }
	public double[] getBaselineX0TypMax() { update(); return baselineX0TypMax; }
	public double[] getBaselineY0Min() { update(); return baselineY0Min; }
	public double[] getBaselineY0Max() { update(); return baselineY0Max; }
	public double[] getBaselineY0Value() { update(); return baselineY0Value; }
	public double[] getBaselineY0TypMin() { update(); return baselineY0TypMin; }
	public double[] getBaselineY0TypMax() { update(); return baselineY0TypMax; }
	public double[] getBaselineAMin() { update(); return baselineAMin; }
	public double[] getBaselineAMax() { update(); return baselineAMax; }
	public double[] getBaselineAValue() { update(); return baselineAValue; }
	public double[] getBaselineATypMin() { update(); return baselineATypMin; }
	public double[] getBaselineATypMax() { update(); return baselineATypMax; }
	public double[] getTpulseX0Min() { update(); return tpulseX0Min; }
	public double[] getTpulseX0Max() { update(); return tpulseX0Max; }
	public double[] getTpulseX0Value() { update(); return tpulseX0Value; }
	public double[] getTpulseX0TypMin() { update(); return tpulseX0TypMin; }
	public double[] getTpulseX0TypMax() { update(); return tpulseX0TypMax; }
	public double[] getTpulseY0Min() { update(); return tpulseY0Min; }
	public double[] getTpulseY0Max() { update(); return tpulseY0Max; }
	public double[] getTpulseY0Value() { update(); return tpulseY0Value; }
	public double[] getTpulseY0TypMin() { update(); return tpulseY0TypMin; }
	public double[] getTpulseY0TypMax() { update(); return tpulseY0TypMax; }
	public double[] getTpulseAMin() { update(); return tpulseAMin; }
	public double[] getTpulseAMax() { update(); return tpulseAMax; }
	public double[] getTpulseAValue() { update(); return tpulseAValue; }
	public double[] getTpulseATypMin() { update(); return tpulseATypMin; }
	public double[] getTpulseATypMax() { update(); return tpulseATypMax; }
	public double[] getSwitchOnX0Min() { update(); return switchOnX0Min; }
	public double[] getSwitchOnX0Max() { update(); return switchOnX0Max; }
	public double[] getSwitchOnX0Value() { update(); return switchOnX0Value; }
	public double[] getSwitchOnX0TypMin() { update(); return switchOnX0TypMin; }
	public double[] getSwitchOnX0TypMax() { update(); return switchOnX0TypMax; }
	public double[] getSwitchOnY0Min() { update(); return switchOnY0Min; }
	public double[] getSwitchOnY0Max() { update(); return switchOnY0Max; }
	public double[] getSwitchOnY0Value() { update(); return switchOnY0Value; }
	public double[] getSwitchOnY0TypMin() { update(); return switchOnY0TypMin; }
	public double[] getSwitchOnY0TypMax() { update(); return switchOnY0TypMax; }
	public double[] getSwitchOnAMin() { update(); return switchOnAMin; }
	public double[] getSwitchOnAMax() { update(); return switchOnAMax; }
	public double[] getSwitchOnAValue() { update(); return switchOnAValue; }
	public double[] getSwitchOnATypMin() { update(); return switchOnATypMin; }
	public double[] getSwitchOnATypMax() { update(); return switchOnATypMax; }
	public double[] getAmbientLevelMin() { update(); return ambientLevelMin; }
	public double[] getAmbientLevelMax() { update(); return ambientLevelMax; }
	public double[] getAmbientLevelValue() { update(); return ambientLevelValue; }
	public double[] getAmbientLevelTypMin() { update(); return ambientLevelTypMin; }
	public double[] getAmbientLevelTypMax() { update(); return ambientLevelTypMax; }
	public double[] getAmbientBeginMin() { update(); return ambientBeginMin; }
	public double[] getAmbientBeginMax() { update(); return ambientBeginMax; }
	public double[] getAmbientBeginValue() { update(); return ambientBeginValue; }
	public double[] getAmbientBeginTypMin() { update(); return ambientBeginTypMin; }
	public double[] getAmbientBeginTypMax() { update(); return ambientBeginTypMax; }
	public double[] getBackwallX0Min() { update(); return backwallX0Min; }
	public double[] getBackwallX0Max() { update(); return backwallX0Max; }
	public double[] getBackwallX0Value() { update(); return backwallX0Value; }
	public double[] getBackwallX0TypMin() { update(); return backwallX0TypMin; }
	public double[] getBackwallX0TypMax() { update(); return backwallX0TypMax; }
	public double[] getBackwallAMin() { update(); return backwallAMin; }
	public double[] getBackwallAMax() { update(); return backwallAMax; }
	public double[] getBackwallAValue() { update(); return backwallAValue; }
	public double[] getBackwallATypMin() { update(); return backwallATypMin; }
	public double[] getBackwallATypMax() { update(); return backwallATypMax; }
	public double[] getBackwallScaleLengthMin() { update(); return backwallScaleLengthMin; }
	public double[] getBackwallScaleLengthMax() { update(); return backwallScaleLengthMax; }
	public double[] getBackwallScaleLengthValue() { update(); return backwallScaleLengthValue; }
	public double[] getBackwallScaleLengthTypMin() { update(); return backwallScaleLengthTypMin; }
	public double[] getBackwallScaleLengthTypMax() { update(); return backwallScaleLengthTypMax; }
	public double getUberConstMean() { update(); return uberConstMean; }
	public double getUberConstSigma() { update(); return uberConstSigma; }
	public double getUberConstValue() { update(); return uberConstValue; }
	public double getUberConstMin() { update(); return uberConstMin; }
	public double getUberConstMax() { update(); return uberConstMax; }
	public double getUberConstTypMin() { update(); return uberConstTypMin; }
	public double getUberConstTypMax() { update(); return uberConstTypMax; }
	public double getElecSigma() { update(); return elecSigma; }
	public double[] getWindowTransmission() { update(); return windowTransmission; }
	public double[] getVignettingProfileX() { update(); return vignettingProfileX; }
	public double[] getVignettingProfileY() { update(); return vignettingProfileY; }
	public double[] getRelativeSensitivitiesMean() { update(); return relativeSensitivitiesMean; }
	public double[] getRelativeSensitivitiesSigma() { update(); return relativeSensitivitiesSigma; }
	public double[] getRelativeSensitivitiesValue() { update(); return relativeSensitivitiesValue; }
	public double[] getRelativeSensitivitiesMin() { update(); return relativeSensitivitiesMin; }
	public double[] getRelativeSensitivitiesMax() { update(); return relativeSensitivitiesMax; }
	public double[] getRelativeSensitivitiesTypMin() { update(); return relativeSensitivitiesTypMin; }
	public double[] getRelativeSensitivitiesTypMax() { update(); return relativeSensitivitiesTypMax; }
	public double getTimingAdjustMean() { update(); return timingAdjustMean; }
	public double getTimingAdjustSigma() { update(); return timingAdjustSigma; }
	public double getTimingAdjustValue() { update(); return timingAdjustValue; }
	public double getTimingAdjustMin() { update(); return timingAdjustMin; }
	public double getTimingAdjustMax() { update(); return timingAdjustMax; }
	public double getTimingAdjustTypMin() { update(); return timingAdjustTypMin; }
	public double getTimingAdjustTypMax() { update(); return timingAdjustTypMax; }
	public double getLaserWidthAdjustMean() { update(); return laserWidthAdjustMean; }
	public double getLaserWidthAdjustSigma() { update(); return laserWidthAdjustSigma; }
	public double getLaserWidthAdjustValue() { update(); return laserWidthAdjustValue; }
	public double getLaserWidthAdjustMin() { update(); return laserWidthAdjustMin; }
	public double getLaserWidthAdjustMax() { update(); return laserWidthAdjustMax; }
	public double getLaserWidthAdjustTypMin() { update(); return laserWidthAdjustTypMin; }
	public double getLaserWidthAdjustTypMax() { update(); return laserWidthAdjustTypMax; }	
	public double[] getTpulseToBackwallTimeMean() { update(); return tpulseToBackwallTimeMean; }
	public double[] getTpulseToBackwallTimeSigma() { update(); return tpulseToBackwallTimeSigma; }
	public double[] getTpulseToBackwallTimeValue() { update(); return tpulseToBackwallTimeValue; }
	public double[] getTpulseToBackwallTimeMin() { update(); return tpulseToBackwallTimeMin; }
	public double[] getTpulseToBackwallTimeMax() { update(); return tpulseToBackwallTimeMax; }
	public double[] getTpulseToBackwallTimeTypMin() { update(); return tpulseToBackwallTimeTypMin; }
	public double[] getTpulseToBackwallTimeTypMax() { update(); return tpulseToBackwallTimeTypMax; }
	public double[] getPhotoElectronFactors() { update(); return photoElectronFactors; }
	public double[][] getLineOfSightCoords() { update(); return lineOfSightCoords; }
	public int[][] getPlasmaRegion() { update(); return plasmaRegion; }
	public boolean[] getChannelEnable () { update(); return channelEnable ; }
	public double getVignetPivotR(){ update(); return vignetPivotR; }
	public double getVignetShiftMean(){ update(); return vignetShiftMean; }
	public double getVignetShiftSigma(){ update(); return vignetShiftSigma; }
	public double getVignetShiftValue(){ update(); return vignetShiftValue; }
	public double getVignetShiftMin(){ update(); return vignetShiftMin; }
	public double getVignetShiftMax(){ update(); return vignetShiftMax; }
	public double getVignetShiftTypMin(){ update(); return vignetShiftTypMin; }
	public double getVignetShiftTypMax(){ update(); return vignetShiftTypMax; }	
	public double getVignetAdjustGradMean(){ update(); return vignetAdjustGradMean; }
	public double getVignetAdjustGradSigma(){ update(); return vignetAdjustGradSigma; }
	public double getVignetAdjustGradValue(){ update(); return vignetAdjustGradValue; }
	public double getVignetAdjustGradMin(){ update(); return vignetAdjustGradMin; }
	public double getVignetAdjustGradMax(){ update(); return vignetAdjustGradMax; }
	public double getVignetAdjustGradTypMin(){ update(); return vignetAdjustGradTypMin; }
	public double getVignetAdjustGradTypMax(){ update(); return vignetAdjustGradTypMax; }

	public int[] getStrayLightChs(){
		update();
		int n = strayLightPeaks.size();
		int ch[] = new int[n];
		for(int i=0; i < n; i++)
			ch[i] = (int)strayLightPeaks.get(i)[0][0];
		return ch;
	}
	
	private double[] getStrayLightEntries(int row, int col){		
		int n = strayLightPeaks.size();
		double ret[] = new double[n];
		for(int i=0; i < n; i++)
			ret[i] = (int)strayLightPeaks.get(i)[row][col];
		return ret;
	}
	
	public double[] getStrayLightAValue(){ update(); return getStrayLightEntries(1, 0); }
	public double[] getStrayLightAMean(){ update(); return getStrayLightEntries(1, 1); }
	public double[] getStrayLightASigma(){ update(); return getStrayLightEntries(1, 2); }
	public double[] getStrayLightAMin(){ update(); return getStrayLightEntries(1, 3); }
	public double[] getStrayLightAMax(){ update(); return getStrayLightEntries(1, 4); }
	public double[] getStrayLightX0Value(){ update(); return getStrayLightEntries(2, 0); }
	public double[] getStrayLightX0Mean(){ update(); return getStrayLightEntries(2, 1); }
	public double[] getStrayLightX0Sigma(){ update(); return getStrayLightEntries(2, 2); }
	public double[] getStrayLightX0Min(){ update(); return getStrayLightEntries(2, 3); }
	public double[] getStrayLightX0Max(){ update(); return getStrayLightEntries(2, 4); }
	public double[] getStrayLightFWHMValue(){ update(); return getStrayLightEntries(3, 0); }
	public double[] getStrayLightFWHMMean(){ update(); return getStrayLightEntries(3, 1); }
	public double[] getStrayLightFWHMSigma(){ update(); return getStrayLightEntries(3, 2); }
	public double[] getStrayLightFWHMMin(){ update(); return getStrayLightEntries(3, 3); }
	public double[] getStrayLightFWHMMax(){ update(); return getStrayLightEntries(3, 4); }	
	
}
