/** This file is part of LIDAR Analysis work carried as part of a PhD with Imperial College.  
 * Copyright (c) 2010 Oliver Ford
 *
 * The content, concept and implementation are the intellectual property of Oliver Ford but permission
 * is given for it's use and further development by Jakob Svensson or any person or organisation
 * directly given such permissions by Oliver Ford or by Jakob Svensson.
 */
package seed.minerva.lidar.dataProvision;

/**
 * Provides nice interface for LIDAR signals and calibration data 
 * See KE3DataProvider or KE9DDataProvider
 *
 * @author oliford <codes@oliford.co.uk>
 *
 */

public interface LidarDataProvider {
	
	/** Things to pass to a data providers constructor to request things */
	public static final int LD_SIGNAL = 0x01;
	public static final int LD_SPECTRA = 0x02;
	public static final int LD_VIGNET = 0x04;
	public static final int LD_PROFILES = 0x08;
	public static final int LD_BASELINE = 0x10;
	public static final int LD_GAINS = 0x20;
	public static final int LD_TIMING = 0x40;
	public static final int LD_ENERGY_MON = 0x80;
	
	
	/** @return the pulse number the provider was initilaised with */
	public long getPulse();
	
	/** Sets the pulse number (reinits the provider) */
	public void setPulse(int pulse);
	
	/** @param frame Frame number (i.e laser pulse index)
	 * @return 2D array: array of data points of the specified frame: signal[channel number][]
	 */
	public double[][] signalFrame(int frame);
		
	/** @return Laser energy for all frames */
	public double[] laserEnergies();
	
	/** @return Laser energy for a specified frame */
	public double laserEnergy(int frame);
	
	/** @return signal for all channels in all frames: signal[frame][channel][] */
	public double[][][] signalFull();
	
	/** @return baseline for all channels in all frames */ 
	public double[][] baseLine();
	
	/** @return number of channels active. */
	public  int nChannels();
	
	/** @return number of frames (laser pulses) */
	public  int nFrames();
	
	/** @return Frame (laser pulses) time points */
	public  double[] frameTimes();
	
	/** @return vignetting[R / data][data point] */
	public  double[][] vignetting();
	
	/** @return the detector spectrums [ch][wave/amp][] - normalsied multiplied by relative sensitivity*/
	public  double[][][] spectrums();
	
	/** @return normalised detector spectrums [ch][wave/amp][] wavelengths in angstrom*/
	public  double[][][] normSpectrums();
	
	/** @return the relative sensitivities [ch]*/
	public  double[] relativeSensitivities();
	
	/** @return signal conversion  in mv per (photoelectron per nanosecond). */
	public  double[] gains();
	
	/** @return time between timing pulse and backwall in seconds */
	public   double[] timing();
	
	/** @return The ADC unit size in mV */
	public  double ADCUnit();
	
	/** @return The maximum value the signal can take in ADC Units (The saturation level) */
	public  double signalMax();
	
	/** @return length of signal data array */
	public  int signalLength();
	
	/** @return data's sampling period in seconds  */
	public  double signalDt();
	
	/** @return The level of the most severe waning issued:
	 * 0=No problems
	 * 1=Some data is unavaliable (PPF or Laser Energies) 
	 * 2=Some unusual values were found, data may be suspect.
	 * 3=Some values were not sensible , but these may or may not have an impact on output, 
	 * 4=Some very unusual values were found/things were not as expected, data is very likely to not be sensible.
	 * 5=Severe irregularlites were seen, outputs are probably complete rubbish. (the provider will most likely crash before finishing if a level 5 is issued)
	 */
	public  int warningLevel();
		
	public  double[] ppfFramene(int ppfFrame);
	public  double[] ppfFrameTe(int ppfFrame);
	public  double[] ppfR();
	public  double[] ppfZ();
	
	/** @return the frame index into the PPF (their frame number) that sits nearest in time to a frame in our arrays, these do not necceasrily line up exactly. */ 
	public  int pffNearestFrame(int frame);
	
	

}
