/** This file is part of LIDAR Analysis work carried as part of a PhD with Imperial College.  
 * Copyright (c) 2010 Oliver Ford
 *
 * The content, concept and implementation are the intellectual property of Oliver Ford but permission
 * is given for it's use and further development by Jakob Svensson or any person or organisation
 * directly given such permissions by Oliver Ford or by Jakob Svensson.
 */
package seed.minerva.lidar.dataProvision;

import seed.minerva.ConnectionPoint;
import seed.minerva.NodeImpl;
import seed.minerva.nodetypes.*;

/** Util node to take extended line of sight from KE9 data source and cut down to a more sensible length
 * @deprecated Dropped, replaced by code in LidarConfigSource 
 * */
public class LineOfSightSlicer extends NodeImpl implements DoubleMatrix {
	public static final String defaultName = "LineOfSightSlicer";
	
	//Parents
	/** Line of sight coords which are too long (e.g from DataSource) */
	DoubleMatrix lineOfSightCoords;
	/** Minimum R to allow LOS to go to */
	DoubleValue minimumR;
	/** Maximum R to allow LOS to go to */
	DoubleValue maximumR;
	
	public LineOfSightSlicer() { this(defaultName); }
	
	public LineOfSightSlicer(String name) {
		super(name);
		addConnectionPoint(new ConnectionPoint("lineOfSightCoords", DoubleMatrix.class, false, getField("lineOfSightCoords")));
		addConnectionPoint(new ConnectionPoint("minimumR", DoubleValue.class, true, getField("minimumR")));
		addConnectionPoint(new ConnectionPoint("maximumR", DoubleValue.class, true, getField("maximumR")));
	}

	@Override
	public double[][] getDoubleMatrix() {
		if(minimumR != null && maximumR != null && minimumR.getDouble() >= maximumR.getDouble())
			throw new IllegalArgumentException("minimumR >= maximumR ("+minimumR.getDouble()+" >= "+maximumR.getDouble()+").");
	
		double los[][] = lineOfSightCoords.getDoubleMatrix();
		double R0 = los[0][0], Z0 = los[1][0];		
		double R1 = los[0][1], Z1 = los[1][1];
		double Rbackwall = los[0][2], Zbackwall = los[1][2];

		if(minimumR != null){
			double minR = minimumR.getDouble();
			if(R0 < minR){
				if(R1 < minR)
					throw new IllegalArgumentException("Both ends of LOS R=("+R0+", "+R1+") are < minimumR ("+minR+").");
				Z0 += (minR - R0) * (Z1 - Z0) / (R1 - R0);
				R0 = minR;
			}else if(R1 < minR){
				Z1 += (minR - R1) * (Z0 - Z1) / (R0 - R1);
				R1 = minR;
			}
		}
		if(maximumR != null){
			double maxR = maximumR.getDouble();
			if(R0 > maxR){
				if(R1 > maxR)
					throw new IllegalArgumentException("Both ends of LOS R=("+R0+", "+R1+") are > maximumR ("+maxR+").");
				Z0 += (maxR - R0) * (Z1 - Z0) / (R1 - R0);
				R0 = maxR;
			}else if(R1 > maxR){
				Z1 += (maxR - R1) * (Z0 - Z1) / (R0 - R1);
				R1 = maxR;
			}
		}
		
		return new double[][]{ {R0, R1, Rbackwall }, {Z0, Z1, Zbackwall} };
	}

}
