/** This file is part of LIDAR Analysis work carried as part of a PhD with Imperial College.  
 * Copyright (c) 2010 Oliver Ford
 *
 * The content, concept and implementation are the intellectual property of Oliver Ford but permission
 * is given for it's use and further development by Jakob Svensson or any person or organisation
 * directly given such permissions by Oliver Ford or by Jakob Svensson.
 */
package seed.minerva.lidar.dataProvision;

import mds.JetMDSFetcher;
import oneLiners.DataConvertPureJava;
import oneLiners.OneLiners;
import seed.minerva.ConnectionPoint;
import seed.minerva.DataSource;
import seed.minerva.MinervaSettings;
import seed.minerva.StateFullNodeImpl;
import seed.minerva.diagnostics.ServiceManager;
import seed.minerva.nodetypes.DoubleValue;
import seed.minerva.nodetypes.IntegerValue;


/** Provides the Minerva data source for LIDAR systems. 
 * The interface LidarDataProvider remains independent of minerva as they are also used outside the project!
 * This is just a Minervication of that interface.
 */
public class MinervaLidarDataSource extends StateFullNodeImpl implements DataSource {
	public final static int DIAG_KE3_CORE_LIDAR = 1;
	public final static int DIAG_KE9_EDGE_LIDAR = 2;
	public final static String defaultNamePerDiag[] = { 
		"MinervaLidarDataSource-ERROR",  
		"MinervaLidarDataSource-KE3CoreLIDAR",
		"MinervaLidarDataSource-KE9EdgeLIDAR",
	};
	
	// Parents
	public IntegerValue pulse;
	public DoubleValue time;
	
	//state
	private LidarDataProvider dataProv;
	private int frame;
	private int diag;
	
	public MinervaLidarDataSource(int diag) { this(defaultNamePerDiag[diag],diag); }	
	public MinervaLidarDataSource(String name, int diag) {
		super(name);
		this.diag = diag;
		
		if(diag != DIAG_KE3_CORE_LIDAR && diag != DIAG_KE9_EDGE_LIDAR)
			throw new IllegalArgumentException("diag argument must be DIAG_KE3_CORE_LIDAR or DIAG_KE9_EDGE_LIDAR.");
		
		dataProv = null; //pick that up in updateState();
		
		addConnectionPoint(new ConnectionPoint("pulse", IntegerValue.class, false, getField("pulse")));
		addConnectionPoint(new ConnectionPoint("time", DoubleValue.class, false, getField("time")));
		
	}
	
	public int getDiag(){ return diag; }
	
	@Override
	public void updateState() {		
		if(dataProv == null){
			JetMDSFetcher jmds = ServiceManager.getInstance().getJETMDSFetcher();
			switch(diag){
			case DIAG_KE3_CORE_LIDAR: dataProv = new KE3DataProvider(jmds); break;
			case DIAG_KE9_EDGE_LIDAR: dataProv = new KE9DDataProvider(jmds); break;
			default:
				throw new IllegalArgumentException("diag argument must be DIAG_KE3_CORE_LIDAR or DIAG_KE9_EDGE_LIDAR.");
			}
		}
		
		if( isAncestorChanged("pulse") || dataProv.getPulse() <= 0 ){
			if(MinervaSettings.getDbgLevel() > 1)
				System.out.println("MinervaLidarDataSource moving to pulse " + pulse.getInteger() + "... ");		
			dataProv.setPulse(pulse.getInteger());
		}
		
		if( isAncestorChanged("pulse") || isAncestorChanged("time")){			
			frame = OneLiners.getNearestIndex(dataProv.frameTimes(), time.getDouble());
			if(MinervaSettings.getDbgLevel() > 1)
				System.out.println("MinervaLidarDataSource moving to time " + time.getDouble() +
						"s. Nearest is frame " + frame + " at " + dataProv.frameTimes()[frame] + "s.");
		}	
	}
	
	@Override
	public void tidyUpState() {
		super.tidyUpState();
		
		dataProv = null;
	}
	
	/** @returns 6940 - the laser wavelength in Angstrom (The JET lidar systems are Ruby) */
	public double laserWavelength(){ return 6940; }; 
	
	/** @returns The currently selected frame */
	public int getFrame(){ update(); return frame; }
	
	/** Gets the actual time of the current frame */ 
	public double frameTime(){ update(); return dataProv.frameTimes()[frame]; }

	/** Returns the entire actual data provider */
	public LidarDataProvider getDataProv(){ update(); return dataProv; }
	
	/** @return 2D array: array of data points in the current frame: signal[channel number][] */
	public double[][] getSignal(){ update(); return dataProv.signalFrame(frame); }
	
	/** @return 2D array: flat version of getSignal	 */
	public double[] getSignalSingleArray(){ update(); return DataConvertPureJava.doubleFlatten2DTo1D(dataProv.signalFrame(frame)); }
			
	/** @return Laser energy for a current frame */
	public double laserEnergy(){ update(); return dataProv.laserEnergy(frame); }
	
	/** @return baseline X vector  for all channels  */ 
	public double[][] baselineX(){ 
		update();
		double ret[][] = new double[dataProv.nChannels()][dataProv.signalLength()];
		for(int ch=0;ch<dataProv.nChannels();ch++)
			for(int i=0;i<dataProv.signalLength();i++)
				ret[ch][i] = i;
		return ret;
	}
	
	/** @return baseline ADCU amplitude for all channels  */ 
	public double[][] baselineY(){ update(); return dataProv.baseLine(); }
		
	/** @return The photo electron factors (ADCU per photoelectron per ADCU bin) */
	public double[] photoElectronFactors(){
		update();
		double peFacts[] = new double[dataProv.nChannels()];
		double gains[] = dataProv.gains();				// mv per (photoelectron per ns)
		double ADCUnit = dataProv.ADCUnit();			// mv per ADCU  
		double signalDt = dataProv.signalDt() / 1e-9;	// dt in ns
		
		for(int i=0;i<dataProv.nChannels();i++){
			peFacts[i] = gains[i] / ADCUnit / signalDt;
		}
		
		return peFacts;
	}
	
	/** @return number of channels active. */
	public int nChannels(){ update(); return dataProv.nChannels(); }
	
	/** @return number of channels active. */
	public int signalLength(){ update(); return dataProv.signalLength(); }
	
	/** @return vignetting profile major radial position [data point] */
	public  double[] vignettingR(){ update(); return dataProv.vignetting()[0]; }
	/** @return vignetting profile */
	public  double[] vignettingAmp(){ update(); return dataProv.vignetting()[1]; }
	
	/** @return normalised detector spectrums wavelengths [ch][] wavelengths in angstrom*/
	public  double[][] normalisedSpectrumWavelengths(){ return normSpectSlice(0); }
	
	/** @return normalised detector spectrums amplitudes [ch][] */
	public double[][] normalisedSpectrumAmplitudes(){ return normSpectSlice(1); }
	
	private double[][] normSpectSlice(int dim){
		update();
		double normSpectBoth[][][] = dataProv.normSpectrums();

		double ret[][] = new double[dataProv.nChannels()][];
		for(int ch=0; ch < dataProv.nChannels(); ch++){
			int n = normSpectBoth[ch][dim].length; 
			ret[ch] = new double[n];
			System.arraycopy(normSpectBoth[ch][dim], 0, ret[ch], 0, n);
		}
		return ret;
	}
	
	/** @return the relative sensitivities [ch]*/
	public  double[] relativeSensitivities(){ update(); return dataProv.relativeSensitivities(); }
	
	/** @return signal conversion  in mv per (photoelectron per nanosecond). */
	public  double[] gains(){ update(); return dataProv.gains(); }
	
	/** @return time between timing pulse and backwall in seconds */
	public   double[] timing(){ update(); return dataProv.timing(); }
	
	/** @return The ADC unit size in mV */
	public  double ADCUnit(){ update(); return dataProv.ADCUnit(); }
	
	/** @return The maximum value the signal can take in ADC Units (The saturation level) */
	public  double signalMax(){ update(); return dataProv.signalMax(); }
	
	/** @return data's sampling period in seconds  */
	public  double signalDt(){ update(); return dataProv.signalDt(); }
	
	
	public double[][] lineOfSightCoords(){
		update();
		
		if(!(dataProv instanceof KE9DDataProvider))
			throw new IllegalArgumentException("Can only get line of sight coords from a KE9DDataProvider");
		
		double defR0 = ((KE9DDataProvider)dataProv).getR0();
		double defR1 = ((KE9DDataProvider)dataProv).getR1();
		double defZ0 = ((KE9DDataProvider)dataProv).getZ0();
		double defZ1 = ((KE9DDataProvider)dataProv).getZ1(); //these define the laser line, but go way outside
		
		return new double[][]{ { defR0, defR1, defR1 }, { defZ0, defZ1, defZ1 }};
	}
		
	public double[] ppfR(){ update(); return dataProv.ppfR(); }
	public double[] ppfZ(){ update(); return dataProv.ppfZ(); }
	public double[] ppfNe(){ update(); return dataProv.ppfFramene(dataProv.pffNearestFrame(frame)); }
	public double[] ppfTe(){ update(); return dataProv.ppfFrameTe(dataProv.pffNearestFrame(frame)); }
	
	
}
