/** This file is part of LIDAR Analysis work carried as part of a PhD with Imperial College.  
 * Copyright (c) 2010 Oliver Ford
 *
 * The content, concept and implementation are the intellectual property of Oliver Ford but permission
 * is given for it's use and further development by Jakob Svensson or any person or organisation
 * directly given such permissions by Oliver Ford or by Jakob Svensson.
 */
package seed.minerva.lidar.model;

import oneLiners.DataConvertPureJava;
import seed.minerva.ConnectionPoint;
import seed.minerva.RandomManager;
import seed.minerva.StateFullNodeImpl;
import seed.minerva.nodetypes.BooleanArray;
import seed.minerva.nodetypes.DoubleArray;
import seed.minerva.nodetypes.DoubleMatrix;
import seed.minerva.nodetypes.DoubleValue;
import seed.minerva.nodetypes.IntegerMatrix;

/** Add together the plasma and non plasma photoElec and electronic components
 * and converts it all the ADCU.
 * 
 * @author oliford
 *
 */
public class ADCSignal extends StateFullNodeImpl {
	public final static String defaultName = "ADCSignal";
	
	/** Non plasma signal that contributes to photoelectron count */  
	private DoubleMatrix nonPlasmaPhotoElecsPerADCU, nonPlasmaPhotoElecsPerADCUSigma2;	
	/** Non plasma signal that contributes to ADCU value (i.e baseline) */	
	private DoubleMatrix elecBasline;
	/** plasma signal in photoElecs but on ADC timebase*/
	private DoubleMatrix plasmaPhotoElecsPerADCU, plasmaPhotoElecsPerADCUSigma2;
	
	/** Calibration constants - photoelectron factor/gain (per channel)
	 * This gives the ADC units recordered for a stream of 1 photoelectron every ADC unit bin */
	private DoubleArray photoElectronFactors;
	/** ADCU saturation level */
	private DoubleValue signalSaturationLevel;
	/** electronic sigma */
	private DoubleValue elecSigma;
	/** Region on ADC timebase for each channel of plasma data
	 *  which will be ignore if evalPlasmaRegion is false. */
	private IntegerMatrix plasmaRegion;
	/** Channel enable plags from config */
	private BooleanArray channelEnable;
	
	//Properties
	/** Flag to enable/disable plasma part of signal */ 
	private boolean evalPlasma = true;
	private double inflateSigma = 1; // I'm ... I'm sorry :(	
	/** For simulation purposes, add noise to photoelectron and/or electronic signals */
	private boolean addElecNoise = false;
	/** Reduce sigma to sigmaElec when the predicted signal is saturated.
	 * This stops the logP chaning with light level above the sturation level
	 * due to the sigma in the normalisation. However, it does make a discontinuity
	 * as the prediction meats the saturation level. To avoid this, turn this off
	 * and turn evalSaturatedData off as well.
	 */
	private boolean reduceSigmaForSaturatedPredictions = true;
	/** If set, don't evaluate points past this point. Set to saturation points 
	 * if reduceSigmaForSaturatedPredictions is off*/
	private double[] noEvalPast;
	
	public void setInflateSigma(double inflateSigma){ setChanged("inflateSigma"); this.inflateSigma = inflateSigma;	}
	
	public void setElecNoiseGeneration(boolean addElecNoise){ this.addElecNoise = addElecNoise; setChanged("addElecNoise");	}

	
 
	/** Evaluate plasma region (for fitting non-plasma components) */
	private boolean evalPlasmaRegion;	
	/** Evaluate non-plasma region (for when not fitting non-plasma components) */
	private boolean evalNonPlasmaRegion = true;
	/** Data point enables */
	private boolean dataPointEn[];

	//State
	private double signal[][], sigma2[][];

	public ADCSignal() { this(defaultName); }
	public ADCSignal(String name) {
		super(name);
		addConnectionPoint(new ConnectionPoint("nonPlasmaPhotoElecsPerADCU", DoubleMatrix.class, false, getField("nonPlasmaPhotoElecsPerADCU")));
		addConnectionPoint(new ConnectionPoint("nonPlasmaPhotoElecsPerADCUSigma2", DoubleMatrix.class, false, getField("nonPlasmaPhotoElecsPerADCUSigma2")));
		addConnectionPoint(new ConnectionPoint("elecBasline", DoubleMatrix.class, false, getField("elecBasline")));
		addConnectionPoint(new ConnectionPoint("plasmaPhotoElecsPerADCU", DoubleMatrix.class, false, getField("plasmaPhotoElecsPerADCU")));
		addConnectionPoint(new ConnectionPoint("plasmaPhotoElecsPerADCUSigma2", DoubleMatrix.class, false, getField("plasmaPhotoElecsPerADCUSigma2")));
		addConnectionPoint(new ConnectionPoint("photoElectronFactors", DoubleArray.class, false, getField("photoElectronFactors")));		
		addConnectionPoint(new ConnectionPoint("signalSaturationLevel", DoubleValue.class, false, getField("signalSaturationLevel")));
		addConnectionPoint(new ConnectionPoint("elecSigma", DoubleValue.class, false, getField("elecSigma")));
		addConnectionPoint(new ConnectionPoint("plasmaRegion", IntegerMatrix.class, false, getField("plasmaRegion")));
		addConnectionPoint(new ConnectionPoint("channelEnable", BooleanArray.class, false, getField("channelEnable")));
	}
	
	public double[][] getADCUPerADCU(){ update(); return signal; }
	public double[][] getADCUPerADCUSigma2(){ update(); return sigma2; }
	
	public double[] getADCUPerADCUSingleArray(){ update(); return DataConvertPureJava.doubleFlatten2DTo1D(signal); }
	public double[] getADCUPerADCUSigma2SingleArray(){ update(); return DataConvertPureJava.doubleFlatten2DTo1D(sigma2); }

	
	double oldBaseline;
	double oldNonPlasmaPE;
	double oldNonPlasmaPESigma2;
	double oldPeToADCU;
	double oldPlasmaPE;
	double oldPlasmaPESigma2;		
	double oldSigmaElec2;
	double oldSatLevel;
	        
	public void calcSignal(){
		double baseline[][] = elecBasline.getDoubleMatrix();
		double nonPlasmaPE[][] = nonPlasmaPhotoElecsPerADCU.getDoubleMatrix();
		double nonPlasmaPESigma2[][] = nonPlasmaPhotoElecsPerADCUSigma2.getDoubleMatrix();
		double peToADCU[] = photoElectronFactors.getDoubleArray();
		double sigmaElec2 = elecSigma.getDouble() * elecSigma.getDouble();
		double plasmaPE[][]=null, plasmaPESigma2[][]=null;
		int plRegion[][] = plasmaRegion.getIntegerMatrix();
		if(evalPlasma){			
			plasmaPE = plasmaPhotoElecsPerADCU.getDoubleMatrix();
			plasmaPESigma2 = plasmaPhotoElecsPerADCUSigma2.getDoubleMatrix();
		}
		double satLevel = signalSaturationLevel.getDouble();
		
		int nCh = baseline.length;
		int sigLen = baseline[0].length;
		
		signal = new double[nCh][sigLen];
		sigma2 = new double[nCh][sigLen];
		dataPointEn = new boolean[nCh * sigLen];
		
		RandomManager randGen = RandomManager.instance();
		
		for(int ch=0; ch < nCh; ch++){
			for(int i=0; i < sigLen; i++){
				//add together nonplasma and plasma photoelecs
				signal[ch][i] = nonPlasmaPE[ch][i];
				sigma2[ch][i] = nonPlasmaPESigma2[ch][i];
				
				if(evalPlasma){
					signal[ch][i] += plasmaPE[ch][i];
					sigma2[ch][i] += plasmaPESigma2[ch][i];
				}
				

				//convert those to ADCU and add electronic component
				signal[ch][i] = signal[ch][i] * peToADCU[ch] + baseline[ch][i];
				sigma2[ch][i] = sigma2[ch][i] * peToADCU[ch] * peToADCU[ch] + sigmaElec2;
				
				if(addElecNoise)
					signal[ch][i] += randGen.nextNormal(0, Math.sqrt(sigmaElec2));
								
				if(reduceSigmaForSaturatedPredictions && signal[ch][i] > satLevel){
					signal[ch][i] = satLevel;
					//make sure sigma doesn't still scale with the predicted signal
					sigma2[ch][i] = sigmaElec2;
				}
				
				//should the MVN eval this point?
				boolean inPlasmaRegion = (i >= plRegion[ch][0] && i <= plRegion[ch][1]);
				dataPointEn[ch*sigLen+i] = ( channelEnable.getBooleanArray()[ch] && (inPlasmaRegion ? evalPlasmaRegion : evalNonPlasmaRegion) );
				
				if(noEvalPast != null && i >= noEvalPast[ch])
					dataPointEn[ch*sigLen+i] = false;
	
				sigma2[ch][i] *= inflateSigma * inflateSigma; // TODO: Shoot yourself for this :(
				
				if(dataPointEn[ch*sigLen+i] && (Double.isNaN(signal[ch][i]) || Double.isNaN(sigma2[ch][i])) ){
					System.err.println(getPath() + ": WARNING: ADC Signal likelihood function is NaN. ch = "+ch+", i="+ i);
				}
					
			}
		}
	}
	
	@Override
	public void updateState() {
		if( signal == null ||
				isAncestorChanged("nonPlasmaPhotoElecsPerADCU") ||
				isAncestorChanged("nonPlasmaPhotoElecsPerADCUSigma2") ||
				isAncestorChanged("elecBasline") ||
				isAncestorChanged("plasmaPhotoElecsPerADCU") ||
				isAncestorChanged("plasmaPhotoElecsPerADCUSigma2") ||
				isAncestorChanged("photoElectronFactors") ||
				isAncestorChanged("signalSaturationLevel") ||
				isAncestorChanged("backwallX0") ||
				isAncestorChanged("elecSigma") 
				|| isPropertyChanged("evalPlasma")
				|| isPropertyChanged("evalPlasmaRegion")
				|| isPropertyChanged("evalNonPlasmaRegion")
				|| isPropertyChanged("evalChannel")
				|| isPropertyChanged("plasmaRegion") 
				|| isPropertyChanged("inflateSigma")
				|| isPropertyChanged("addElecNoise")
				|| isPropertyChanged("nonPlasmaNoise")
				|| isPropertyChanged("reduceSigmaForSaturatedPredictions")
				|| isPropertyChanged("noEvalPast"))
			calcSignal();
	}
	
	@Override
	public void tidyUpState() {
		super.tidyUpState();
		
		signal = null;
		sigma2 = null;
		dataPointEn = null;

	}
	
	public int[][] getPlasmaRegion() { return plasmaRegion.getIntegerMatrix(); }
	public boolean getEvalPlasmaRegion() { return evalPlasmaRegion; }
	public boolean[] getEvalChannel() { return channelEnable.getBooleanArray(); }
	public boolean getEvalPlasma() { update(); return evalPlasma; }
	
	public void setEvalPlasmaRegion(boolean evalPlasmaRegion) { this.evalPlasmaRegion = evalPlasmaRegion; setChanged("evalPlasmaRegion"); }
	public void setEvalNonPlasmaRegion(boolean evalNonPlasmaRegion) { this.evalNonPlasmaRegion = evalNonPlasmaRegion; setChanged("evalNonPlasmaRegion"); }
	public void setEvalPlasma(boolean evalPlasma) { this.evalPlasma = evalPlasma; setChanged("evalPlasma"); }
	
	public boolean[] getDataPointEvalEnable(){ update(); return dataPointEn; }
	
	/** Reduce sigma to sigmaElec when the predicted signal is saturated.
	 * This stops the logP changing with light level above the sturation level
	 * due to the sigma in the normalisation. However, it does make a discontinuity
	 * as the prediction meats the saturation level. To avoid this, turn this off
	 * and turn evalSaturatedData off as well.
	 */
	public void setReduceSigmaForSaturatedPredictions(boolean reduceSigmaForSaturatedPredictions){
		this.reduceSigmaForSaturatedPredictions = reduceSigmaForSaturatedPredictions;
		 setChanged("reduceSigmaForSaturatedPredictions");
	}
	
	/** If set, don't evaluate points past this point. Set to saturation points 
	 * if reduceSigmaForSaturatedPredictions is off*/
	public void setNoEvalPast(double noEvalPast[]){ this.noEvalPast = noEvalPast; setChanged("noEvalPast"); }
	
}
