/** This file is part of LIDAR Analysis work carried as part of a PhD with Imperial College.  
 * Copyright (c) 2010 Oliver Ford
 *
 * The content, concept and implementation are the intellectual property of Oliver Ford but permission
 * is given for it's use and further development by Jakob Svensson or any person or organisation
 * directly given such permissions by Oliver Ford or by Jakob Svensson.
 */
package seed.minerva.lidar.model;

import seed.minerva.ConnectionPoint;
import seed.minerva.StateFullNodeImpl;
import seed.minerva.nodetypes.DoubleArray;
import seed.minerva.nodetypes.DoubleMatrix;
import seed.minerva.nodetypes.IntegerArray;
import seed.minerva.nodetypes.IntegerValue;

/** Provides stray light for all channels from a series of Gaussian peaks. Width is FWHM */
public class GaussianPeaksStrayLight extends StateFullNodeImpl implements DoubleMatrix {
	public static final String defaultName = "GaussianPeaksStrayLight";
	public static final double r2p = Math.sqrt(2 * Math.PI);
	public static final double sigToFWHM = 2 * Math.sqrt(2 * Math.log(2));

	DoubleArray peakAs, peakXs, peakWs;
	IntegerArray peakDefChannel;
	IntegerValue nChannels;
	IntegerValue signalLength;

	double strayLight[][];
	
	public GaussianPeaksStrayLight(String name) {
		super(name);
		addConnectionPoint(new ConnectionPoint("nChannels", IntegerValue.class, false, getField("nChannels")));
		addConnectionPoint(new ConnectionPoint("signalLength", IntegerValue.class, false, getField("signalLength")));
		addConnectionPoint(new ConnectionPoint("peakDefChannel", IntegerArray.class, false, getField("peakDefChannel")));
		addConnectionPoint(new ConnectionPoint("peakAs", DoubleArray.class, false, getField("peakAs")));
		addConnectionPoint(new ConnectionPoint("peakXs", DoubleArray.class, false, getField("peakXs")));
		addConnectionPoint(new ConnectionPoint("peakWs", DoubleArray.class, false, getField("peakWs")));
	}

	@Override
	public void updateState() {
		int nCh = nChannels.getInteger();
		int sigLen = signalLength.getInteger();
		
		strayLight = new double[nCh][sigLen];
		
		int ch[] = peakDefChannel.getIntegerArray();
		double A[] = peakAs.getDoubleArray();
		double x[] = peakXs.getDoubleArray();
		double w[] = peakWs.getDoubleArray();
		
		for(int i=0; i < ch.length; i++){
			double sigma = w[i] / sigToFWHM;
			
			int j0 = (int)(x[i] - 5 * w[i]), j1 = (int)(x[i] + 5 * w[i] + 0.5);
			if(j0 < 0)
				j0 = 0;
			if(j1 >= sigLen)
				j1 = sigLen - 1;
			for(int j=j0; j <= j1; j++){
				strayLight[ch[i]][j] += A[i] * Math.exp(-0.5 * Math.pow((j - x[i])/sigma, 2)); // / (sigma * r2p);
			}			
		}
	}
	
	@Override
	public void tidyUpState() {
		super.tidyUpState();
		
		strayLight = null;
	}

	@Override
	public double[][] getDoubleMatrix() { update(); return strayLight; }

}
