/** This file is part of LIDAR Analysis work carried as part of a PhD with Imperial College.  
 * Copyright (c) 2010 Oliver Ford
 *
 * The content, concept and implementation are the intellectual property of Oliver Ford but permission
 * is given for it's use and further development by Jakob Svensson or any person or organisation
 * directly given such permissions by Oliver Ford or by Jakob Svensson.
 */
package seed.minerva.lidar.model;

import oneLiners.OneLiners;
import seed.minerva.ConnectionPoint;
import seed.minerva.StateFullNodeImpl;
import seed.minerva.nodetypes.DoubleMatrix;
import seed.minerva.nodetypes.DoubleValue;
import algorithmrepository.LinearInterpolation1D;

/** Provides handling of a set of instrument functions (i.e one for each channel */
public class InstrumentFunctions extends StateFullNodeImpl implements InsFuncInterface {

		// Parents
		/** Data describing instrument function. DataX can be in seconds or in ADCU. The node will guess.
		 * That way you can read in a proper instrument function definition or just attach it to the
		 * timing pulse nodes */
		private DoubleMatrix dataX, dataY;
		/** ADC time unit in s*/
		private DoubleValue signalDt;
		/** Modification of the instrument function by stretching in x, this is a crude way of adjusting the width and hence convolution strength (both a parent and a property)*/
		private DoubleValue widthAdjust;
		

		//Properties
		/** number of points to interpolate function onto, in order to calc FWHM, peak and C.O.Ms */
		private int nCalcGrids = 1000;
		/** Modification of the instrument function by stretching in x, this is a crude way of adjusting the width and hence convolution strength (both a parent and a property)*/
		private double stretch = 1.0; // artificial width modification
		/** Remove the effect of 'stretch' from the calculation of the convolution compensation for the noise based uncertainty*/
		private boolean unstretchConvolComp = true;

		// State
		private int nCh;
		private double t[][], f[][];
		private LinearInterpolation1D interps[];
		/** Peak, Full-width-half-max and centre of mass for each instrument function 
		 * in seconds. */
		private double peaks[], FWHMs[], COMs[];
		/** Lower and highest values of t (the ends) */
		private double t0[], t1[];
				
		public InstrumentFunctions(String name) {
			super(name);
			addConnectionPoint(new ConnectionPoint("dataX", DoubleMatrix.class, false, getField("dataX")));	
			addConnectionPoint(new ConnectionPoint("dataY", DoubleMatrix.class, false, getField("dataY")));
			addConnectionPoint(new ConnectionPoint("signalDt", DoubleValue.class, false, getField("signalDt")));
			addConnectionPoint(new ConnectionPoint("widthAdjust", DoubleValue.class, true, getField("widthAdjust")));
		}
		
		@Override
		public void updateState() {
			if(t != null && !isAncestorChanged("dataX") && !isAncestorChanged("dataY") && !isPropertyChanged("stretch") && !isAncestorChanged("widthAdjust"))
				return; //nothing todo
			
			if(widthAdjust != null && isAncestorChanged("widthAdjust")){
				stretch = widthAdjust.getDouble();
				setChanged("stretch");
			}
			
			if(t == null || isAncestorChanged("dataX") || isPropertyChanged("stretch")){
				final double tIn[][] = dataX.getDoubleMatrix();
				f = dataY.getDoubleMatrix();
				nCh = tIn.length;
				
				//get the full length of the X vector for the first insFunc 
				double fullX = tIn[0][tIn[0].length-1] - tIn[0][0];
				// if it's bigger than 1 then they are probably in ADCU rather than s
				// we need them in s
				t = new double[nCh][];
				if(fullX > 1) {
					for(int ch=0; ch < nCh; ch++){
						t[ch] = new double[tIn[ch].length];
						for(int i=0; i < tIn[ch].length; i++)
							t[ch][i] = tIn[ch][i] * signalDt.getDouble();
					}
				}
				
				if(stretch != 1.0){
					for(int ch=0; ch < nCh; ch++){						
						for(int i=0; i < tIn[ch].length; i++)
							t[ch][i] *= stretch;
					}
				}
				
				//create the linear interp functions and end times
				interps = new LinearInterpolation1D[nCh];
				t0 = new double[nCh];
				t1 = new double[nCh];
				for(int ch=0; ch < nCh; ch++){
					t0[ch] = t[ch][0];
					t1[ch] = t[ch][t[ch].length - 1];
					interps[ch] = new LinearInterpolation1D(t[ch], f[ch], 0);
				}
				
			}else{  //only Y changed, just update it, no need to clone
				f = dataY.getDoubleMatrix();
				for(int ch=0; ch < nCh; ch++)
					interps[ch].setF(f[ch]);
			}
			
			double highResT[][] = new double[nCh][];
			double highResF[][] = new double[nCh][];
			for(int ch=0; ch < nCh; ch++){
				highResT[ch] = OneLiners.linSpace(t0[ch], t1[ch], nCalcGrids);
				highResF[ch] = interps[ch].eval(highResT[ch]);
			}
			
			findPeaks();
			calcFWHMs(highResT, highResF);
			calcCentreOfMasses(highResT, highResF);
		}
		
		@Override
		public void tidyUpState() {
			super.tidyUpState();
			t = null; f = null; interps = null;
			t0 = null; t1 = null; 
			peaks = null; FWHMs = null; COMs = null;
		}
		
		private void findPeaks(){
			peaks = new double[nCh];
			//Can't be totally sure a binary search would always work here, so just scan it
			for(int ch=0; ch < nCh; ch++){
				double peakF = Double.NEGATIVE_INFINITY;
				for(int i=0; i < t[ch].length; i++){
					if(f[ch][i] > peakF){
						peakF = f[ch][i];
						peaks[ch] = t[ch][i]; 
					}
				}
			}
		}
		
		private void calcFWHMs(double X[][], double Y[][]){ 
			FWHMs = new double[nCh];			
			for(int ch=0; ch < nCh; ch++){
				int peakIdx = OneLiners.getNearestIndex(X[ch], peaks[ch]);
				int n = X[ch].length; 
				
				//we'll work from the centre outwards, and find where the average of both sides falls below 1/2 of the peak height
				double limit = (n - peakIdx) < peakIdx ? (n - peakIdx) : peakIdx;
				
				int i;
				for(i=0; i < limit; i++)
					if( (Y[ch][peakIdx - i] + Y[ch][peakIdx + i])/2  < (Y[ch][peakIdx]/2) )
						break;
				
				if(i==limit)
					System.err.println("WARNING: Instrument function for ch"+ch +" never falls to half it's maximum before hitting the edge. " +
											"I'll use the whole width as averaging period calculation for non-correlated calculation");
				
				FWHMs[ch] = X[ch][peakIdx + i] - X[ch][peakIdx - i];				
			}
		}
		
		private void calcCentreOfMasses(double X[][], double Y[][]){ 
			COMs = new double[nCh];
			
			for(int ch=0; ch < nCh; ch++){
				double sumXY = 0, sumY = 0;
				for(int i=0; i < X[ch].length; i++){
					sumXY += X[ch][i] * Y[ch][i];
					sumY += Y[ch][i];
				}
				COMs[ch] = sumXY / sumY;
			}
		}
		
		public double[] getFWHMSeconds(){ update(); return FWHMs; }
		
		/** @Returns the ins.Func width in ADCU */
		public double[] getApproxFWHMsInADCUs(){
			update();
			double ret[] = new double[nCh];
			for(int ch=0;ch < nCh; ch++)
				ret[ch] = FWHMs[ch] * signalDt.getDouble();
					
			return ret;
		}
		
		/** @Retruns the convolution compensation factors for uncertainty sigma.
		 * For a constant ADC signal of white noise with a Gaussian distribution of 
		 * photoelectron counts around with width width sigmaA in each ADC time bin. 
		 * The marginal distribution on each ADC reading, after the convolution will 
		 * have sigmaC = sigmaA * cheese.
		 * 
		 * 
		 * errorOnConvolved = errorWithout * sqrt( 1 / [2 * sigma * sqrt(pi)] ), this is called 'the factor of cheeses' for no particular reason.
		 * and since sigma = FWHM / (2 * sqrt(2 * log(2))); this gives:
		 * errorOnConvolved^2 = errorWithout^2 * sqrt(2 * log(2)/pi) / FWHM
		 * errorOnConvolved^2 = errorWithout^2 * 0.6642825 / FWHM
		 */
		public double[] getCheeseFactors2(){
			update();
			double ret[] = new double[nCh];
			double sigDt = signalDt.getDouble();
			
			// Approx calculation based on gaussian instrument function
			 for(int ch=0;ch < nCh; ch++)
				ret[ch] = (unstretchConvolComp ? stretch : 1.0) * 0.6642825 / (FWHMs[ch] / sigDt);
			//System.out.print("Approx cheese2: ");
			//OneLiners.dumpArray(ret);
			// */
			
			//More accurate calculation: cheese^2 = sum_i( C_i^2 ) where C would be the normalised (to sum_i C_i = 1) instrument function
			// with each element representing 1 ADC unit.
			ret = new double[nCh];
			for(int ch=0; ch<nCh; ch++){
				double sum = 0, sum2 = 0;
				for(int i=0; i < (f[ch].length-1); i++){
					double dt = (t[ch][i+1] - t[ch][i]) / sigDt;
					sum += f[ch][i];
					sum2 += f[ch][i]*f[ch][i] / dt;
				}
				
				ret[ch] = sum2 / (sum*sum);
			}
				
			//System.out.print("Accurate cheese2: ");
			//OneLiners.dumpArray(ret);
			
			return ret;
		}
		
		
		
		public double[] getPeaks(){ 
			update();
			double ret[] = new double[nCh];
			for(int ch=0;ch < nCh; ch++)
				ret[ch] = (peaks[ch] - t0[ch]);					
			return ret;	
		}
		
		public double[] getCentresOfMasses(){
			update();
			double ret[] = new double[nCh];
			for(int ch=0;ch < nCh; ch++)
				ret[ch] = (COMs[ch] - t0[ch]);					
			return ret;
		}
		
		/** @Returns The indices of the peak of the instrument function, 
		 * 				on the given temporal grid */
		public double[] getPeakIdxs(double dt){
			double ret[] = getPeaks();
			for(int ch=0;ch < nCh; ch++)
				ret[ch] /= dt;					
			return ret;	
		}
		
		/** @Returns The indices of the centre of mass of the instrument function,
		 * 		 on the given temporal grid */
		public double[] getCentreOfMassIdxs(double dt){
			double ret[] = getCentresOfMasses();
			for(int ch=0;ch < nCh; ch++)
				ret[ch] /= dt;					
			return ret;
		}
		
		@Override
		/** @Returns the instrument functions linearly interpolated onto the given temporal grid */
		public double[][] getInsFunc(double dt){
			update();
			double ret[][] = new double[nCh][];
						
			for(int ch=0; ch < nCh; ch++){
				ret[ch] = interps[ch].eval( OneLiners.linSpace(t0[ch], t1[ch], dt) );
				double sum  = 0;
				for(int i=0; i < ret[ch].length; i++)
					sum += ret[ch][i];// * dt;
				for(int i=0; i < ret[ch].length; i++)
					ret[ch][i] /= sum;
			}
				
			
			return ret;
		}
		
		
		public double getStretch() { update(); return stretch; }
		public void setStretch(double stretch) { this.stretch = stretch; setChanged("stretch"); }		
}

