/** This file is part of LIDAR Analysis work carried as part of a PhD with Imperial College.  
 * Copyright (c) 2010 Oliver Ford
 *
 * The content, concept and implementation are the intellectual property of Oliver Ford but permission
 * is given for it's use and further development by Jakob Svensson or any person or organisation
 * directly given such permissions by Oliver Ford or by Jakob Svensson.
 */
package seed.minerva.lidar.model;

import java.util.Arrays;

import oneLiners.OneLiners;

import numericalRecipes.NumericalRecipes;

import algorithmrepository.Algorithms;
import algorithmrepository.LinearInterpolation1D;
import seed.minerva.ConnectionPoint;
import seed.minerva.RandomManager;
import seed.minerva.StateFullNodeImpl;
import seed.minerva.nodetypes.DoubleArray;
import seed.minerva.nodetypes.DoubleMatrix;
import seed.minerva.nodetypes.DoubleValue;
import seed.minerva.nodetypes.IntegerValue;

/** Calculates the non-plasma contribution to the data based on the signal component parameters*/
public class NonPlasmaSignal extends StateFullNodeImpl {
	public final static String defaultName = "NonPlasmaSignal"; 
	
	//Parents
	/** Baseline pulse position and scaling */
	private DoubleArray baselineA, baselineX0, baselineY0;
	/** Timing pulse position and scaling */
	private DoubleArray tpulseA, tpulseX0, tpulseY0;
	/** Switch-On pulse position and scaling */
	private DoubleArray switchOnA, switchOnX0, switchOnY0;
	/** Ambient light level and switch on point */
	private DoubleArray ambientBeginX, ambientLevel;
	/** Backwall position, amplitude and scale length */
	private DoubleArray backwallX0, backwallA, backwallScaleLength;
	/** Number of Channels */
	private IntegerValue nChannels;
	/** Signal length */
	private IntegerValue signalLength;	
	/** Baseline shape data */
	private DoubleMatrix baselineShapeDataX, baselineShapeDataY;
	/** Timing pulse shape data */
	private DoubleMatrix tpulseShapeDataX, tpulseShapeDataY;
	/** Adjustment for timing pulse width */
	private DoubleValue tpulseWidthAdjust;
	/** Switch-on pulse shape data */
	private DoubleMatrix switchonShapeDataX, switchonShapeDataY;
	/** cheeseFactors2 - see InstrumentFunctions.cheeseFactors2() */
	private DoubleArray cheeseFactors2;
	/** Stray Light */
	private DoubleMatrix strayLight;
	/** Instrument function interface, (only needed when adding simulated noise to non-plasma components) */
	private InsFuncInterface instrumentFunctions;
	/** ADC time unit in s (only needed when adding simulated noise to non-plasma components)*/
	private DoubleValue signalDt;
	
	
	
	//Parents as single array
	/** An array of all the parameter DoubleArray objects */
	private DoubleArray paramArrays[] = { baselineA, baselineX0, baselineY0, tpulseA, tpulseX0, tpulseY0,
									switchOnA, switchOnX0, switchOnY0, ambientBeginX, ambientLevel,
									backwallX0, backwallA, backwallScaleLength };
	/** Text names of parameters */
	public static String paramNames[] = { "baselineA", "baselineX0", "baselineY0", "tpulseA", "tpulseX0", "tpulseY0",
									"switchOnA", "switchOnX0", "switchOnY0", "ambientBeginX", "ambientLevel",
									"backwallX0", "backwallA", "backwallScaleLength" };
	

	//Properties
	private boolean tPulseEnabled[];	
	private boolean strayLightInhibit = false;
	private boolean backwallInhibit = false;
	private boolean addNonPlasmaNoise = false, nonPlasmaNoiseConvolved = false;
	
	//State
	private double photoElecs[][], photoElecsSigma2[][];
	private double ADCU[][];
	private int nCh, sigLen;
	private LinearInterpolation1D baselineShape[], tpulseShape[], switchonShape[];
	private double stray[][];
	
	private double cheese2[];
	
	//ins func stuff for convolving simulated noise
	private double insFuncs[][] = null;
	private double insFuncPeaks[];
	private double insFuncCOMs[];
	private int fftLens[];
	private double insFuncFFTs[][];
	
	
	public NonPlasmaSignal() { this(defaultName); }
	public NonPlasmaSignal(String name) {
		super(name);
		for(int i=0;i<paramArrays.length;i++)
			addConnectionPoint(new ConnectionPoint(paramNames[i], DoubleArray.class, false, getField(paramNames[i])));
		addConnectionPoint(new ConnectionPoint("nChannels", IntegerValue.class, false, getField("nChannels")));
		addConnectionPoint(new ConnectionPoint("signalLength", IntegerValue.class, false, getField("signalLength")));
		addConnectionPoint(new ConnectionPoint("baselineShapeDataX", DoubleMatrix.class, false, getField("baselineShapeDataX")));
		addConnectionPoint(new ConnectionPoint("baselineShapeDataY", DoubleMatrix.class, false, getField("baselineShapeDataY")));
		addConnectionPoint(new ConnectionPoint("tpulseShapeDataX", DoubleMatrix.class, false, getField("tpulseShapeDataX")));
		addConnectionPoint(new ConnectionPoint("tpulseShapeDataY", DoubleMatrix.class, false, getField("tpulseShapeDataY")));
		addConnectionPoint(new ConnectionPoint("switchonShapeDataX", DoubleMatrix.class, false, getField("switchonShapeDataX")));
		addConnectionPoint(new ConnectionPoint("switchonShapeDataY", DoubleMatrix.class, false, getField("switchonShapeDataY")));
		addConnectionPoint(new ConnectionPoint("cheeseFactors2", DoubleArray.class, false, getField("cheeseFactors2")));
		addConnectionPoint(new ConnectionPoint("tpulseWidthAdjust", DoubleValue.class, true, getField("tpulseWidthAdjust")));
		addConnectionPoint(new ConnectionPoint("strayLight", DoubleMatrix.class, true, getField("strayLight")));
		addConnectionPoint(new ConnectionPoint("instrumentFunctions", InsFuncInterface.class, true, getField("instrumentFunctions")));
		addConnectionPoint(new ConnectionPoint("signalDt", DoubleValue.class, true, getField("signalDt")));
	}
	
	public boolean forceSwitchOnReinit = false;
	
	@Override
	public void updateState() {
		boolean datasizeChanged = isAncestorChanged("nChannels") || isAncestorChanged("signalLength");
		boolean blChanged = isAncestorChanged("baselineShapeDataX") || isAncestorChanged("baselineShapeDataY");
		boolean tpChanged = isAncestorChanged("tpulseShapeDataX") || isAncestorChanged("tpulseShapeDataY");
		boolean soChanged = isAncestorChanged("switchonShapeDataX") || isAncestorChanged("switchonShapeDataY");
		boolean insfChanged = isAncestorChanged("cheeseFactors2");
		boolean strayChanged = (strayLight != null) && (isAncestorChanged("strayLight") || isPropertyChanged("strayLightInhibit"));

		if(datasizeChanged){ //if these change, all bets are off
			nCh = nChannels.getInteger();
			sigLen = signalLength.getInteger();
			
			tPulseEnabled = new boolean[nCh];
			for(int ch=0; ch < nCh; ch++)
				tPulseEnabled[ch] = true;
			
			photoElecs = null;
		}
				 
		if(baselineShape == null || datasizeChanged || blChanged){
			baselineShape = new LinearInterpolation1D[nCh];
			for(int ch=0; ch < nCh; ch++){
				double blX[] = baselineShapeDataX.getDoubleMatrix()[ch].clone();
				double blY[] = baselineShapeDataY.getDoubleMatrix()[ch].clone();
				baselineShape[ch] = new LinearInterpolation1D(
						blX,
						blY, blY[0]);
			}
		}
		
		if(tpulseShape == null || datasizeChanged || tpChanged){
			tpulseShape = new LinearInterpolation1D[nCh];
			double tpDataX[][] = tpulseShapeDataX.getDoubleMatrix().clone();
			double tpDataY[][] = tpulseShapeDataY.getDoubleMatrix().clone();
			for(int ch=0; ch < nCh; ch++){
				if(tpDataX.length > ch && tpDataX[ch] != null)
					tpulseShape[ch] = new LinearInterpolation1D(
							tpDataX[ch],
							tpDataY[ch], 
							0);
			}
		}		
		
		if(switchonShape == null || datasizeChanged || soChanged){
			switchonShape = new LinearInterpolation1D[nCh];			
			double soDataX[][] = switchonShapeDataX.getDoubleMatrix().clone();
			double soDataY[][] = switchonShapeDataY.getDoubleMatrix().clone();
			for(int ch=0; ch < nCh; ch++){
				if(soDataX.length > ch && soDataX[ch] != null)
					switchonShape[ch] = new LinearInterpolation1D(
							soDataX[ch],
							soDataY[ch], 
							0);
			}
		}		

		if(insfChanged)
			cheese2 = cheeseFactors2.getDoubleArray();
		
		if(strayLight == null || strayLightInhibit)
			stray = null;
		else if(stray == null || strayChanged){
			stray = strayLight.getDoubleMatrix();			
		}
		
		boolean noiseChanged = isPropertyChanged("nonPlasmaNoise");
		if(addNonPlasmaNoise && nonPlasmaNoiseConvolved){
			if(instrumentFunctions == null || signalDt == null)
				throw new RuntimeException("Instrument functions and signalDt need to be connected to add convolved noise to no plasma signal.");
			else if(insFuncs == null || isAncestorChanged("instrumentFunctions") || isAncestorChanged("signalDt")){
				prepNoiseConvol();
				noiseChanged = true;
			}
		}
		
		if(photoElecs == null ||
				datasizeChanged || tpChanged || soChanged || blChanged || insfChanged ||
				forceSwitchOnReinit || isAncestorChanged("tpulseWidthAdjust") || strayChanged ||
				isPropertyChanged("backwallInhibit") || noiseChanged){
			forceSwitchOnReinit = false;
			calcNonPlasma();
			return;
		}
		
		for(int i=0;i<paramArrays.length;i++)
			if(isAncestorChanged(paramNames[i])){
				calcNonPlasma();
				return;
			}
		
		
	}
	
	@Override
	public void tidyUpState() {
		super.tidyUpState();
		
		baselineShape = null; tpulseShape = null; switchonShape = null;
		stray = null; photoElecs = null; photoElecsSigma2 = null; ADCU = null;		
		insFuncs = null; insFuncPeaks = null; insFuncCOMs = null;
		fftLens = null; insFuncFFTs = null;
	}
	
	public void calcNonPlasma(){
		if(photoElecs == null){
			photoElecs = new double[nCh][sigLen];
			photoElecsSigma2 = new double[nCh][sigLen];
			ADCU = new double[nCh][sigLen];
		}
		
		double blX0[] = baselineX0.getDoubleArray();
		double blY0[] = baselineY0.getDoubleArray();
		double blA[] = baselineA.getDoubleArray();
		double tpX0[] = tpulseX0.getDoubleArray();
		double tpY0[] = tpulseY0.getDoubleArray();
		double tpA[] = tpulseA.getDoubleArray();
		double soX0[] = switchOnX0.getDoubleArray();
		double soY0[] = switchOnY0.getDoubleArray();
		double soA[] = switchOnA.getDoubleArray();
		double bwX0[] = backwallX0.getDoubleArray();		
		double bwA[] = backwallA.getDoubleArray();
		double bwL[] = backwallScaleLength.getDoubleArray();
		double alBegin[] = ambientBeginX.getDoubleArray();
		double alLevel[] = ambientLevel.getDoubleArray();
		double wAdjust = (tpulseWidthAdjust!=null) ? tpulseWidthAdjust.getDouble() : 1.0;
		
		for(int ch=0;ch<nCh;ch++){
			//get interpolation grid and calc first/last that will be inside it
			double blShapeX[] = baselineShape[ch].getX(); 
			double blShapeY[] = baselineShape[ch].getF();
			double tpShapeX[] = (tpulseShape[ch] != null) ? tpulseShape[ch].getX() : null;
			double soShapeX[] = (switchonShape[ch] != null) ? switchonShape[ch].getX() : null;
			
			int blI0=0, blI1=0, tpI0=0, tpI1=0, soI0=0, soI1=0;			
			try{
				blI0 = (int)(blShapeX[0] + blX0[ch]);
				blI1 = (int)(blShapeX[blShapeX.length-1] + blX0[ch] + 1);
				if(tpShapeX != null && tpX0.length > ch){ //may not have a tpulse on this channel
					tpI0 = (int)(tpShapeX[0]/wAdjust + tpX0[ch]);
					tpI1 = (int)(tpShapeX[tpShapeX.length-1]/wAdjust + tpX0[ch] + 1);
				}else{
					tpI0 = -1; tpI1 = -1;
				}
				if(soShapeX != null && soX0.length > ch){
					soI0 = (int)(soShapeX[0] + soX0[ch]);
					soI1 = (int)(soShapeX[soShapeX.length-1] + soX0[ch] + 1);
				}else {
					soI0 = -1; soI1 = -1;
				}
			}catch(NullPointerException err){	
				err.printStackTrace();
				throw new RuntimeException("NonPlasmaSignal is being used on uninitialised signal components. You must manually call SignalComponentsContainer.createParameterArrays() first.");
			}catch(ArrayIndexOutOfBoundsException err){
				err.printStackTrace();
				throw new RuntimeException("NonPlasmaSignal is being used on uninitialised signal components. You must manually call SignalComponentsContainer.createParameterArrays() first.");			
			}
			
			
			for(int i=0; i < sigLen; i++){
				ADCU[ch][i] = 0;
				photoElecs[ch][i] = 0;
				photoElecsSigma2[ch][i] = 0;
				
				//electronic baseline is an ACDU quantity, not photoelecs
				ADCU[ch][i] += blY0[ch];
				if(i < blI0)
					ADCU[ch][i] += blA[ch] * blShapeY[0];
				else if(i > blI1)
					ADCU[ch][i] += blA[ch] * blShapeY[blShapeY.length-1];
				else
					ADCU[ch][i] += blA[ch] * baselineShape[ch].eval((double)i - blX0[ch]);
				
				photoElecs[ch][i] += ((tpY0.length > ch) ? tpY0[ch] : 0) + soY0[ch]; //they are usually 0 anyway
				
				//add ambient light if past its begin point
				if(i > alBegin[ch])
					photoElecs[ch][i] += alLevel[ch];
				
				//add the timing pulse to the photoElecs
				if(i > tpI0 && i <= tpI1)
					photoElecs[ch][i] += tpA[ch] * tpulseShape[ch].eval( (i - tpX0[ch])/wAdjust);
			
				//add the switch-on pulse to the photoElecs
				if(i > soI0 && i <= soI1)
					photoElecs[ch][i] += soA[ch] * switchonShape[ch].eval(i - soX0[ch]);
				
				//Add stray light
				if(stray != null)
					photoElecs[ch][i] += stray[ch][i];
				
				//Add the backwall and simulated saturation past it
				if(backwallInhibit){
					photoElecsSigma2[ch][i] = cheese2[ch] * photoElecs[ch][i];
				}else{// if(i < bwX0[ch]){
					photoElecs[ch][i] += bwA[ch] * Math.exp((i - bwX0[ch]) / bwL[ch]);
					photoElecsSigma2[ch][i] = cheese2[ch] * photoElecs[ch][i];
				}/*else{
					photoElecs[ch][i] = Double.POSITIVE_INFINITY;
					photoElecsSigma2[ch][i] = Double.NaN;
				}*/
				
					
				/* negative photoelectron counts here arn't actually too bad.
				 * the instrument response (which isn't seperated for non-plasma signal) can
				 * take the signal negative. The sqrt(N) noise on this doesn't make any sense
				 * so just set it to 0 and let it get picked up in the electronic
				 *  noise allowance later on. */				
				if(photoElecsSigma2[ch][i] <= 0)
					photoElecsSigma2[ch][i] = 0;
			}

			
			if(addNonPlasmaNoise){ //Noise addition
				double noise[] = new double[sigLen];
				
				//only the noise on the ambient light gets added here since noise on the pulses (which were a delta-function, cannot cause deformation of that shape).
				for(int i=0; i < sigLen; i++)					
					if(i > alBegin[ch] && alLevel[ch] > 0 && i < bwX0[ch])
							noise[i] = RandomManager.instance().nextNormal(0, Math.sqrt(alLevel[ch]));
				
				if(nonPlasmaNoiseConvolved)
					convolveNoise(ch, noise);
				
				for(int i=0; i < sigLen; i++){					
					photoElecs[ch][i] += noise[i];						                           
				}
			}
			
		
		}
	}
	
	
	
	
	private void prepNoiseConvol(){
		double dt = signalDt.getDouble();
		insFuncs = instrumentFunctions.getInsFunc(dt);
		insFuncPeaks = instrumentFunctions.getPeaks();
		insFuncCOMs = instrumentFunctions.getCentresOfMasses();
		
		
					
		//find the biggest required convolving length (los + insfunc)
		fftLens = new int[nCh];
		insFuncFFTs = new double[nCh][];
		for(int ch=0; ch < nCh; ch++){							
			//calculate the required fft length (round up to  it a power of 2)
			fftLens[ch] = sigLen + insFuncs[ch].length;
			int npow = (int)Math.ceil( Math.log(fftLens[ch]) / Math.log(2) );
			fftLens[ch] = (int)Math.pow(2, npow);
				
			insFuncFFTs[ch] = new double[fftLens[ch]];
			System.arraycopy(insFuncs[ch], 0, insFuncFFTs[ch], 0, insFuncs[ch].length);
				
			NumericalRecipes.drealft(insFuncFFTs[ch], fftLens[ch], 1);
		}
	}
	
	private void convolveNoise(int ch, double noise[]){
		
		double dt = signalDt.getDouble();
		double lightSig[] = new double[fftLens[ch]];
				
		System.arraycopy(noise, 0,  lightSig, 0, noise.length);				
		NumericalRecipes.drealft(lightSig, fftLens[ch], 1); //fft the signal
		lightSig = NumericalRecipes.fftMul(insFuncFFTs[ch], lightSig); //convolve the with the insFunc
		NumericalRecipes.drealft(lightSig, fftLens[ch], -1); //de-fft
		for(int i=0; i < fftLens[ch]; i++)	//and a small correction for drealft routine
			lightSig[i] *= 2.0 / (double)fftLens[ch];
	
		// we have to note that the los time/dist axis will have been shifted by the convolution
		//losT0s[ch] -= insFuncCOMs[ch];
		int i0 = (int)(insFuncCOMs[ch] / dt);
		
		System.arraycopy(lightSig, i0, noise, 0, noise.length);
		
	}

	/** @returns the signal[channel][ADCindex] */ 
	public double[][] getNonPlasmaPhotoElecsPerADCU(){ update(); return photoElecs; }
	/** @returns the signal[channel][ADCindex] */ 
	public double[][] getNonPlasmaPhotoElecsPerADCUSigma2(){ update(); return photoElecsSigma2; }
	/** @returns the non-photoelectron ADC unit count for non-plasma signal: signal[channel][ADCindex] */
	public double[][] getElecBasline(){ update(); return ADCU; }
	/** @returns The interpolation objects used for the timing pulse shapes - array over channels */ 
	public LinearInterpolation1D[] getTPulseShapes(){ update(); return tpulseShape; };
	/** @returns The interpolation objects used for the switch-on pulse shapes - array over channels */ 
	public LinearInterpolation1D[] getSwitchOnShapes(){ update(); return switchonShape; };
	/** @returns The interpolation objects used for the baseline shapes - array over channels */ 
	public LinearInterpolation1D[] getBaselineShapes(){ update(); return baselineShape; };
	
	public boolean getStrayLightInhibit(){ /*update(); not needed */ return this.strayLightInhibit;  }
	public void setStrayLightInhibit(boolean inhibit){ this.strayLightInhibit = inhibit; setChanged("strayLightInhibit"); }
	public boolean getBackwallInhibit(){ /*update(); not needed */ return this.backwallInhibit; }
	public void setBackwallInhibit(boolean inhibit){ this.backwallInhibit = inhibit; setChanged("backwallInhibit"); }
	
	public void setNonPlasmaNoiseGeneration(boolean addNonPlasmaNoise, boolean nonPlasmaNoiseConvolved){
		this.addNonPlasmaNoise = addNonPlasmaNoise;
		this.nonPlasmaNoiseConvolved = nonPlasmaNoiseConvolved;
		setChanged("nonPlasmaNoise");
	}
	
		
}
