/** This file is part of LIDAR Analysis work carried as part of a PhD with Imperial College.  
 * Copyright (c) 2010 Oliver Ford
 *
 * The content, concept and implementation are the intellectual property of Oliver Ford but permission
 * is given for it's use and further development by Jakob Svensson or any person or organisation
 * directly given such permissions by Oliver Ford or by Jakob Svensson.
 */
package seed.minerva.lidar.model;

import java.util.Arrays;

import seed.minerva.ConnectionPoint;
import seed.minerva.StateFullNodeImpl;
import seed.minerva.diagnostics.ServiceManager;
import seed.minerva.diagnostics.UnitManager;
import seed.minerva.nodetypes.DoubleArray;
import seed.minerva.nodetypes.DoubleMatrix;
import seed.minerva.nodetypes.DoubleValue;
import seed.minerva.nodetypes.ScalarND;

/** Provides the number of photoelectrons leaving each PMT's photocathode per L.O.S time unit 
 * but assuming a delta function laser pulse. */  
public class PhotoElectronPlasmaSignal extends StateFullNodeImpl implements DoubleMatrix {
	public final static String defaultName = "PhotoElectronPlasmaSignal";
	
	/** Electron density profile */
	ScalarND electronDensity;
	/** Electron temperature profile */
	ScalarND electronTemperature;
	/** Laser energy */
	DoubleValue laserEnergy;
	/** Vignetting */
	ScalarND vignetting;
	/** Calibration constant - the overall optical constant for the whole transmission path */
	DoubleValue uberConst;
	/** Calibration constants - transmission factors for torus windows */
	DoubleArray windowTransmission;
	/** Scattering table - thomson scattering calculation */
	DoubleMatrix thomsonScatteringTable;
	/** Calibration constants - relative sensitivity (per channel) */
	DoubleArray relativeSensitivity;
	/** Line of sight coordinates [R/Z][{start, end, beamdump} */
	DoubleMatrix lineOfSightCoords;
	/** Vignetting modifications */
	DoubleValue vignettingShift;
	DoubleValue vignettingAdjustGradient;
	DoubleValue vignettingAdjustPivotR;
	//Properties
	/** interval along long of sight at which to calculate scattering */
	private double losDl = 0.005; 	//default 5mm
	
	
	//State
	private double table[][];
	private int nCh;

	private double losR[], losZ[];
	private double losDistFromBackwall[];
	/** first row of this is losDistFromBackwall */
	private double losPhotoElecs[][];
	private double losNe[], losTe[];
	private double losVignet[];
	private int nLos;
	
	private double winTrans[],relSens[];
	private double laserEng, U;
	
	
	public PhotoElectronPlasmaSignal() { this(defaultName); }
	
	private void createLOS(){
		double losCoords[][] = lineOfSightCoords.getDoubleMatrix();
		
		double r0 = losCoords[0][0], r1 = losCoords[0][1], rBackwall = losCoords[0][2];
		double z0 = losCoords[1][0], z1 = losCoords[1][1], zBackwall = losCoords[1][2];
		double losLength = Math.sqrt((r1-r0)*(r1-r0) + (z1-z0)*(z1-z0));
		nLos = (int)(losLength / losDl) + 1;
		losR = new double[nLos];
		losZ = new double[nLos];
		losDistFromBackwall = new double[nLos];
		double dR = (r1-r0)/(nLos-1);
		double dZ = (z1-z0)/(nLos-1);
		double lBackwall = Math.sqrt((rBackwall-r0)*(rBackwall-r0) + (zBackwall-z0)*(zBackwall-z0)); //+ve dist of backwall from (r0,z0)
		
		for(int i=0; i < nLos; i++){
			losR[i] = r0 + i * dR;
			losZ[i] = z0 + i * dZ;
			double l = Math.sqrt( i*i*(dR*dR + dZ*dZ) ); //dist from (r0,z0) of this point
			losDistFromBackwall[i] = l - lBackwall; //dist past backwall, travlleing up the los in +ve steps
		}
	}	
	
	public PhotoElectronPlasmaSignal(String name) {
		super(name);
		addConnectionPoint(new ConnectionPoint("laserEnergy", DoubleValue.class, false, getField("laserEnergy")));
		addConnectionPoint(new ConnectionPoint("vignetting", ScalarND.class, false, getField("vignetting")));
		addConnectionPoint(new ConnectionPoint("electronDensity", ScalarND.class, false, getField("electronDensity")));
		addConnectionPoint(new ConnectionPoint("electronTemperature", ScalarND.class, false, getField("electronTemperature")));
		addConnectionPoint(new ConnectionPoint("uberConst", DoubleValue.class, false, getField("uberConst")));
		addConnectionPoint(new ConnectionPoint("windowTransmission", DoubleArray.class, false, getField("windowTransmission")));
		addConnectionPoint(new ConnectionPoint("thomsonScatteringTable", DoubleMatrix.class, false, getField("thomsonScatteringTable")));
		addConnectionPoint(new ConnectionPoint("relativeSensitivity", DoubleArray.class, false, getField("relativeSensitivity")));
		addConnectionPoint(new ConnectionPoint("lineOfSightCoords", DoubleMatrix.class, false, getField("lineOfSightCoords")));
		addConnectionPoint(new ConnectionPoint("vignettingShift", DoubleValue.class, true, getField("vignettingShift")));
		addConnectionPoint(new ConnectionPoint("vignettingAdjustGradient", DoubleValue.class, true, getField("vignettingAdjustGradient")));
		addConnectionPoint(new ConnectionPoint("vignettingAdjustPivotR", DoubleValue.class, true, getField("vignettingAdjustPivotR")));
	}
	
	private void calcScattering(){

		double logTeTableParam = table[0][0];
		
		if(losPhotoElecs == null)
			losPhotoElecs = new double[nCh][losDistFromBackwall.length];
		
		//calc scattering on LOS grid
		for(int i=0;i<nLos;i++) {
			
			//find nearest entry in the scattering table (use our own interp here since Te is logarithmic)
			double teIdx = Math.log(losTe[i]) / logTeTableParam;
			int j = (int)Math.floor(teIdx);
			double f = teIdx - j; 
						
			if(j >= (table[0].length-2)){
				System.err.println("Temperature table overrun: losTe["+i+"] = " + losTe[i]);
				j=0; f=0;
				for(int ch=0; ch < nCh; ch++)
					losPhotoElecs[ch][i] = Double.POSITIVE_INFINITY;
			}
			
			if(j < 1){
				for(int ch=0; ch < nCh; ch++)
					losPhotoElecs[ch][i] = 0;
			}else{
				//linear interp the table (nearest would do but linear interp makes it change continuously for optimisers)
				for(int ch=0; ch < nCh; ch++)
					losPhotoElecs[ch][i] += U * laserEng * losNe[i] * 
								(((1.0-f) * table[ch+1][j+1]) + (f * table[ch+1][j+2])); 
			}
				
			// losPhotoElecs[ch+1][i+1] is now number of photons leaving the plasma region in question, that fall within this channels spectral range
			// per unit time, per unit solid angle

			// So now apply vignetting, sensitivity (all 1D) etc for each channel for this los point
			for(int ch=0; ch < nCh; ch++){
				
				double v = losVignet[i];
				/*double z0 = los.z[(int)(los.nz / 2.0)]; //central z				
				if(evalVignetCorrection)
					v *= (1 + vignetCorrectGradient.get() * (los.z[i] - z0) );
					*/  
				losPhotoElecs[ch][i] *= losDl;
				losPhotoElecs[ch][i] *= v;	//apply vignetting - solid angle - so it's now in photons that will be collected by optics per unit time
				losPhotoElecs[ch][i] *= winTrans[ch]; //apply window transmission factors - this converts it to number of photons actually hitting the mirrors per unit time
				losPhotoElecs[ch][i] *= relSens[ch]; //this converts it to number of photoelectrons leaving this PMTs photocathode per unit time
				
				if(losPhotoElecs[ch][i] < 0) //can't have -ve intensity
					losPhotoElecs[ch][i] = 0;
			}
		}
	}

	@Override
	public void updateState() {				
		if(table == null || isAncestorChanged("thomsonScatteringTable")){
			table = thomsonScatteringTable.getDoubleMatrix();			
			nCh = table.length - 1;
			losPhotoElecs = null;
		}
		
		if(losR == null || isAncestorChanged("lineOfSightCoords") || isPropertyChanged("losDl") || losDistFromBackwall == null){
			createLOS();
			losNe = null;
			losTe = null;
			losVignet = null;
		}
		
		if( losVignet == null || isAncestorChanged("vignetting") ||
				(vignettingShift != null && isAncestorChanged("vignettingShift")) ||
				(vignettingAdjustPivotR != null && isAncestorChanged("vignettingAdjustPivotR")) ||
				(vignettingAdjustGradient != null && isAncestorChanged("vignettingAdjustGradient")) ){
			
			double vigR[] = losR;
			
			if(vignettingShift != null){
				double Rshift = vignettingShift.getDouble();				
				vigR = new double[losR.length];
				for(int i=0; i < losR.length; i++)
					vigR[i] = losR[i] - Rshift;
			}
			//get vignetting on los grid (could be 3D in theory)
			losVignet = vignetting.eval(new double[][]{ vigR, new double[nLos], losZ });
			
			if(vignettingAdjustGradient != null){
				if(vignettingAdjustPivotR == null)
					throw new RuntimeException("Must have both or niether " +
							"vignettingAdjustGradient and vignettingAdjustPivotR");
				
				double R0 = vignettingAdjustPivotR.getDouble();
				double grad = vignettingAdjustGradient.getDouble();
				
				for(int i=0; i < losR.length; i++){
					losVignet[i] *= 1 + (losR[i] - R0) * grad;
				}
				
			}
			losPhotoElecs = null;
		}
		
		if( isAncestorChanged("windowTransmission")){
			winTrans = windowTransmission.getDoubleArray();
			losPhotoElecs = null;
		}
		
		if (isAncestorChanged("relativeSensitivity")){
			relSens = relativeSensitivity.getDoubleArray();
			losPhotoElecs = null;
		}
		
		if (isAncestorChanged("laserEnergy")){
			laserEng = laserEnergy.getDouble();
			losPhotoElecs = null;
		}
		
		if( isAncestorChanged("uberConst")){
			U = uberConst.getDouble();
			if(U < 0) U = 0; //if uberconst is -ve, just don't bother, any sensible prior will kill this anyway
			losPhotoElecs = null;
		}
		
		if(losNe == null || isAncestorChanged("electronDensity") || isAncestorChanged("electronTemperature")){
			double losCoords[][] = new double[][]{ losR, new double[nLos], losZ };
		
			//get Te and Ne on the los
			losNe = electronDensity.eval(losCoords);
			losTe = electronTemperature.eval(losCoords);
			
			//we want these in their S.I unit
			UnitManager unitMan = ServiceManager.getInstance().getUnitManager();
			double neUnit = unitMan.getUnit("ElectronDensity");
			double TeUnit = unitMan.getUnit("ElectronTemperature");
			for(int i=0; i < losNe.length; i++){
				losNe[i] *= neUnit;
				losTe[i] *= TeUnit;
			}
			
			losPhotoElecs = null;
		}
	
		if( losPhotoElecs == null)				
			calcScattering();
	}
	
	@Override
	public void tidyUpState() {
		super.tidyUpState();
		
		losPhotoElecs = null;
		losNe = null; losTe = null;
		losVignet = null;
		losR = null;
		losZ = null;
		losDistFromBackwall = null;
		table = null;
	}

	@Override
	/** los photoElecs [{dist from backwall, photoElecs}][i] */
	public double[][] getDoubleMatrix() { 
		update();
		double ret[][] = new double[nCh+1][];
		ret[0] = losDistFromBackwall;
		System.arraycopy(losPhotoElecs, 0, ret, 1, nCh);
		return ret;
	}
	
	/** Getter for LOS quantities, mostly for debug/state output */
	public double[] getLosR() { update(); return losR; }
	public double[] getLosZ() { update(); return losZ; }
	public double[] getLosNe() { update(); return losNe; }
	public double[] getLosTe() { update(); return losTe; }
		
	public void setLosDl(double losDl){ setChanged("losDl"); this.losDl = losDl; }
	public double getLosDl(double losDl){ return losDl; }
	
}
