/** This file is part of LIDAR Analysis work carried as part of a PhD with Imperial College.  
 * Copyright (c) 2010 Oliver Ford
 *
 * The content, concept and implementation are the intellectual property of Oliver Ford but permission
 * is given for it's use and further development by Jakob Svensson or any person or organisation
 * directly given such permissions by Oliver Ford or by Jakob Svensson.
 */
package seed.minerva.lidar.model;

import oneLiners.OneLiners;
import algorithmrepository.Algorithms;
import numericalRecipes.NumericalRecipes;
import seed.minerva.ConnectionPoint;
import seed.minerva.RandomManager;
import seed.minerva.StateFullNodeImpl;
import seed.minerva.diagnostics.ServiceManager;
import seed.minerva.nodetypes.DoubleArray;
import seed.minerva.nodetypes.DoubleMatrix;
import seed.minerva.nodetypes.DoubleValue;
import seed.minerva.nodetypes.IntegerValue;

/** Provides the plasma part of the ADC signal, on the ADC timebase in ADC units
 * This does the convolution and the timing.
 * @author oliford
 * 
 * TODO: Port from SoMuchForSubtlety a) 2D L.O.S and maybe b) correlated evaluation mode. 
 */
public class PlasmaSignalOnADCTimebase extends StateFullNodeImpl {
	public final static String defaultName = "PlasmaSignalOnADCTimebase";
	private  final static double c = 299792458;		//speed of light
		
	/** Calibration constants - time between timing pulse and backwall (per channel)*/
	DoubleArray tpulseToBackwallTime;
	/** Calibration constant - constant adjustment to timing for all channels */ 
	DoubleValue timingAdjust;
	/** ADC time unit in s*/
	DoubleValue signalDt;
	/** Timing pulse position for timing */
	DoubleArray tpulseX0;
	/** Plasma signal in photoelectrons */
	DoubleMatrix photoElectronPlasmaSignal;
	/** Signal length */
	private IntegerValue signalLength;
	/** Instrument functions */
	InsFuncInterface instrumentFunctions;
	
	
	//Properties
	/* *vvv* Minerva/SMFS common code *vvv* */
	private int evalMode = 0;	
	public int getEvalMode() { return evalMode; }
	public void setEvalMode(int evalMode) { setChanged("evalMode"); this.evalMode = evalMode; }

	/** Evaluation mode as if there was no instrument function (Typically for simulation). 
	 * LOS scattering is sampled at the centre of each ADC bin for that ADC value. */
	public final static int EVALMODE_NO_CONVOLUTION = 1;	
	/** Evaluation mode without convolving the predicted signal but including the error reducing effect of the convolution
	 * (Typically for fast and crude anaylsis of real data) */
	public final static int EVALMODE_CONVOLUTION_ERROR_COMPENSATED = 2;
	/** LOS Convolved analysis - ignore correlated errors effect but convolve prediction and include error reducing effect
	 * Convolution is done on the high-resolution LOS grid and then sampled. */
	public final static int EVALMODE_LOS_CONVOLVE_AND_SAMPLE = 3;
	/** Data grid Convolved analysis - ignore correlated errors effect but convolve prediction and include error reducing effect
	 * SMFS Only - not supported in Minerva. */
	public final static int EVALMODE_INTEGATE_GRID_AND_CONVOLE_ON_DATA = 4;
	/** Full correlated calculation mode - Convolution effectively performed on data grid.
	 * SMFS Only - not supported in Minerva. */
	public final static int EVALMODE_CORRELATED = 5;
	/* *^^^* Minerva/SMFS common code *^^^* */
	
	private boolean addPlasmaPENoise = false;
	
	/** Minimum values of photo elecs to use in calc of Gaussian approx to Poisson likelihood 
	 * minimum N in sigma = sqrt(N). Use if youreally think low light level blips are messing up results, 
	 * otherwise leave at 0  */
	private double minPhotoElecsInPoissonNoise = 0;
	
	/** For simulation purposes, add noise to photoelectron signals.
	 * 
	 * This doesn't fit in the usual graph way of doing this, because that would require
	 * most of the LIDAR nodes to generate and pass the full covariance matrix that describe 
	 * the possible noise created from the addition of noise before the convolution. 
	 * This is a very costly process and achieves little benefit over just using the 
	 * uncorrelated covariance for evaluating, and actually doing the noise addition before the convolution. 
	 */
	public void setPhotoElecNoiseGeneration(boolean addPlasmaPENoise){ this.addPlasmaPENoise = addPlasmaPENoise; setChanged("addPlasmaPENoise"); }
	
	
	//State
	/** Current los size/dt */
	double insFuncDt;
	int losN;
	/** instrument function properties */
	double insFuncs[][], insFuncPeaks[], insFuncCOMs[], cheese2[], insFuncFFTs[][];
	int fftLens[];
	
	double signal[][];
	double sigma2[][];
	
	public PlasmaSignalOnADCTimebase() { this(defaultName); }
	public PlasmaSignalOnADCTimebase(String name) {
		super(name);
		addConnectionPoint(new ConnectionPoint("tpulseToBackwallTime", DoubleArray.class, false, getField("tpulseToBackwallTime")));
		addConnectionPoint(new ConnectionPoint("timingAdjust", DoubleValue.class, false, getField("timingAdjust")));
		addConnectionPoint(new ConnectionPoint("signalDt", DoubleValue.class, false, getField("signalDt")));
		addConnectionPoint(new ConnectionPoint("tpulseX0", DoubleArray.class, false, getField("tpulseX0")));
		addConnectionPoint(new ConnectionPoint("photoElectronPlasmaSignal", DoubleMatrix.class, false, getField("photoElectronPlasmaSignal")));
		addConnectionPoint(new ConnectionPoint("signalLength", IntegerValue.class, false, getField("signalLength")));
		addConnectionPoint(new ConnectionPoint("instrumentFunctions", InsFuncInterface.class, false, getField("instrumentFunctions")));
	}
	
	public double[][] getPlasmaPhotoElecsPerADCU(){ update(); return signal; }
	public double[][] getPlasmaPhotoElecsPerADCUSigma2(){ update(); return sigma2; }
	
	private void calcSignal(){
		double photoElecs[][] = photoElectronPlasmaSignal.getDoubleMatrix().clone(); //1st row has dist from backwall
		// photoElecs itself is now cloned, i.e we have a new container for the rows, it hasn't cloned the rows themselves
		// but the only modification we do (the convolution) we copy the row data anyway
		
		int nCh = photoElecs.length - 1;
		int sigLen = signalLength.getInteger();
		
		double losz[] = photoElecs[0];
		double losDz = losz[1] - losz[0]; //los should be constant
		double losDt = 2 * losDz / c;
		losN = photoElecs[0].length;
		
		double losT0s[] = new double[nCh]; // we will want to know the time 'after' the backwall of the first  
		for(int ch=0; ch < nCh; ch++)	// element of the los. (This will probaly be -ve). We define it 
			losT0s[ch] = 2 * losz[0] / c;	//as an array here as the convolution might change it.
		
		
				
		if(evalMode == EVALMODE_LOS_CONVOLVE_AND_SAMPLE){ //if signal convolution requested...
						
			if(addPlasmaPENoise) //add noise before convolution				
				photoElecs = addNoise(photoElecs);
			
			//Need instrument functions (also need to re-get/recalc fft if los Dt or length has changed)
			if(insFuncs == null || insFuncDt != losDt || losN != photoElecs.length){
				//Get the instrument functions on the LOS timebase, and the index of it's centre/peak
				insFuncDt = losDt;
				insFuncs = instrumentFunctions.getInsFunc(Math.abs(losDt));
				insFuncPeaks = instrumentFunctions.getPeaks();
				insFuncCOMs = instrumentFunctions.getCentresOfMasses();
							
				//find the biggest required convolving length (los + insfunc)
				fftLens = new int[nCh];
				insFuncFFTs = new double[nCh][];
				for(int ch=0; ch < nCh; ch++){
						//calculate the required fft length (round up to  it a power of 2)
						fftLens[ch] = losN + insFuncs[ch].length;
						int npow = (int)Math.ceil( Math.log(fftLens[ch]) / Math.log(2) );
						fftLens[ch] = (int)Math.pow(2, npow);
						
						insFuncFFTs[ch] = new double[fftLens[ch]];
						System.arraycopy(insFuncs[ch], 0, insFuncFFTs[ch], 0, insFuncs[ch].length);
						
						//do the FFT now to speed up when only the los changes
						NumericalRecipes.drealft(insFuncFFTs[ch], fftLens[ch], 1);
				}
			}
			
			for(int ch=0; ch < nCh; ch++){
				
				//actually do the convolution			
				double lightSig[] = new double[fftLens[ch]];
				
				System.arraycopy(photoElecs[ch+1], 0,  lightSig, 0, losN);				
				NumericalRecipes.drealft(lightSig, fftLens[ch], 1); //fft the signal
				lightSig = NumericalRecipes.fftMul(insFuncFFTs[ch], lightSig); //convolve the with the insFunc
				NumericalRecipes.drealft(lightSig, fftLens[ch], -1); //de-fft
				for(int i=0; i < fftLens[ch]; i++)	//and a small correction for drealft routine
					lightSig[i] *= 2.0 / (double)fftLens[ch];
				
				//put that back into the photoElecs array
				photoElecs[ch+1] = lightSig; 
				// we have to note that the los time/dist axis will have been shifted by the convolution
				losT0s[ch] -= insFuncCOMs[ch];
				
			}
		}
		
		// now sample the (possibly convolved) scattering onto the data grid
		signal = new double[nCh][sigLen];
		sigma2 = new double[nCh][sigLen];
		
		double tpX0s[] = tpulseX0.getDoubleArray();
		double tpbw[] = tpulseToBackwallTime.getDoubleArray();
		
		if((evalMode == EVALMODE_CONVOLUTION_ERROR_COMPENSATED || evalMode == EVALMODE_LOS_CONVOLVE_AND_SAMPLE)
				&& cheese2 == null)
			cheese2 = instrumentFunctions.getCheeseFactors2();
		
		double tAdjust = timingAdjust.getDouble();
		
		double sigDt = signalDt.getDouble();
		
		
		for(int ch=0; ch < nCh; ch++){
			for(int i=0; i < sigLen; i++){				
				double tpX0 = (tpX0s.length > ch) ? tpX0s[ch] : tpX0s[0]; //use ch0 if this one is invalid (not there)
				double timeAfterBackwall = (i - tpX0) * sigDt - tpbw[ch] + tAdjust;
				
				// linear interp the los photoelecs (it's actually an integration of the ADC time period
				// but we assume that the instrument function includes that in the sense that e.g. doing a
				// 1ms gaussian window integration of a function is the same as convolving it with a 1ms gaussian
				// and sampling at the centre.
				
				// However... the convolution is normalised (should be), so we do need to apply 
				// the difference between the losDt and the sigDt to get from the number of photoelectrons in the
				// the los bin, to the number of photoelectrons in the ADC bin
				
				double f = (timeAfterBackwall - losT0s[ch]) / losDt;
				int j = (int)f;
				f -= j;
				
				if(j>=0 && j < (photoElecs[ch+1].length-1)){					
					signal[ch][i] = (1.0-f)*photoElecs[ch+1][j] + f*photoElecs[ch+1][j+1];
					//that is _average_ over period of the ADC bin of the number of photoelecs /per los unit/, 
					
					signal[ch][i] *= sigDt / losDt; 	//now number of photoelectrons in the ADC bin

					// This is where we create the light-signal part of the likliehood disitribution
					// sigma2 = N, or sigma = sqrt(N) along the lines of the Gaussian approximated Possion likelihood.
					sigma2[ch][i] = (signal[ch][i] < minPhotoElecsInPoissonNoise) ? minPhotoElecsInPoissonNoise : signal[ch][i];  
					
					//for non-convolved noise simulation, we add the noise now, so that it's on the correct number
					if(addPlasmaPENoise && (evalMode == EVALMODE_NO_CONVOLUTION || evalMode == EVALMODE_CONVOLUTION_ERROR_COMPENSATED)){						
						signal[ch][i] += RandomManager.instance().nextNormal(0, Math.sqrt(signal[ch][i]));
						if(signal[ch][i] < 0)
							signal[ch][i] = 0;
					}
					
					if(evalMode == EVALMODE_CONVOLUTION_ERROR_COMPENSATED || evalMode == EVALMODE_LOS_CONVOLVE_AND_SAMPLE)
						sigma2[ch][i] *= cheese2[ch];
				}
			}	
		}
	}
	
	private double[][] addNoise(double photoElecs[][]){
		double withNoise[][] = new double[photoElecs.length][];
		withNoise[0] = photoElecs[0];
		RandomManager randMan = RandomManager.instance();
	
				
		for(int i=1; i < photoElecs.length; i++){
			withNoise[i] = new double[photoElecs[i].length];
			for(int j=0; j < photoElecs[i].length; j++)
			withNoise[i][j] = randMan.nextNormal(photoElecs[i][j], Math.sqrt(photoElecs[i][j]));
		}
		
		return withNoise;
	}
	
	@Override
	public void updateState() {
		
		if(evalMode != EVALMODE_NO_CONVOLUTION && evalMode != EVALMODE_LOS_CONVOLVE_AND_SAMPLE && evalMode != EVALMODE_CONVOLUTION_ERROR_COMPENSATED)
			throw new IllegalArgumentException("Unknown/unsupported evalMode "+evalMode+". Note that EVALMODE_CORRELATED and EVALMODE_INTEGATE_GRID_AND_CONVOLE_ON_DATA are not supported in Minerva.");
		
		//make sure we invalidate the instrument function if they changed
		if(isAncestorChanged("instrumentFunctions")){
			insFuncs = null;
			cheese2 = null;
		}
		
		if( signal == null || 
				isAncestorChanged("tpulseToBackwallTime") ||
				isAncestorChanged("timingAdjust") ||
				isAncestorChanged("signalDt") ||
				isAncestorChanged("tpulseX0") ||
				isAncestorChanged("photoElectronPlasmaSignal") ||
				isAncestorChanged("signalLength") ||
				isPropertyChanged("addPlasmaPENoise") ||
				isPropertyChanged("evalMode") )
			calcSignal();
	
	}
	
	@Override
	public void tidyUpState() {
		super.tidyUpState();
		
		insFuncs = null;
		insFuncPeaks = null;
		insFuncCOMs = null;
		cheese2 = null;
		insFuncFFTs = null;
		
		signal = null;
		sigma2 = null;
		
	}
 
}
