/** This file is part of LIDAR Analysis work carried as part of a PhD with Imperial College.  
 * Copyright (c) 2010 Oliver Ford
 *
 * The content, concept and implementation are the intellectual property of Oliver Ford but permission
 * is given for it's use and further development by Jakob Svensson or any person or organisation
 * directly given such permissions by Oliver Ford or by Jakob Svensson.
 */
package seed.minerva.lidar.model;

import static java.lang.Math.cos;
import static java.lang.Math.exp;
import static java.lang.Math.sin;
import static java.lang.Math.sqrt;

import java.io.IOException;
import java.nio.ByteOrder;
import java.util.zip.CRC32;

import descriptors.SignalDesc;

import mds.MdsPlus;

import algorithmrepository.LinearInterpolation1D;
import oneLiners.BinaryMatrixFile;
import oneLiners.DataConvertPureJava;
import seed.minerva.ConnectionPoint;
import seed.minerva.MinervaSettings;
import seed.minerva.StateFullNodeImpl;
import seed.minerva.diagnostics.ServiceManager;
import seed.minerva.nodetypes.DoubleMatrix;
import seed.minerva.nodetypes.DoubleValue;

/** Provides table of thomson scattering coefficients into given normalised wavelength bands */
public class ThomsonScatteringTable extends StateFullNodeImpl implements DoubleMatrix {
	private  final static double e = 1.60217646E-19;		//electron charge
	private  final static double me = 9.10938188E-31;		//electron mass
	private  final static double c = 299792458;				//speed of light
	public final static String defaultName = "ThomsonScatteringTable";
	
	private String tableCacheFile;
	
	/** Normalised spectrums */
	DoubleMatrix normalisedSpectrumWavelengths, normalisedSpectrumAmplitudes;
	/** Laser wavelength - laser wavelength in same units as spectral functions */	
	DoubleValue laserWavelength;
	
	public ThomsonScatteringTable(String tableCacheFile) { this(defaultName, tableCacheFile); }
	
	public ThomsonScatteringTable(String name, String tableCacheFile) {
		super(name);
		this.tableCacheFile = tableCacheFile;
		addConnectionPoint(new ConnectionPoint("normalisedSpectrumWavelengths", DoubleMatrix.class, false, getField("normalisedSpectrumWavelengths")));
		addConnectionPoint(new ConnectionPoint("normalisedSpectrumAmplitudes", DoubleMatrix.class, false, getField("normalisedSpectrumAmplitudes")));
		addConnectionPoint(new ConnectionPoint("laserWavelength", DoubleValue.class, false, getField("laserWavelength")));		
	}
	
	//should be Properties (needs setters, getters, setChanged() and updateState isPropChanges etc)
	private int nEntries = 10000;
	private double TeMax = 50000;
	private int nl = 200; //number of grid point over detector wavelength ranges
	
	//State
	private double table[][];
	private double logTeTableParam;

	@Override
	public void updateState() {
		if( table != null &&
				!isAncestorChanged("normalisedSpectrumWavelengths") &&
				!isAncestorChanged("normalisedSpectrumAmplitudes") &&
				!isAncestorChanged("laserWavelength") ) return;
		
		double ll = laserWavelength.getDouble();
		
		//get normspects, checksum and lookup cache file
		//if no exist generate cache file
		//set into table;
		
		double normSpecWave[][] = normalisedSpectrumWavelengths.getDoubleMatrix();
		double normSpecAmp[][] = normalisedSpectrumAmplitudes.getDoubleMatrix();
		int nCh = normSpecWave.length;
		
		CRC32 crcs[] = new CRC32[nCh];
		for(int ch=0; ch < nCh; ch++){
			crcs[ch] = new CRC32();
			crcs[ch].update(DataConvertPureJava.arrayToData1D(new double[]{ ll }, double.class, ByteOrder.LITTLE_ENDIAN));
			crcs[ch].update(DataConvertPureJava.arrayToData1D(normSpecWave[ch], double.class, ByteOrder.LITTLE_ENDIAN));
			crcs[ch].update(DataConvertPureJava.arrayToData1D(normSpecAmp[ch], double.class, ByteOrder.LITTLE_ENDIAN));
		}
				
		double l[][];
		double detSpect[][];
		
						
		if(tableCacheFile != null){			
			String fileName = tableCacheFile.replaceAll("%SIGNALS_CACHE_PATH%", SignalDesc.getDefaultCachePath());
			try{
				table = BinaryMatrixFile.load(fileName, true); //load temp table instead
				int ch;
				for(ch=0; ch < nCh; ch++){
					if(table[ch+1][0] != (double)crcs[ch].getValue())
						break;
				}
				if(ch >= nCh){
					logTeTableParam = table[0][0];
					System.out.println("Preloaded "+nEntries+" scattering table entires from '" + fileName + "' - Spectra CRCs all matched.");
					return;
				}else
					System.out.println("Scattering table in '" + fileName + "' has different checksum for channel "+ch);
			}catch(IOException err){
				System.out.print("Cannot read cached temperature table from file '"+fileName+"', ");
			}
		}
		
		System.out.print("Generating "+nEntries+" scattering table entires to "+TeMax+": ");
		table = new double[nCh+1][nEntries+1];
		
		//get the detector spectrums - normalised - we take into account the relative sensitivities in calcLosGridScattering()
		//double spectra[][][] = dataProv.normSpectrums(); // [ch][wave/amp][i]
		
		detSpect = new double[nCh][];
		double dl[] = new double[nCh];
		l = new double[nCh][nl];
		for(int ch=0; ch < nCh; ch++){
			//define grid from lowest to highest wavelength of the spectrum
			double l0 = normSpecWave[ch][0];
			double l1 = normSpecWave[ch][normSpecWave[ch].length-1 ];
			
			dl[ch] = (l1-l0)/(nl-1);
			for(int i=0;i<nl;i++)
				l[ch][i] = l0 + i*dl[ch];
			
			//and regrid the spectrum
			//detSpect[ch] = Grids.reGrid(l[ch], normSpecWave[ch], normSpecAmp[ch], false);
			detSpect[ch] = (new LinearInterpolation1D(normSpecWave[ch], normSpecAmp[ch], 0)).eval(l[ch]);
		}
		
		double teTableParam = Math.pow(TeMax, 1.0 / (nEntries-1));
		logTeTableParam = Math.log(teTableParam);
		
		table[0][0] = logTeTableParam;
		for(int ch=0; ch < nCh; ch++)
			table[ch+1][0] = (double)crcs[ch].getValue();
			
		//Scattering scat = new Scattering();
		double Te = 1.0;
		for(int i=0;i<nEntries;i++){
			Te *= teTableParam;
			for(int ch=0; ch < nCh; ch++)
				for(int j=0; j < l[ch].length; j++){ // integrate w.r.t dl
					if(l[ch][j] > 0 && detSpect[ch][j] > 0){
						table[0][i+1] = Te;						
						table[ch+1][i+1] += detSpect[ch][j] * S(Te, ll, l[ch][j], Math.PI) * dl[ch];						
					}
				}

			if(i % (nEntries/50) == 0)
				System.out.print(".");
		}

		System.out.println(" Done.");
		
		if(tableCacheFile != null){
			String fileName = tableCacheFile.replaceAll("%SIGNALS_CACHE_PATH%", SignalDesc.getDefaultCachePath());
			try{
				BinaryMatrixFile.write(fileName, table, true);
			}catch(IOException e){
				System.err.println("Unable to write to thomson table cache file '"+fileName+"':");
				e.printStackTrace();
			}
		}
	
	}
	
	@Override
	public void tidyUpState() {
		// TODO Auto-generated method stub
		super.tidyUpState();
		
		table = null;
	}
	
	/** Returns the scattering intensity
	 * @param Te Electron temperature / eV
	 * @param ll (lambda_laser) - laser input wavelength / any units
	 * @param l - (lambda) scattering wavelength / any units
	 * @param theta - scattering angle / rad
	 * @return
	 */
	public final static double S(double Te, double ll, double l, double theta){
		double ep = (l - ll) / ll;
		double alpha = me * c*c / (2 * Te * e);
		
		double u = sin(theta)  / (1 - cos(theta));
		
		double x = sqrt(1.0 + ep*ep / (2.0 * (1.0-cos(theta)) * (1.0 + ep)) );
		double y = 1.0 / sqrt(x*x + u*u);
		
		double eta = y / (2.0 * alpha);
		double zeta = x * y;
		
		double bes;
		
		//from IDL thomson code
		//bes = Math.sqrt(3.1416/(4*alpha)) /
		//			(  1- (15.0/(16.0 * alpha)) + (345.0 / (512.0*alpha*alpha)));
		
		//from chain1's JETSP.F - seems to match Scilab's besselk2() better but not much difference for < 30KeV
		bes = Math.sqrt(3.1416/(4*alpha)) * 
					(  1 + (15.0/(16.0 * alpha)) + (105.0/(512.0*alpha*alpha)));

		
		double Sz = exp(-2.0 * alpha * (x - 1.0)) / 
//					(2.0 * bes *(1.0+ep)*(1.0+ep)*(1.0+ep));
					(2.0 * bes *(1.0+ep)*(1.0+ep));
		
		
		Sz = Sz / sqrt(2.0 * (1.0 - cos(theta)) * (1+ep) + ep*ep);
		//Sz = Sz * 1e9;

		//double q00 =  1 - 4 * eta * zeta;
		//double S00 = Sz * q00;
		
		double nom = 2.0 * zeta - (2.0 - 3.0 * zeta*zeta) * eta;
		double den = 2.0 * zeta - (2.0 - 15.0 * zeta*zeta) * eta;
		double q11 =  1.0 - 4.0 * eta * zeta * nom / den;
		double S11 = Sz * q11;
		
		/* double p0 = 4.0 + 30.0*zeta*zeta - 55.0*pow(zeta,4);
		double q0 = p0;
		double p1 = -zeta * (24 - 545.0 *zeta*zeta + 720.0 *pow(zeta,4));
		double p2 = 2.0 * (33.0 - 165.0 * zeta*zeta + 240.0 * pow(zeta,4) + 100.0 * pow(zeta,6));
		double q1 = 25.0 * pow(zeta,3) * (29.0 - 42.0*zeta*zeta);
		double q2 = 5.0 * (18.0 - 66.0*zeta*zeta + 630.0 * pow(zeta,4) - 805.0*pow(zeta,6));
		nom = p0 + p1 * eta + p2 * eta*eta;
		den = q0 + q1 * eta + q2 * eta*eta;
		double q22 =  1.0 - 4.0 * eta * zeta * nom / den;
		double S22 = Sz * q22;*/

		return S11; //use (1,1) approximation		
	}

	@Override
	public double[][] getDoubleMatrix() { update(); return table; }
	

}
